
-- msf.lua
-- mission scripting foundation
--
-- дата создания: 14.11.2012
-- автор: Седов А.Г. (BAntDit)

Factory = {
	setBaseClass = function (class, baseClass) 
		setmetatable(class, {__index = baseClass});
	end, 
	create = function (class, parameters) 
		local newObject = {}
		setmetatable(newObject, {__index = class});
		
		if (nil ~= newObject.construct) then
			newObject:construct(parameters);
		end
		return newObject;
	end
}

mission = {
	["model"] = {
		["flags"] = {},
		["phases"] = {},
		["sounds"] = {},
		["groups"] = {},
		["units"] = {},
		["zones"] = {},
		["cargoPlaces"] = {},
		["airDefenceGroupments"] = {},
		["coalitions"] = {
			["red"] = {
				["messages"] = {
					["queue"] = {},
					["push"] = function(_message)
						table.insert(mission.model.coalitions.red.messages.queue, _message);
					end,
					["fetch"] = function()
						return table.remove(mission.model.coalitions.red.messages.queue, 1);
					end
				}
			},
			["blue"] = {
				["messages"] = {
					["queue"] = {},
					["push"] = function(_message)
						table.insert(mission.model.coalitions.blue.messages.queue, _message);
					end,
					["fetch"] = function()
						return table.remove(mission.model.coalitions.blue.messages.queue, 1);
					end
				}
			}
		},
		["messages"] = {
			["queue"] = {},
			["push"] = function(message) 
				table.insert(mission.model.messages.queue, message);
			end,
			["fetch"] = function()
				return table.remove(mission.model.messages.queue, 1);
			end
		},
		["pickupingThreads"] = {}
	},
	["view"] = {
		["updateFrequency"] = 1,
		["maxMessageDelay"] = 10,
		["displayQueueSize"] = 4,
		["frameCounter"] = 0,
		["frameCounter_2"] = 0,
		["hasMessageBeenAdded"] = false,
		["displayTextQueue"] = {},
		["displayTextGroupsQueues"] = {},
		["displayTextCoalitionsQueues"] = {
			["red"] = {
				["hasMessageBeenAdded"] = false,
				["frameCounter_2"] = 0,
				["frameCounter"] = 0,
				["messages"] = {} -- to display
			},
			["blue"] = {
				["hasMessageBeenAdded"] = false,
				["frameCounter_2"] = 0,
				["frameCounter"] = 0,
				["messages"] = {} -- to display
			}
		},
		["displayText"] = function()
			-- all:
			local _new_message_for_all = '';
			local _new_red_coalition_message = '';
			local _new_blue_coalition_message = '';
			local _has_coalition_message_added = false;
			local _has_red_coalition_message_added = false;
			local _has_blue_coalition_message_added = false;
			local _frame_counter = mission.view.frameCounter;
			
			if not mission.view.hasMessageBeenAdded then
				if mission.view.frameCounter_2 < mission.view.maxMessageDelay then 
					mission.view.frameCounter_2 = mission.view.frameCounter_2 + 1;
				else
					mission.view.frameCounter_2 = 0;
					table.remove(mission.view.displayTextQueue, 1);
				end
			end
		
			if mission.view.frameCounter < mission.view.updateFrequency then 
				mission.view.frameCounter = mission.view.frameCounter + 1;
			else
				mission.view.frameCounter = 0;
				mission.view.hasMessageBeenAdded = false;
				
				local _new_message = mission.model.messages.fetch();
				
				if (nil ~= _new_message and '' ~= _new_message) then 
					table.insert(mission.view.displayTextQueue, _new_message);
					mission.view.hasMessageBeenAdded = true;
					
					_new_message_for_all = _new_message;
				end
				
				if (#mission.view.displayTextQueue > mission.view.displayQueueSize) then 
					table.remove(mission.view.displayTextQueue, 1);
				end
				
				local _display_text = '';
				
				for _message_index = 1, #mission.view.displayTextQueue do
					if '' ~= _display_text then 
						_display_text = mission.view.displayTextQueue[_message_index] .. '\n\r------\n\r\n\r' .. _display_text;
					else
						_display_text = mission.view.displayTextQueue[_message_index];
					end
				end
				
				if nil ~= _display_text and '' ~= _display_text then 
					trigger.action.outText(_display_text, 20);
				end
			end
			
			-- coalitions:
			for _coalition_key, _coalition in pairs(mission.view.displayTextCoalitionsQueues) do				
				if not _coalition.hasMessageBeenAdded then 
					if _coalition.frameCounter_2 < mission.view.maxMessageDelay then 
						_coalition.frameCounter_2 = _coalition.frameCounter_2 + 1;
					else
						_coalition.frameCounter_2 = 0;
						table.remove(_coalition.messages, 1);
					end
				end
				
				if _coalition.frameCounter < mission.view.updateFrequency then 
					_coalition.frameCounter = _coalition.frameCounter + 1;
				else
					_coalition.frameCounter = 0;
					_coalition.hasMessageBeenAdded = false;
					
					local _new_message = '';
					
					if not mission.view.hasMessageBeenAdded then						
						if 'red' == _coalition_key then 
							_new_message = mission.model.coalitions.red.messages.fetch();
							_new_red_coalition_message = _new_message;
							if _new_message then _has_red_coalition_message_added = true; end
						end
						if 'blue' == _coalition_key then 
							_new_message = mission.model.coalitions.blue.messages.fetch();
							_new_blue_coalition_message = _new_message;
							if _new_message then _has_blue_coalition_message_added = true; end 
						end
					else
						_new_message = _new_message_for_all;
					end
					
					if (nil ~= _new_message and '' ~= _new_message) then 
						table.insert(_coalition.messages, _new_message);
						_coalition.hasMessageBeenAdded = true;
					
						if (#_coalition.messages < (mission.view.displayQueueSize + 1) and mission.view.hasMessageBeenAdded) then
							if 'red' == _coalition_key then 
								_new_message = mission.model.coalitions.red.messages.fetch();
								_new_red_coalition_message = _new_message;
								if  _new_message then _has_red_coalition_message_added = true; end
							end
							if 'blue' == _coalition_key then 
								_new_message = mission.model.coalitions.blue.messages.fetch();
								_new_blue_coalition_message = _new_message;
								if _new_message then _has_blue_coalition_message_added = true; end
							end
							
							if (nil ~= _new_message and '' ~= _new_message) then 
								table.insert(_coalition.messages, _new_message);
							end
						end
					end
					
					if (#_coalition.messages > mission.view.displayQueueSize) then 
						table.remove(_coalition.messages, 1);
					end
				
					local _display_text = '';
					
					for _message_index = 1, #_coalition.messages do
						if '' ~= _display_text then 
							_display_text = _coalition.messages[_message_index] .. '\n\r------\n\r\n\r' .. _display_text;
						else
							_display_text = _coalition.messages[_message_index];
						end
					end
				
					if nil ~= _display_text and '' ~= _display_text then
						if _coalition_key == 'red' then
							trigger.action.outTextForCoalition(coalition.side.RED, _display_text, 20);
						end
						if _coalition_key == 'blue' then
							trigger.action.outTextForCoalition(coalition.side.BLUE, _display_text, 20);
						end
					end
				end
			end
			
			-- groups:
			for _group_key, _group in pairs(mission.model.groups) do
				local _dcsw_group = _group:getDCSWorldGroup();
				if _dcsw_group then 					
					local _group_coalition = _dcsw_group:getCoalition(); 
					
					if (_has_red_coalition_message_added == true and coalition.side.RED == _group_coalition) 
						or (_has_blue_coalition_message_added == true and coalition.side.BLUE == _group_coalition) then 
						_has_coalition_message_added = true;
					end
					
					if not mission.view.displayTextGroupsQueues[_group_key] then 
						mission.view.displayTextGroupsQueues[_group_key] = {
							hasMessageBeenAdded = false,
							messages = {},
							frameCounter = _frame_counter,
							frameCounter_2 = 0
						};
					end
					if not mission.view.displayTextGroupsQueues[_group_key].hasMessageBeenAdded then
						if mission.view.displayTextGroupsQueues[_group_key].frameCounter_2 < mission.view.maxMessageDelay then
							mission.view.displayTextGroupsQueues[_group_key].frameCounter_2 = 
								mission.view.displayTextGroupsQueues[_group_key].frameCounter_2 + 1;	
						else
							mission.view.displayTextGroupsQueues[_group_key].frameCounter_2 = 0;
							table.remove(mission.view.displayTextGroupsQueues[_group_key].messages, 1);
						end
					end
					
					if mission.view.displayTextGroupsQueues[_group_key].frameCounter < mission.view.updateFrequency then
						mission.view.displayTextGroupsQueues[_group_key].frameCounter = 
							mission.view.displayTextGroupsQueues[_group_key].frameCounter + 1;
					else
						mission.view.displayTextGroupsQueues[_group_key].frameCounter = 0;
						mission.view.displayTextGroupsQueues[_group_key].hasMessageBeenAdded = false;
					
						local _new_message = '';
					
						if not mission.view.hasMessageBeenAdded then						
							if not _has_coalition_message_added then 
								_new_message = _group:recieveMessage();
							else
								if coalition.side.RED == _group_coalition then 
									_new_message = _new_red_coalition_message;
								end
								if coalition.side.BLUE == _group_coalition then 
									_new_message = _new_blue_coalition_message;
								end
							end
						else
							_new_message = _new_message_for_all;
						end
						
						if (nil ~= _new_message and '' ~= _new_message) then 
							table.insert(mission.view.displayTextGroupsQueues[_group_key].messages, _new_message);
							mission.view.displayTextGroupsQueues[_group_key].hasMessageBeenAdded = true;
						
							if (#mission.view.displayTextGroupsQueues[_group_key].messages < (mission.view.displayQueueSize + 1) and mission.view.hasMessageBeenAdded) then
								if _has_coalition_message_added then 
									if coalition.side.RED == _group_coalition then 
										_new_message = _new_red_coalition_message;
									end
									if coalition.side.BLUE == _group_coalition then 
										_new_message = _new_blue_coalition_message;
									end
									if (nil ~= _new_message and '' ~= _new_message) then 
										table.insert(mission.view.displayTextGroupsQueues[_group_key].messages, _new_message);
									end
								end
							end
							
							if (#mission.view.displayTextGroupsQueues[_group_key].messages < (mission.view.displayQueueSize + 1) and 
								(_has_coalition_message_added or mission.view.hasMessageBeenAdded)) then
								_new_message = _group:recieveMessage();

								if (nil ~= _new_message and '' ~= _new_message) then
									table.insert(mission.view.displayTextGroupsQueues[_group_key].messages, _new_message);
								end
							end
						end
						
						if (#mission.view.displayTextGroupsQueues[_group_key].messages > mission.view.displayQueueSize) then 
							table.remove(mission.view.displayTextGroupsQueues[_group_key].messages, 1);
						end
						
						local _display_text = '';
						
						for _message_index = 1, #mission.view.displayTextGroupsQueues[_group_key].messages do
							if '' ~= _display_text then 
								_display_text = mission.view.displayTextGroupsQueues[_group_key].messages[_message_index] .. '\n\r------\n\r\n\r' .. _display_text;
							else
								_display_text = mission.view.displayTextGroupsQueues[_group_key].messages[_message_index];
							end
						end
						
						if nil ~= _display_text and '' ~= _display_text then
							local _dcsw_group_id = _dcsw_group:getID();
						
							trigger.action.outTextForGroup(_dcsw_group_id, _display_text, 20);
						end
					end
				end -- else - раскоментарить при отладке
				--[[	if not _dcsw_status then 
						env.info('MSF: function displayText could not get _dcsw_group: ' .. _group_key, false);
					end
				end --]]
			end
		end,
		["printDebugMessage"] = function(_message)
			if slmod then
				slmod.msg_out(_message, 40, 'chat', 'all');
			end
		end
	},
	["controller"] = {
		["iteration"] = 0,
		["iterationTime"] = 0,
		["mainCycleStorage"] = function(_storag_arguments)
			local _main_cycle_iteration_test = math.abs(timer.getTime() - mission.controller.iteration);
			if (main_cycle_iteration_test > 100) then 
				trigger.action.outText('.' .. timer.getTime() .. ' Основной цикл востановлен после ошибки.', 5);
				env.info('.' .. timer.getTime() .. ' Основной цикл востановлен после ошибки.', false);
				timer.scheduleFunction(mission.controller.mainCycle, {}, (timer.getTime() + 1)); 
			end
			return (timer.getTime() + 15);
		end,
		["moveOnToWayPoint"] = function(_group_name, _waypoint_id)
			local _wp = mission.model.groups[_group_name]:getWaypointByID(_waypoint_id);
			if _wp then _wp:moveOn();
			else mission.model.messages.push('MSF error: msfWayPoint:moveOn(): could not find Way point with id: ' .. _waypoint_id); end
		end,
		["reachWayPoint"] = function(_group_name, _waypoint_id)
			local _wp = mission.model.groups[_group_name]:getWaypointByID(_waypoint_id);
			if _wp then _wp:reach();
			else mission.model.messages.push('MSF error: msfWayPoint:reach(): could not find Way point with id: ' .. _waypoint_id); end
		end,
		["sendMessageForAll"] = function(_message)
			mission.model.messages.push(_message);
		end,
		["sendMessageToCoalition"] = function(_coalition, _message)
			if coalition.side.RED == _coalition then 
				mission.model.coalitions.red.messages.push(_message);
			end
			
			if coalition.side.BLUE == _coalition then
				mission.model.coalitions.blue.messages.push(_message);
			end
		end,
		["sendMessageToGroup"] = function(_group_name, _message)
			if not mission.model.groups[_group_name] then 
				mission.model:addGroup(_group_name);
			end
			
			if not mission.model.groups[_group_name] then
				env.error('MSF: Could not send message to ' .. _group_name .. '. The Group does not exist.', false)
				return;
			else 
				local _dcsw_group = mission.model.groups[_group_name]:getDCSWorldGroup();
				local _dcsw_group_status, _dcsw_group_units = pcall(function() return _dcsw_group:getUnits(); end);
				if not _dcsw_group_status then 
					env.info('MSF: Could not send message to ' .. _group_name .. '. The Group has not been created yet or does not exists.', false);
					return;
				else
					if #_dcsw_group_units < 1 then 
						env.info('MSF: Could not send message to ' .. _group_name .. '. The Group have already been destroyed.', false);
						return;
					end
				end
			end
			table.insert(mission.model.groups[_group_name].messagesQueue, _message);
		end
	},
	["utils"] = {
		["bitops"] = {},
		["getDistance"] = function(_point_1, _point_2)
			return math.sqrt(((_point_1.x - _point_2.x)^2) + ((_point_1.z - _point_2.z)^2));
		end,
		["getHeading"] = function(_position)
			local _rad_heading = -1 * math.asin(_position.x.z);
			local _deg_heading = math.deg(_rad_heading);
			local _do_not_adjust_heading = false;
			
			if _deg_heading < 0 and _position.x.x >= 0 and not _do_not_adjust_heading then _deg_heading = -_deg_heading; _do_not_adjust_heading = true; end
			if _deg_heading > 0 and _position.x.x >= 0 and not _do_not_adjust_heading then _deg_heading = 360 - _deg_heading; _do_not_adjust_heading = true; end
			if _position.x.x < 0 and not _do_not_adjust_heading then _deg_heading = 180 + _deg_heading; _do_not_adjust_heading = true; end
			
			return _deg_heading;
		end,
		["transformPoint"] = function(_position, _local_point)
			return { x = _local_point.x * _position.x.x + _local_point.y * _position.x.y + _local_point.z * _position.x.z + 1 * _position.p.x, 
				y = _local_point.x * _position.y.x + _local_point.y * _position.y.y + _local_point.z * _position.y.z + 1 * _position.p.y,
				z = _local_point.x * _position.z.x + _local_point.y * _position.z.y + _local_point.z * _position.z.z + 1 * _position.p.z };
		end,
		["getLogic"] = function(_probability)
			local _random_value = math.random();
			if (_random_value < _probability) then return true; end
			return false;
		end, 
		["getEnumItemNameById"] = function(_enum_item_id, _enum)
			for _item_key, _item_id in pairs(_enum) do
				if _enum_item_id == _item_id then return _item_key;
				end
			end
			if nil == _enum_item_id then 
				env.error('could not get enum item name by id, cause _enum_item_id has nil value', true);
			else
				env.error('could not get enum item name by id: ' .. _enum_item_id, true);
			end
			return "Unknown";
		end,
		["clampAngle"] = function(_angle)
			_angle = _angle % 360;
			if _angle < 0 then _angle = _angle + 360; end
			return _angle;
		end,
		["getRCS"] = function(_unit_type, _has_external_weapon)
			if 'A-10' == _unit_type:sub(1, 4) or 'MiG-25' == _unit_type:sub(1, 6) or 'MiG-31' == _unit_type then 
				if _has_external_weapon then return 18; 
				else return 10; end 
			end
			if 'F-4E' == _unit_type or 'Su-17M4' == _unit_type or 'Su-25' == _unit_type:sub(1, 5) then 
				if _has_external_weapon then return 14; 
				else return 7; end 
			end
			if 'F-5E' == _unit_type or 'MiG-27K' == _unit_type then if _has_external_weapon then return 10; else return 5; end end
			if 'Su-27' == _unit_type or 'Su-33' == _unit_type then if _has_external_weapon then return 9; else return 5.5; end end
			if 'MiG-23MLD' == _unit_type or 'Tornado' == _unit_type:sub(1, 7) then if _has_external_weapon then return 9; else return 4; end end
			if 'B-1B' == _unit_type then if _has_external_weapon then return 5; else return 5; end end
			if 'С-130' == _unit_type or 'С-17A' == _unit_type or 'E-3A' == _unit_type or 'KC-135' == _unit_type or 'A-50' == _unit_type 
				or 'IL-76MD' == _unit_type or 'IL-78M' == _unit_type then 
				if _has_external_weapon then return 80; 
				else return 80; end 
			end
			if 'S-3B' == _unit_type or 'Mi-26' == _unit_type or 'CH-47D' == _unit_type then if _has_external_weapon then return 30; else return 30; end end
			if 'E-2D' == _unit_type or 'An-26B' == _unit_type or 'An-30M' == _unit_type or 'YAK-40' == _unit_type then if _has_external_weapon then return 50; else return 50; end end
			if 'S-3B Tanker' == _unit_type then if _has_external_weapon then return 55; else return 55; end end
			if 'F-117A' == _unit_type then if _has_external_weapon then return 0.01; else return 0.01; end end
			if 'F-14A' == _unit_type or 'Su-30' == _unit_type or 'Su-34' == _unit_type then if _has_external_weapon then return 12; else return 6; end end
			if 'F-15C' == _unit_type or 'F-15E' == _unit_type or 'F/A-18C' == _unit_type or 'Mirage 2000-5' == _unit_type then 
				if _has_external_weapon then return 10; 
				else return 5; end 
			end
			if 'F-16' == _unit_type:sub(1, 4) then  if _has_external_weapon then return 8; else return 4; end end
			if 'RQ-1A Predator' == _unit_type then if _has_external_weapon then return 1; else return 0.65; end end
			if 'L-39ZA' == _unit_type then if _has_external_weapon then return 6; else return 2.5; end end
			if 'MiG-29' == _unit_type:sub(1, 6) then if _has_external_weapon then return 9; else return 5; end end
			if 'Su-24' == _unit_type:sub(1, 5) then if _has_external_weapon then return 12; else return 7.5; end end
			if 'Tu-160' == _unit_type then if _has_external_weapon then return 15; else return 15; end end
			if 'Tu-22M3' == _unit_type then if _has_external_weapon then return 60; else return 60; end end
			if 'P-51D' == _unit_type or 'FW-190D9' == _unit_type then if _has_external_weapon then return 3; else return 2.5; end end
			if 'Ka-27' == _unit_type then if _has_external_weapon then return 18; else return 15; end end
			if 'Ka-50' == _unit_type:sub(1, 5) or 'Mi-28N' == _unit_type or 'AH-64' == _unit_type:sub(1, 5)  then 
				if _has_external_weapon then return 8; 
				else return 5; end 
			end
			if 'Mi-24' == _unit_type:sub(1, 5) or 'Mi-8' == _unit_type:sub(1, 4) then if _has_external_weapon then return 14; else return 12; end end
			if 'UH-1H' == _unit_type or 'SH-60B' == _unit_type or 'UH-60A' == _unit_type then if _has_external_weapon then return 11; else return 10; end end
			if 'CH-53E' == _unit_type then if _has_external_weapon then return 25; else return 25; end end
			if 'OH-58D' == _unit_type then if _has_external_weapon then return 6; else return 5; end end
			if 'AH-1W' == _unit_type then if _has_external_weapon then return 8; else return 7; end end
			return 100;
		end,
		["getDetectionRangeForASRadar"] = function(_dcsw_unit, _sensor_desc, _ewr)
			if not (_sensor_desc.detectionDistanceAir and _sensor_desc.detectionDistanceAir.upperHemisphere
				and _sensor_desc.detectionDistanceAir.upperHemisphere.headOn) then return 0;
			end
			local _distance0 = _sensor_desc.detectionDistanceAir.upperHemisphere.headOn;
			local _ammo = _dcsw_unit:getAmmo();
			local _has_external_weapon = false;
			local _ewr_type = _ewr:getTypeName();
			local _unit_point, _ewr_point = _dcsw_unit:getPoint(), _ewr:getPoint();
			local _altitude = math.abs(_unit_point.y - land.getHeight({ x = _unit_point.x, y = _unit_point.z }));
			
			if _altitude < 5 then return 0; end -- !!!
			if _altitude > 50 then _altitude = math.abs(_unit_point.y - _ewr_point.y); end
			if not _ammo then _ammo = {}; end 
			
			for _ammo_item_index, _ammo_item in ipairs(_ammo) do
				if (_ammo_item.desc.category == Weapon.Category.MISSILE and _ammo_item.count > 0) 
					or (_ammo_item.desc.category == Weapon.Category.ROCKET and _ammo_item.count > 0)
					or (_ammo_item.desc.category == Weapon.Category.BOMB and _ammo_item.count > 0) then 
					_has_external_weapon = true; break;
				end
			end
			local _rcs = mission.utils.getRCS(_dcsw_unit:getTypeName(), _has_external_weapon);
			local _distance = _distance0 * math.pow(_rcs, 0.25); 
		
			if 'EWR 1L13' == _ewr_type then 
				if _altitude >= 5 and _altitude < 100 then if _distance > 25000 then _distance = 25000; end return _distance; end
				if _altitude >= 100 and _altitude < 500 then if _distance > 51000 then _distance = 51000; end return _distance; end
				if _altitude >= 500 and _altitude < 10000 then if _distance > 273000 then _distance = 273000; end return _distance; end
				if _altitude >= 500 and _altitude < 10000 then if _distance > 273000 then _distance = 273000; end return _distance; end
				if _altitude >= 10000 and _altitude < 27000 then if _distance > 400000 then _distance = 400000; end return _distance; end
				if _altitude >= 27000 then return 0; end
			end
			if 'EWR 55G6' == _ewr_type then 
				if _altitude >= 5 and _altitude < 500 then if _distance > 67500 then _distance = 67500; end return _distance; end
				if _altitude >= 500 and _altitude < 3000 then if _distance > 170000 then _distance = 170000; end return _distance; end
				if _altitude >= 3000 and _altitude < 10000 then if _distance > 310000 then _distance = 310000; end return _distance; end
				if _altitude >= 10000 and _altitude < 20000 then if _distance > 400000 then _distance = 400000; end return _distance; end
				if _altitude >= 20000 and _altitude < 50000 then if _distance > 490000 then _distance = 490000; end return _distance; end
				if _altitude >= 50000 then return 0; end
			end
			if 'Dog Ear radar' == _ewr_type then 
				if _altitude < 25 then _distance = 0; return _distance; end
				if _altitude >= 25 and _altitude < 500 then if _distance > 22000 then _distance = 22000; end return _distance; end
				if _altitude >= 500 and _altitude < 1000 then if _distance > 34000 then _distance = 34000; end return _distance; end
			end
			if 'SA-11 Buk SR 9S18M1' == _ewr_type then 
				if _altitude >= 5 and _altitude < 10 then if _distance > 30000 then _distance = 30000; end return _distance; end
				if _altitude >= 10 and _altitude < 15 then if _distance > 35000 then _distance = 35000; end return _distance; end
				if _altitude >= 15 and _altitude < 3000 then if _distance > 120000 then _distance = 120000; end return _distance; end
			end
			if 'p-19 s-125 sr' == _ewr_type then 
				if _altitude >= 5 and _altitude < 500 then if _distance > 11000 then _distance = 11000; end return _distance; end
			end
			if 'Kub 1S91 str' == _ewr_type or 'Hawk sr' == _ewr_type then 
				if _altitude < 30 then _distance = 0; return _distance; end
			end
			return _distance;
		end
	}
}

-- sounds:
mission.controller["outSound"] = function(...)
	local _time_lag = 0.5;
	for _sound_key_index = 1, arg.n do
		if mission.model.sounds[arg[_sound_key_index]] then timer.scheduleFunction(function(_sound_file_path)
				trigger.action.outSound(_sound_file_path); return nil;
			end, mission.model.sounds[arg[_sound_key_index]].path, timer.getTime() + _time_lag);
			_time_lag = _time_lag + mission.model.sounds[arg[_sound_key_index]].length;
		end
	end
end;

mission.controller["outSoundForGroup"] = function(_group_name, ...)
	if not mission.model.groups[_group_name] then mission.model:addGroup(_group_name); end
	local _play_list, _sound_length = {}, 0; 
	for _index = 1, arg.n do 
		table.insert(_play_list, arg[_index]);
		if mission.model.sounds[arg[_index]] then _sound_length = _sound_length + mission.model.sounds[arg[_index]].length; end
	end
	mission.model.groups[_group_name]:startAudioReciving(_sound_length + 2);
	mission.model.groups[_group_name]:outSound(unpack(_play_list));
end;

mission.controller["outSoundForCountry"] = function(_country_id, ...)
	local _time_lag = 0.5;
	for _sound_key_index = 1, arg.n do
		if mission.model.sounds[arg[_sound_key_index]] then timer.scheduleFunction(function(_data)
					trigger.action.outSoundForCountry(_data.country, _data.path); return nil;
				end, { path = mission.model.sounds[arg[_sound_key_index]].path, country = _country_id }, timer.getTime() + _time_lag);
			_time_lag = _time_lag + mission.model.sounds[arg[_sound_key_index]].length;
		end
	end
end;

mission.controller["outSoundForCoalition"] = function(_coalition_id, ...)
	local _time_lag = 0.5;
	for _sound_key_index = 1, arg.n do
		if mission.model.sounds[arg[_sound_key_index]] then 
			timer.scheduleFunction(function(_data)
					trigger.action.outSoundForCoalition(_data.coalition, _data.path); return nil;
				end, { path = sounds[arg[_sound_key_index]].path, coalition = _coalition_id }, timer.getTime() + _time_lag);
			_time_lag = _time_lag + sounds[arg[_sound_key_index]].length;
		end
	end
end;
--

-- реализация bitwise операций:
mission.utils.bitops["lshift"] = function(_i32, _n) 
	_i32, _ = math.modf(_i32 / (2^_n)); 
	return _i32; 
end;

mission.utils.bitops["rshift"] = function(_i32, _n) 
	_i32 = _i32 * 2^_n;
	if _i32 > 2^32 - 1 then _i32 = _i32 - (2^32 - 1); end
	return _i32; 
end;

mission.utils.bitops["bnot"] = function(_i32)
	local _i32_result = 0;
	for _i = 1, 32 do
		local _ext_b = 0;
		_i32, _i32_result = _i32 * 2, _i32_result * 2;
		if _i32 > 2^32 - 1 then _ext_b = 0; _i32 = _i32 - (2^32 - 1); 
		else _ext_b = 1; end
		_i32_result = _i32_result + _ext_b;
	end
	return _i32_result; 
end;

mission.utils.bitops["bor"] = function(_i32, ...)
	for _arg_index = 1, arg.n do
        local _mask = arg[_arg_index];
		local _i32_result = 0;
		for _i = 1, 32 do
			local _ext_b1, _ext_b2 = 0, 0;
			_i32, _mask, _i32_result = _i32 * 2, _mask * 2, _i32_result * 2;
			
			if _i32 > 2^32 - 1 then _ext_b1 = 1; _i32 = _i32 - (2^32 - 1); 
			else _ext_b1 = 0; end
			if _mask > 2^32 - 1 then _ext_b2 = 1; _mask = _mask - (2^32 - 1); 
			else _ext_b2 = 0; end
			if 1 == _ext_b1 or 1 == _ext_b2 then _i32_result = _i32_result + 1; end 
		end
		_i32 = _i32_result;
	end
	return _i32; 
end;

mission.utils.bitops["bxor"] = function(_i32, ...)
	local _first_number, _rest_of_numbers = 0, {};
	for _arg_index = 1, arg.n do
		if 1 == _arg_index then _first_number = arg[_arg_index];
		else table.insert(_rest_of_numbers, arg[_arg_index]); end
	end
	
    local _mask, _i32_result = mission.utils.bitops.bor(_first_number, unpack(_rest_of_numbers)), 0;
	
	for _i = 1, 32 do
		local _ext_b1, _ext_b2 = 0, 0;
		_i32, _mask, _i32_result = _i32 * 2, _mask * 2, _i32_result * 2;
			
		if _i32 > 2^32 - 1 then _ext_b1 = 1; _i32 = _i32 - (2^32 - 1); 
		else _ext_b1 = 0; end
		if _mask > 2^32 - 1 then _ext_b2 = 1; _mask = _mask - (2^32 - 1); 
		else _ext_b2 = 0; end
		if _ext_b1 ~= _ext_b2 then _i32_result = _i32_result + 1; end 
	end

	return _i32_result; 
end;

mission.utils.bitops["band"] = function(_i32, ...)
	local _first_number, _rest_of_numbers = 0, {};
	for _arg_index = 1, arg.n do
		if 1 == _arg_index then _first_number = arg[_arg_index];
		else table.insert(_rest_of_numbers, arg[_arg_index]); end
	end
	
	local _mask, _i32_result = mission.utils.bitops.bor(_first_number, unpack(_rest_of_numbers)), 0;
	
	for _i = 1, 32 do
		local _ext_b1, _ext_b2 = 0, 0;
		_i32, _mask, _i32_result = _i32 * 2, _mask * 2, _i32_result * 2;
			
		if _i32 > 2^32 - 1 then _ext_b1 = 1; _i32 = _i32 - (2^32 - 1); 
		else _ext_b1 = 0; end
		if _mask > 2^32 - 1 then _ext_b2 = 1; _mask = _mask - (2^32 - 1); 
		else _ext_b2 = 0; end
		if 1 == _ext_b1 and 1 == _ext_b2 then _i32_result = _i32_result + 1; end 
	end
	
	return _i32_result; 
end;

local _msf_lshift, _msf_rshift = mission.utils.bitops.lshift, mission.utils.bitops.rshift;
local _msf_not, _msf_or = mission.utils.bitops.bnot, mission.utils.bitops.bor;  
local _msf_and, _msf_xor = mission.utils.bitops.band, mission.utils.bitops.bxor;
-----------------------------------------------------------------------------------------

function mission.controller:setTimeout(_functionToCall, _argument, _timeout)
	return timer.scheduleFunction(_functionToCall, _argument, _timeout);
end; 

mission.utils["getGuidanceAtAir"] = function(_from, _to)
	local _position_status, _positions = pcall(function()
			if not (_from and _to and _from:isExist() and _to:isExist()) then return nil; end
			local _position_from_to = { from = _from:getPosition(), to = _to:getPosition() };
			return _position_from_to;
		end);	
	if not _position_status or not _positions then 
		env.info('MSF Info: mission.utils.getGuidanceAtAir() could not get guidance, couse of: ' .. tostring(_positions), false);
		return 0, 0, 0, 0;
	end
	
	local _velocity_status, _velocities = pcall(function()
			if not (_from and _to and _from:isExist() and _to:isExist()) then return nil; end
			local _velocity_from_to = { from = _from:getVelocity(), to = _to:getVelocity() };
			return _velocity_from_to;
		end);
	if not _velocity_status or not _velocities then 
		env.info('MSF Info: mission.utils.getGuidanceAtAir() could not get closing speed, couse of: ' .. tostring(_velocities), false);
		return 0, 0, 0, 0;
	end

	local _guidance_v = { x = _positions.to.p.x - _positions.from.p.x, y = _positions.to.p.y - _positions.from.p.y, z = _positions.to.p.z - _positions.from.p.z };
	local _distance, _range = math.sqrt(_guidance_v.x^2 + _guidance_v.y^2 + _guidance_v.z^2),  math.sqrt(_guidance_v.x^2 + _guidance_v.z^2);
	_guidance_v.x, _guidance_v.y, _guidance_v.z = _guidance_v.x / _distance, _guidance_v.y / _distance, _guidance_v.z / _distance;
	
	local _rad_heading = -1 * math.asin(_guidance_v.z);
	local _deg_heading = math.deg(_rad_heading);
	local _do_not_adjust_heading = false;
			
	if _deg_heading < 0 and _guidance_v.x >= 0 and not _do_not_adjust_heading then _deg_heading = -_deg_heading; _do_not_adjust_heading = true; end
	if _deg_heading > 0 and _guidance_v.x >= 0 and not _do_not_adjust_heading then _deg_heading = 360 - _deg_heading; _do_not_adjust_heading = true; end
	if _guidance_v.x < 0 and not _do_not_adjust_heading then _deg_heading = 180 + _deg_heading; _do_not_adjust_heading = true; end

	local _from_heading = mission.utils.getHeading(_positions.from);
	
	local _abs_closing_speed = math.sqrt((_velocities.from.x + _velocities.to.x)^2 + (_velocities.from.y + _velocities.to.y)^2);
	local _changed_range = mission.utils.getDistance(
		{ x = _positions.from.p.x + _velocities.from.x, y = _positions.from.p.y + _velocities.from.y, z = _positions.from.p.z + _velocities.from.z },
		{ x = _positions.to.p.x + _velocities.to.x, y = _positions.to.p.y + _velocities.to.y, z = _positions.to.p.z + _velocities.to.z });
			
	if _changed_range > _range then _abs_closing_speed = _abs_closing_speed * -1; end
	
	return _range, (_deg_heading - _from_heading), (_positions.to.p.y - _positions.from.p.y), _abs_closing_speed;
end;

mission.utils["getGuidanceAt"] = function(_from, _to)
	local _status, _positions = pcall(function()
			if not (_from and _to and _from:isExist() and _to:isExist()) then return nil; end
			local _position_from_to = { from = _from:getPosition(), to = _to:getPosition() };
			return _position_from_to;
		end);	
	if not _status or not _positions then 
		env.info('MSF Info: mission.utils.getGuidanceAt() could not get guidance, couse of: ' .. tostring(_positions), false);
		return 0, 0;
	end
	
	local _guidance_v = { x = _positions.to.p.x - _positions.from.p.x, y = _positions.to.p.y - _positions.from.p.y, z = _positions.to.p.z - _positions.from.p.z };
	local _distance, _range = math.sqrt(_guidance_v.x^2 + _guidance_v.y^2 + _guidance_v.z^2), math.sqrt(_guidance_v.x^2 + _guidance_v.z^2);
	_guidance_v.x, _guidance_v.y, _guidance_v.z = _guidance_v.x / _distance, _guidance_v.y / _distance, _guidance_v.z / _distance;
	
	local _rad_heading = -1 * math.asin(_guidance_v.z);
	local _deg_heading = math.deg(_rad_heading);
	local _do_not_adjust_heading = false;
			
	if _deg_heading < 0 and _guidance_v.x >= 0 and not _do_not_adjust_heading then _deg_heading = -_deg_heading; _do_not_adjust_heading = true; end
	if _deg_heading > 0 and _guidance_v.x >= 0 and not _do_not_adjust_heading then _deg_heading = 360 - _deg_heading; _do_not_adjust_heading = true; end
	if _guidance_v.x < 0 and not _do_not_adjust_heading then _deg_heading = 180 + _deg_heading; _do_not_adjust_heading = true; end

	return _range, _deg_heading;
end;

mission.utils["getPYR"] = function(_unit_position)
	local _rad_yaw = -1 * math.asin(_unit_position.x.z);
	local _deg_yaw = math.deg(_rad_yaw);
	local _c = math.cos(_rad_yaw);
	local _deg_pitch = 0;
	local _deg_roll = 0;
	local _roll_side = 0;
	local _do_not_adjust_yaw = false;
	local _do_not_adjust_roll = false;
	local _dd_not_adjust_pitch = false;
	
	if math.abs(_rad_yaw) > 0.001 then 
		local _trx = _unit_position.z.z / _c;
		local _try = -1 * (_unit_position.y.z / _c);
		
		local _rad_roll = math.atan2(_try, _trx);
		
		_deg_roll = math.abs(math.deg(_rad_roll) % 360);
		
		
		local _trx = _unit_position.x.x / _c;
		local _try = -1 * (_unit_position.x.y / _c);
		
		local _rad_pitch = math.atan2(_try, _trx);
		
		_deg_pitch = mission.utils.clampAngle(math.deg(_rad_pitch));
	else
		_deg_roll = 0;
		local _trx = _unit_position.y.y;
		local _try = _unit_position.x.y;
		local _rad_pitch = math.atan2(_try, _trx);
		_deg_pitch = mission.utils.clampAngle(math.deg(_rad_pitch));
	end
	if _deg_yaw < 0 and _unit_position.x.x >= 0 and not _do_not_adjust_yaw then _deg_yaw = -_deg_yaw; _do_not_adjust_yaw = true; end
	if _deg_yaw > 0 and _unit_position.x.x >= 0 and not _do_not_adjust_yaw then _deg_yaw = 360 - _deg_yaw; _do_not_adjust_yaw = true; end
	if _unit_position.x.x < 0 and not _do_not_adjust_yaw then _deg_yaw = 180 + _deg_yaw; _do_not_adjust_yaw = true; end
	if _deg_roll > 180 and _unit_position.x.x >= 0 and not _do_not_adjust_roll then _deg_roll = 360 - _deg_roll; _do_not_adjust_roll = true; end
	if _unit_position.x.x < 0 and not _do_not_adjust_roll then _deg_roll = math.abs(_deg_roll - 180); _do_not_adjust_roll = true; end
	if _deg_pitch > 90 and _deg_pitch < 180 then _deg_pitch = 90 - (_deg_pitch % 90); end
	if _deg_pitch == 180 or _deg_pitch == 360 then _deg_pitch = 0; end
	if _deg_pitch > 180 and _deg_pitch < 270 then _deg_pitch = -1 * (_deg_pitch % 90); end
	if _deg_pitch == 270 then _deg_pitch = -90; end
	if _deg_pitch > 270 and _deg_pitch < 360 then _deg_pitch = (_deg_pitch % 90) - 90; end
	
	return -_deg_pitch, _deg_yaw, _deg_roll;
end;

mission.utils["getCargoTemplateByGroupName"] = function(_group_name, _offset_point, _cargo_template)
	local _dcsw_group = Group.getByName(_group_name);

	if not (_dcsw_group and _dcsw_group:isExist()) then return; 
	end
	
	local _category = _dcsw_group:getCategory();
	local _units = _dcsw_group:getUnits();
	
	if #_units < 1 then return; end; 
	
	local _coalition_data = nil;
	local _country_data = nil;
	local _country_id = _units[1]:getCountry();
	local _coalition = coalition.getCountryCoalition(_country_id);
	
	if coalition.side.RED == _coalition then 
		_coalition_data = env.mission.coalition.red;
	end
	if coalition.side.BLUE == _coalition then
		_coalition_data = env.mission.coalition.blue;
	end
	if not _coalition_data then 
		env.error('MSF Error: mission.utils.getCargoTemplateByGroupName() could not get coalition data', false);
		error('MSF Error: mission.utils.getCargoTemplateByGroupName() could not get coalition data');
	end
	
	for _country_index, _country in pairs(_coalition_data.country) do
		if ((_country.id + 0) == (_country_id + 0)) then 
			_country_data = _country; break;
		end
	end
	
	if not _country_data then 
		env.error('MSF Error: mission.utils.getCargoTemplateByGroupName() could not get country data', false);
		error('MSF Error: mission.utils.getCargoTemplateByGroupName() could not get country data');
	end
	
	if Group.Category.GROUND == _category then
		local _group_data = nil;
		for _group_index, _group in ipairs(_country_data.vehicle.group) do
			if _group.name == _group_name then _group_data = _group; break;
			end
		end
		if not _group_data then 
			env.error('MSF Error: mission.utils.getCargoTemplateByGroupName() could not get group data', false);
			error('MSF Error: mission.utils.getCargoTemplateByGroupName() could not get group data');
		end
		
		local _initial_point = { x = 0, y = 0 };
		
		_cargo_template.units = {};
		_unload_time = 0;
		for _unit_template_index, _unit_template in ipairs(_group_data.units) do
			local _player_can_drive = false;
			if 'true' == _unit_template.playerCanDrive then _player_can_drive = true; end
			
			if #_cargo_template.units < 1 then 
				_initial_point.x, _initial_point.y  = _unit_template.x, _unit_template.y;
			end
			local _x = 0;
			local _y = 0;
			
			_x, _y = (_offset_point.x + (_unit_template.x - _initial_point.x)), (_offset_point.y + (_unit_template.y - _initial_point.y));
			_unload_time = _unload_time + 2;
			
			table.insert(_cargo_template.units, { unitType = _unit_template.type, playerCanDrive = _player_can_drive, pointOffset = { x = _x, y = _y } });
		end
		_cargo_template.unloadTime = _unload_time;
		_cargo_template.groupName = _group_name;
		_cargo_template.cargoPlace = { point = { x = 100, y = 100, z = 100 }, radius = 20 };
		_cargo_template.countryId = _country_id;
		_cargo_template.conditions = { altitudeLessThen = 5, speedLessThen = 4, command = 0 };
	else
		env.error('MSF Error: mission.utils.getCargoTemplateByGroupName() not implemented for group category: ' .. mission.utils.getEnumItemNameById(_category, Group.Category), false);
		error('MSF Error: mission.utils.getCargoTemplateByGroupName() not implemented for group category: ' .. mission.utils.getEnumItemNameById(_category, Group.Category));
	end
end;

-- groups descriptor: 
msfGroupDescriptor = {
	["route"] = {},
    ["tasks"] = {},
	["units"] = {}
};

function msfGroupDescriptor:construct(parameters)
	self.visible = false;
	self.taskSelected = true;
	self.route.spans = {};
	self.route.points = {};
	
	table.insert(self.route.points, { 
		alt = parameters.point.y,  
		type = "Turning Point",
		ETA = 0, 
		["alt_type"] = "BARO",
		formation_template = "", 
		y = parameters.point.z, 
		x = parameters.point.x,
		ETA_locked = true, 
		speed = 5, 
		action = "Off Road",
		task = { id = 'ComboTask', params = { tasks = {} } },
		speed_locked = true
	});
	self.groupId = 1; 
	self.tasks = {};
	self.hidden = true;
	self.units = {};
	
	local _dx = 0;
	local _dy = 0;
	
	for _index = 1, #parameters.units do 
		
		local _y = 0;
		local _x = 0;
		
		if parameters.units[_index].point then 
			_x, _y = parameters.units[_index].point.x, parameters.units[_index].point.y;
		else
			_x, _y = (parameters.point.x + _dx), (parameters.point.z + _dy);
		end
		
		table.insert(self.units, {
			y = _y, 
			type = parameters.units[_index].unitType,
			name = parameters.units[_index].unitName,
			unitId = _index,
			heading = 0,
			playerCanDrive = parameters.units[_index].playerCanDrive,
			skill = 'Average',
			x = _x
		});
		
		_dx = _dx + 2;
		_dy = _dy + 2; 
	end
	
	self.x = parameters.point.x;
	self.y = parameters.point.z;
	self.name = parameters.groupName;
	self.start_time = 0; 
	self.task = 'Ground Nothing';
end;

mission.utils["getGroupDescriptor"] = function(_group_name, _point, _units)
	return Factory.create(msfGroupDescriptor, { groupName = _group_name, point = _point, units = _units });
end;

mission.controller["removeRadioCommandsForTransport"] = function(_transport_unit)
	if not _transport_unit:hasTransportTask() then return; end
	local _dcsw_transport_unit = _transport_unit:getDCSWorldUnit();
	if not _dcsw_transport_unit then return; end
	
	local _dcsw_transport_group = _dcsw_transport_unit:getGroup();
	if not (_dcsw_transport_group and _dcsw_transport_group:isExist()) then return; end
	
	if Group.Category.GROUND ~= _dcsw_transport_group:getCategory() then return; end
	
	if _transport_unit:isExistRadioMenuItem('msfCargo_' .. _transport_unit:getUnitName() .. '_pickUpGroup') then 
		_transport_unit:removeRadioCommandByName('msfCargo_' .. _transport_unit:getUnitName() .. '_pickUpGroup');
	end	
	if _transport_unit:isExistRadioMenuItem('uloadCargoRadioCommand') then 
		_transport_unit:removeRadioCommandByName('uloadCargoRadioCommand');
	end	
	if _transport_unit:isExistRadioMenuItem('msfCargo_' .. _dcsw_transport_group:getName() .. '_' .. _dcsw_transport_unit:getName() .. '_linkGroupsToPickUp') then 
		_transport_unit:removeRadioCommandByName('msfCargo_' .. _dcsw_transport_group:getName() .. '_' .. _dcsw_transport_unit:getName() .. '_linkGroupsToPickUp');
	end
end

mission.controller["cancelCommunicationWithInfantry"] = function(_transport_unit) -- transport unit, for ground units only
	local _dcsw_transport_unit = _transport_unit:getDCSWorldUnit();
	if not _dcsw_transport_unit then return; end
	
	local _dcsw_transport_group = _dcsw_transport_unit:getGroup();
	if not (_dcsw_transport_group and _dcsw_transport_group:isExist()) then return; end
	
	if _transport_unit:isExistRadioMenuItem('msfCargo_' .. _transport_unit:getUnitName() .. '_pickUpGroup') then
		_transport_unit:removeRadioCommandByName('msfCargo_' .. _transport_unit:getUnitName() .. '_pickUpGroup');
	end
	if not _transport_unit:isExistRadioMenuItem('msfCargo_' .. _dcsw_transport_group:getName() .. '_' .. _dcsw_transport_unit:getName() .. '_linkGroupsToPickUp') then
		_transport_unit:addRadioCommand('msfCargo_' .. _dcsw_transport_group:getName() .. '_' .. _dcsw_transport_unit:getName() .. '_linkGroupsToPickUp',
			_dcsw_transport_unit:getTypeName() .. ' (' .. _dcsw_transport_unit:getName() .. '): связаться с пехотой', false, mission.controller.getGroupsToPickUp, _transport_unit, nil);
	end
end;

mission.controller["getGroupsToPickUp"] = function(_transport_unit) -- transport unit, for ground units only
	local _dcsw_transport_unit = _transport_unit:getDCSWorldUnit();
	if not _dcsw_transport_unit then return; end
	
	local _dcsw_transport_group = _dcsw_transport_unit:getGroup();
	
	_transport_unit:removeRadioCommandByName('msfCargo_' .. _dcsw_transport_group:getName() .. '_' .. _dcsw_transport_unit:getName() .. '_linkGroupsToPickUp');
	
	local _unit_cargo = _transport_unit:getCargo();
	
	if _transport_unit:hasTransportTask() and not _unit_cargo.units then 
		local _pickup_group_submenu = nil;
		local _unit_type = _dcsw_transport_unit:getTypeName();
		
		if _transport_unit:isExistRadioMenuItem('msfCargo_' .. _transport_unit:getUnitName() .. '_pickUpGroup') then
			_transport_unit:removeRadioCommandByName('msfCargo_' .. _transport_unit:getUnitName() .. '_pickUpGroup');
		end
		_pickup_group_submenu = _transport_unit:addRadioCommand('msfCargo_' .. _transport_unit:getUnitName() .. '_pickUpGroup', 
			_unit_type .. ' (' .. _transport_unit:getUnitName() .. '): Подобрать группу', true, nil, nil, nil);
		
		--env.info('MSF Debug: mission.controller.getGroupsToPickUp(): renew command msfCargo_' .. _transport_unit:getUnitName() .. '_pickUpGroup', false);
		
		local _found_groups = {};
		local _transport_unit_point = _dcsw_transport_unit:getPoint();
		
		world.searchObjects(Object.Category.UNIT, { id = world.VolumeType.SPHERE, params = {
					point = _transport_unit_point,
					radius = 500} },
			function(_object, _data)											
				env.info('MSF Debug: enter to search function', false);
				env.info('MSF Debug: new object has been found: ' .. _object:getName(), false);
				if _object:isExist() then 	
					local _object_descriptor = _object:getDesc();
					if Unit.Category.GROUND_UNIT == _object_descriptor.category then
						local _found_ground_unit = _object;
						local _found_ground_group = _found_ground_unit:getGroup();
						local _found_group_name = _found_ground_group:getName();
						
						if nil ~= mission.model.groups[_found_group_name] and nil == _data.foundGroups[_found_group_name] then 
							if  mission.model.groups[_found_group_name]:doesTransportSuit(_data.tarnsportType) then 
								-- если транспорт подходит, то можно добавить радио комманду на перевозку:
								if not _data.transportUnit:isExistRadioMenuItem('msfCargo_toPickup' .. _found_group_name) then
									_data.transportUnit:addRadioCommand('msfCargo_toPickup' .. _found_group_name, _found_group_name, false, 
										mission.controller.requestToPickupGroup, { 
											pickupingGroup = mission.model.groups[_found_group_name], 
											pickupingUnit = _found_ground_unit,  
											transport = _data.transportUnit }, _data.pickupGroupsSubMenu);
									_data.foundGroups[_found_group_name] = _found_group_name;
								end
								_data.foundGroups[_found_group_name] = _found_group_name;
							end -- does transport suit
						end
					end
				end
				local _count_found_groups = 0;
				for _found_group_key, _found_group in pairs(_data.foundGroups) do
					_count_found_groups = _count_found_groups + 1;
				end
				if _count_found_groups > 4 then return false; end
				return true;
			end, { 
				foundGroups = _found_groups, 
				pickupGroupsSubMenu = _pickup_group_submenu, 
				transportUnit = _transport_unit, 
				tarnsportType = _unit_type
			});
		
		-- меню отмены подбора группы:
		_transport_unit:addRadioCommand('msfCargo_' .. _transport_unit:getUnitName()  .. '_closeCommunicationWithInfantry',
			'Конец связи', false, mission.controller.cancelCommunicationWithInfantry, _transport_unit, _pickup_group_submenu);
	end
end;

mission.controller["updatePickupingThread"] = function(_thread_data) -- _pickuping_thread_data
	local _dcsw_pickuping_group = _thread_data.pickupingGroup:getDCSWorldGroup();
	local _dcsw_transport = _thread_data.transport:getDCSWorldUnit();
	
	if _dcsw_transport then 
		local _dcsw_transport_group = _dcsw_transport:getGroup();
		
		if _thread_data.pickupingThread.attemptTime > 120 then 
			trigger.action.outTextForGroup(_dcsw_transport_group:getID(), 'Не удалось подобрать группу, время попытки превысило допустимое', 10);
			return true;
		end
		if _dcsw_pickuping_group then
			local _pickuping_units = _dcsw_pickuping_group:getUnits();
			
			if #_pickuping_units < 1 then 
				if _thread_data.pickupingThread.loadingTime > 0 then 
					trigger.action.outTextForGroup(_dcsw_transport_group:getID(), 'Не удалось подобрать группу - группа уничтожена.', 10);
				end
				return true; -- завершение процесса подбора группы из-за её уничтожения
			end
			
			local _transport_point = _dcsw_transport:getPoint();
			local _transport_point_height = land.getHeight({ x = _transport_point.x, y = _transport_point.z });
			
			if math.abs(_transport_point.y - _transport_point_height) > 4.5 and Group.Category.GROUND ~= _dcsw_transport_group:getCategory() then 
				if _thread_data.pickupingThread.loadingTime > 0 then 
					_thread_data.pickupingThread.loadingTime = 0;
					trigger.action.outTextForGroup(_dcsw_transport_group:getID(), 'Посадка группы прервана. Высота не удерживается.', 10);
				end
				_thread_data.pickupingThread.attemptTime = _thread_data.pickupingThread.attemptTime + _thread_data.delay;
				return false;
			end
			
			for _dcsw_unit_index, _dcsw_unit in ipairs(_pickuping_units) do
				local _distance = mission.utils.getDistance(_transport_point, _dcsw_unit:getPoint());
				if _distance > 100 then
					if _thread_data.pickupingThread.loadingTime > 0 then 
						_thread_data.pickupingThread.loadingTime = 0;
						trigger.action.outTextForGroup(_dcsw_transport_group:getID(), 'Посадка группы прервана. Значительное отклонение от точки подбора.', 10);
					end
					_thread_data.pickupingThread.attemptTime = _thread_data.pickupingThread.attemptTime + _thread_data.delay;
					return false;
				end
			end
			
			-- все условия подбора выполнены:
			if _thread_data.pickupingThread.loadingTime == 0 then 
				trigger.action.outTextForGroup(_dcsw_transport_group:getID(), 'Начало посадки группы ' 
					.. _dcsw_pickuping_group:getName() .. ' на транспорт: ' .. _dcsw_transport_group:getName(), _thread_data.delay);
			else
				local _load_time = (#_pickuping_units) * 2;
				if _load_time > _thread_data.pickupingThread.loadingTime then 
					trigger.action.outTextForGroup(_dcsw_transport_group:getID(), 'Продолжается посадка группы. Осталось ' 
						.. (_load_time - _thread_data.pickupingThread.loadingTime) .. ' секунд.', _thread_data.delay);
				else -- посдка группы:
					local _cargo = _thread_data.pickupingGroup:getGroupTemplate();
					
					if not _cargo.units then mission.utils.getCargoTemplateByGroupName(_dcsw_pickuping_group:getName(), { x = 8, y = 8 }, _cargo); 
					end
					if not _cargo.units then env.info('MSF Info: could not get group cargo template.', false); return true; 
					end
					
					local _unit_index = 0;
					local _new_units_templates = {};
					
					for _dcsw_unit_key, _dcsw_unit in pairs(_pickuping_units) do
						_unit_index = _unit_index + 1;
						table.insert(_new_units_templates, { unitType = _cargo.units[_unit_index].unitType, 
							playerCanDrive = _cargo.units[_unit_index].playerCanDrive,
							pointOffset = { x = _cargo.units[_unit_index].pointOffset.x, y = _cargo.units[_unit_index].pointOffset.y } });
						_dcsw_unit:destroy();
					end
					
					_thread_data.transport:loadCargo({ 	
						units = _new_units_templates, 	
						unloadTime = _unit_index * 2,
						groupName = _dcsw_pickuping_group:getName(),
						cargoPlace = { point = _cargo.cargoPlace.point, radius = _cargo.cargoPlace.radius },
						countryId = _cargo.countryId,	
						conditions = { 
							altitudeLessThen = _cargo.conditions.altitudeLessThen,
							speedLessThen = _cargo.conditions.speedLessThen, 
							command = 0 } });
					
					trigger.action.outTextForGroup(_dcsw_transport_group:getID(), 'Посадка группы выполнена.', 5);
					return true;
				end
			end
		else
			env.info('MSF Info: Group pickuping thread had been interrupted, couse it left access to pickuping group.', false);
			return true;
		end
	else
		env.info('MSF Info: Group pickuping thread had been interrupted, couse it left access to transport unit.', false);
		return true; 
	end
			
	_thread_data.pickupingThread.loadingTime = _thread_data.pickupingThread.loadingTime + _thread_data.delay;	
	_thread_data.pickupingThread.attemptTime = _thread_data.pickupingThread.attemptTime + _thread_data.delay;
	return false;
end;

mission.controller["requestToPickupGroup"] = function(_data) -- pickupingGroup, pickupingUnit,  transport
	local _dcsw_trnasport = _data.transport:getDCSWorldUnit();	
	if not _dcsw_trnasport then 
		env.info('MSF Info: pickup group task failed couse of transport unit ' .. _data.transport:getUnitName() .. ' does not exists', false);
		return;
	end
	local _transport_point = _dcsw_trnasport:getPoint();
	local _transport_desc = _dcsw_trnasport:getDesc();
	local _transport_position = _dcsw_trnasport:getPosition();
	local _transport_group = _dcsw_trnasport:getGroup();
	
	local _pickuping_unit_exist_status, _pickuping_unit_exist = pcall(function()
			local _does_object_exist = _data.pickupingUnit:isExist();
			return _does_object_exist;
		end);
	if not _pickuping_unit_exist_status or not _pickuping_unit_exist then 
		trigger.action.outTextForGroup(_transport_group:getID(), 'Группа не выходит на связь. Отмена задачи подбора группы', 5);
		return;
	end
	local _pickuping_group_point = _data.pickupingUnit:getPoint();
	
	if _data.transport:isExistRadioMenuItem('msfCargo_' .. _data.transport:getUnitName() .. '_pickUpGroup') then -- удалить радиоменю подбора группы
		_data.transport:removeRadioCommandByName('msfCargo_' .. _data.transport:getUnitName() .. '_pickUpGroup'); --'msfCargo_' .. _unit_key .. '_pickUpGroup'
	end
	
	local _distance = mission.utils.getDistance(_transport_point, _pickuping_group_point);
	
	if _distance > 30 then 
		_data.pickupingGroup:atEase();
		
		-- вычисление точки посадки: 
		local _left_door_point = mission.utils.transformPoint(_transport_position, { x = 0, y = 0, z = (_transport_desc.box.min.z - 0.75) });
		local _right_door_point = mission.utils.transformPoint(_transport_position, { x = 0, y = 0, z = (_transport_desc.box.max.z + 0.75) });
		local _left_distance = mission.utils.getDistance(_left_door_point, _pickuping_group_point);
		local _right_distance = mission.utils.getDistance(_right_door_point, _pickuping_group_point);
		local _wp_pickuping_point = { x = _transport_point.x + 8, y = _transport_point.z + 8 }; 
		
		if _right_distance <= _left_distance then _wp_pickuping_point = { x = _right_door_point.x, y = _right_door_point.z };
		else _wp_pickuping_point = { x = _left_door_point.x, y = _left_door_point.z };
		end
		
		-- конец вычисления точки посадки:
		local _pickuping_area_wp = _data.pickupingGroup:addWayPoint('pickuping area', _wp_pickuping_point.x, _wp_pickuping_point.y);
		
		_pickuping_area_wp:addOnReachEventHandler(function(sender, eventArgs) -- выполнить погрузку
				table.insert(mission.model.pickupingThreads, { pickupingGroup = eventArgs.arrivedGroup, transport = _data.transport, 
					pickupingThread = { attemptTime = 0, loadingTime = 0 }, delay = 2 });
			end);
		
		local _pickuping_group = _data.pickupingGroup:getDCSWorldGroup();
		local _pickuping_units = _pickuping_group:getUnits();
		local _pickuping_group_speed = -1;
		local _lock_speed = false;
		
		-- определение возможной скорости группы:
		for _unit_index, _unit in ipairs(_pickuping_units) do
			local _unit_desc = _unit:getDesc();
			
			if nil ~= _unit_desc.speedMax then 
				if -1 == _pickuping_group_speed then _pickuping_group_speed = _unit_desc.speedMax;
				else if _pickuping_group_speed > _unit_desc.speedMax then _pickuping_group_speed = _unit_desc.speedMax; end
				end
			end
			if _pickuping_group_speed > 0 then _lock_speed = true;
			else _pickuping_group_speed = 5; _lock_speed = false; end
		end
		
		if _pickuping_group_speed > 0 then
			_data.pickupingGroup:moveTo(_pickuping_area_wp:getUniqueID(), 5, _pickuping_group_speed, AI.Task.VehicleFormation.OFF_ROAD, _lock_speed);
		else
			if Group.Category.GROUND == _transport_group:getCategory() then 
				trigger.action.outTextForGroup(_transport_group:getID(), 'Группа не может выдвинуться к точке подбора. Необходимо подъехать ближе', 5);
			else 
				trigger.action.outTextForGroup(_transport_group:getID(), 'Группа не может выдвинуться к точке подбора. Необходимо подлететь ближе', 5); 
			end
		end
		_data.transport:requestToPickUp(_data.pickupingGroup);
		return;
	end
	_data.transport:requestToPickUp(_data.pickupingGroup);
	table.insert(mission.model.pickupingThreads, { pickupingGroup = _data.pickupingGroup, transport = _data.transport, 
		pickupingThread = { attemptTime = 0, loadingTime = 0 }, delay = 2 });
end;

-- flags: 

msfFlag = {
}

function msfFlag:construct(parameters)
	self.name = parameters.flagName;
	self.number = parameters.flagNumber;
	self.value = parameters.flagValue;
	trigger.action.setUserFlag(parameters.flagNumber, parameters.flagValue);
	
	self.onChange = Factory.create(msfEvent, {});
	self.onFlagOn = Factory.create(msfEvent, {});
	self.onFlagOff = Factory.create(msfEvent, {});
end;

function msfFlag:getNumber()
	return self.number;
end;

function msfFlag:getName()
	return self.name;
end;

function msfFlag:update()
	self.value = trigger.misc.getUserFlag(self.number);
end;

function msfFlag:getValue()
	return self.value;
end;

function msfFlag:setValue(_flag_value)
	trigger.action.setUserFlag(self.number, _flag_value);
end;

function msfFlag:setTrue()
	trigger.action.setUserFlag(self.number, true);
end;

function msfFlag:setFalse()
	trigger.action.setUserFlag(self.number, false);
end;

function msfFlag:increase()
	local _flag_value = trigger.misc.getUserFlag(self.number);
	
	if _falg_value then 
		_flag_value = _flag_value + 1;
	else
		_flag_value = 1;
	end
	trigger.action.setUserFlag(self.number, _flag_value);
end;

function msfFlag:decrease()
	local _flag_value = trigger.misc.getUserFlag(self.number);
		
	if _flag_value > 1 then 
		_flag_value = _flag_value - 1;
	else
		_flag_value = 0;
	end
	trigger.action.setUserFlag(self.number, _flag_value);
end;

function msfFlag:addValue(_value_to_add)
	local _flag_value = trigger.misc.getUserFlag(self.number);
	
	if _falg_value then 
		_flag_value = _flag_value + _value_to_add;
	else
		_flag_value = _value_to_add;
	end
	trigger.action.setUserFlag(self.number, _flag_value);
end;

function msfFlag:removeValue(_value_to_remove)
	local _flag_value = trigger.misc.getUserFlag(self.number);
	
	if _falg_value > _value_to_remove then 
		_flag_value = _flag_value - _value_to_remove;
	else
		_flag_value = 0;
	end
	trigger.action.setUserFlag(self.number, _flag_value);
end;

function msfFlag:change(eventArgs)
	for _onChangeEventHandlerKey, _onChangeEventHandler in pairs(self.onChange) do
		_onChangeEventHandler(self, eventArgs);
	end
end;

function msfFlag:flagOn(eventArgs)
	for _onFlagOnEventHandlerKey, _onFlagOnEventHandler in pairs(self.onFlagOn) do
		_onFlagOnEventHandler(self, eventArgs);
	end
end;

function msfFlag:flagOff(eventArgs)
	for _onFlagOffEventHandlerKey, _onFlagOffEventHandler in pairs(self.onFlagOff) do
		_onFlagOffEventHandler(self, eventArgs);
	end
end;

function msfFlag:addOnFlagOffEventHandler(eventHandler)
	self.onFlagOff[eventHandler] = eventHandler;
end;

function msfFlag:removeOnFlagOffEventHandler(eventHandler)
	self.onFlagOff[eventHandler] = nil;
end;

function msfFlag:addOnFlagOnEventHandler(eventHandler)
	self.onFlagOn[eventHandler] = eventHandler;
end;

function msfFlag:removeOnFlagOnEventHandler(eventHandler)
	self.onFlagOn[eventHandler] = nil;
end;

function msfFlag:addOnChangeEventHandler(eventHandler)
	self.onChange[eventHandler] = eventHandler;
end;

function msfFlag:removeOnChangeEventHandler(eventHandler)
	self.onChange[eventHandler] = nil;
end;

-- tasks:
msfTask = {
}

function msfTask:construct(parameters) 
	self.functionToCall = parameters.functionToCall;
	self.argument = parameters.functionArgument;
	self.id = timer.scheduleFunction(self.functionToCall, self.argument, parameters.startTime);
end;

function msfTask:Stop()
	pcall(function() timer.removeFunction(self.id); return 0; end);
end;

function msfTask:Resum(newTime)
	self.id = timer.scheduleFunction(self.functionToCall, self.argument, newTime);
end;

function msfTask:ResumWithArgument(argument, newTime)
	self.argument = argument;
	self.id = timer.scheduleFunction(self.functionToCall, self.argument, newTime);
end;

function msfTask:StartImmediately()
	timer.setFunctionTime(self.id, (timer.getTime() + 1));
end;

function msfTask:StartAfter(delay)
	-- mission.model.messages.push('Отладка: перепланирование задачи ИД: ' .. self.id .. ', старт через: ' .. delay  .. ' секунд.');
	timer.setFunctionTime(self.id, (timer.getTime() + delay));
end;

function msfTask:GetId() return self.id; end;

-- events:
msfEvent = {
}

-- mission phase description:
msfMissionPhaseDescription = {
};

function msfMissionPhaseDescription:construct(parameters)
	self.before = parameters.before;
	self.during = parameters.during;
	self.after = parameters.after;
	self.task = parameters.task;
end;

function msfMissionPhaseDescription:getBefore() return self.before; end;
function msfMissionPhaseDescription:getDuring() return self.during; end;
function msfMissionPhaseDescription:getAfter() return self.after; end;
function msfMissionPhaseDescription:getTask() return self.task; end;

function msfMissionPhaseDescription:setBefore(_before) self.before = _before; end;
function msfMissionPhaseDescription:setDuring(_during) self.during = _during; end;
function msfMissionPhaseDescription:setAfter(_after) self.after = _after; end;
function msfMissionPhaseDescription:setTask(_task) self.task = _task; end;

-- misson phases
msfMissionPhase = {
};

function msfMissionPhase:construct(parameters)
	self.isPhaseActivated = false;
	self.hasPhaseBegun = false;
	self.hasPahseDone = false;
	self.onActivate = Factory.create(msfEvent, {});
	self.onDeactivate = Factory.create(msfEvent, {});
	self.description = Factory.create(msfMissionPhaseDescription, parameters.description);
	self.tasks = {};
end;

function msfMissionPhase:addOnDeactivateEventHandler(eventHandler)
	local _nextEventHandlerIndex = (#self.onDeactivate + 1);
	self.onDeactivate[_nextEventHandlerIndex] = eventHandler;
end;

function msfMissionPhase:addOnActivateEventHandler(eventHandler)
	local _nextEventHandlerIndex = (#self.onActivate + 1);
	self.onActivate[_nextEventHandlerIndex] = eventHandler;
end;

function msfMissionPhase:deactivate(eventArgs)
	self.isPhaseActivated = false;
	self.hasPahseDone = true;

	for _eventHandlerIndex = 1, #self.onDeactivate do
		self.onDeactivate[_eventHandlerIndex](self, eventArgs);
	end
	
	for _taskIndex = 1, #self.tasks do
		self.task[_taskIndex]:Stop();
	end
end;

function msfMissionPhase:activate(eventArgs)
	
	if self.isPhaseActivated then return; end
	
	self.isPhaseActivated = true;
	self.hasPhaseBegun = true;
	
	for _eventHandlerIndex = 1, #self.onActivate do
		self.onActivate[_eventHandlerIndex](self, eventArgs);
	end
	
	for _taskIndex = 1, #self.tasks do
		self.task[_taskIndex]:StartImmediately();
	end
end;

function msfMissionPhase:addTask(_functionToCall, _argument)
	local _nextTaskIndex = (#self.tasks + 1);
	self.tasks[_nextTaskIndex] = Factory.create(msfTask, { functionToCall = _functionToCall, functionArgument = _argument, startTime = 999999 });
end

function msfMissionPhase:isActivated() return self.isPhaseActivated; end;

function msfMissionPhase:getDescription() 
	if ((false == self.hasPhaseBegun) and (false == self.hasPahseDone)) then
		return self.description:getBefore(), nil;
	else 
		if ((true == self.hasPhaseBegun) and (false == self.hasPahseDone)) then 
			return self.description:getDuring(), self.description:getTask(); 
		end
	end
	
	return self.description:getAfter(), nil;
end;

function msfMissionPhase:getDescriptionBefore() return self.description:getBefore(); end;
function msfMissionPhase:getDescriptionDuring() return self.description:getDuring(); end;
function msfMissionPhase:getDescriptionAfter() return self.description:getAfter(); end;
function msfMissionPhase:getDescriptionTask() return self.description:getTask(); end;

function msfMissionPhase:setDescriptionBefore(_description) self.description:setBefore(_description); end;
function msfMissionPhase:setDescriptionDuring(_description) self.description:setDuring(_description); end;
function msfMissionPhase:setDescriptionAfter(_description) self.description:setAfter(_description); end;
function msfMissionPhase:setDescriptionTask(_description) self.description:setTask(_description); end;

function msfMissionPhase:begun() return self.hasPhaseBegun; end;
function msfMissionPhase:done() return self.hasPahseDone; end;

-- unit's wrapper:

function msfOnPrepareToUnloadRadioCommand(_data)
	if _data.cargo.conditions and 0 == _data.cargo.conditions.command then _data.cargo.conditions.command = 1;
	end
	_data.unit:removeRadioCommandByName('uloadCargoRadioCommand');
end;

msfUnit = {
	["InterceptorEventType"] = { NO_TARGET = 1, NEW_TARGET = 2, LOST_TARGET = 3, UPDATE_TARGET = 4 };
}

function msfUnit:construct(parameters) 
	self.dcsWorldUnit = parameters.newUnit;
	self.unitName = parameters.newUnitName;
	self.onHit = Factory.create(msfEvent, {});
	self.onShot = Factory.create(msfEvent, {});
	self.onDead = Factory.create(msfEvent, {});
	self.onPlayerEnterUnit = Factory.create(msfEvent, {});
	self.onPlayerLeaveUnit = Factory.create(msfEvent, {});
	self.onTakeOff = Factory.create(msfEvent, {});
	self.onHitting = Factory.create(msfEvent, {});
	self.onBirth = Factory.create(msfEvent, {});
	self.onUnloadCargo = Factory.create(msfEvent, {});
	self.onLoadCargo = Factory.create(msfEvent, {});
	self.onHover = Factory.create(msfEvent, {});
	self.onEndHover = Factory.create(msfEvent, {});
	self.onRequestToPickUp = Factory.create(msfEvent, {});
	self.onEngineStartUp = Factory.create(msfEvent, {});
	self.detectIntersectionZones = false;
	
	-- for artillery observers only:
	self.onEnemySpotted = Factory.create(msfEvent, {});
	self.onLostVisualContact = Factory.create(msfEvent, {});
	self.cargo = {}; -- conditions, uload_time, units
	self.selectedCargoTemplateName = '';
	self.unloadThread = { uloadingTime = 0, unloadingStatus = false }; -- time, status
	self.hoverThread = { hoverTime = 0, hoverStatus = false, hoverEventRaised = false };
	self.radioCommands = {}; -- command id, path, executed or not, subItem or not, subMenu
	self.isTransport = false;
	
	-- для перехватчиков и истребителей завоевания превосходства: 
	self.interceptor = false;
	self.interceptingThread = nil; -- conext: target, air defence unit, unit owner
	self.onTargeting = Factory.create(msfEvent, {});
	-- 
end;

function msfUnit:getTargetForIntercepting()
	if self.interceptingThread and self.interceptingThread.target.object and self.interceptingThread.target.object:isExist() then 
		return self.interceptingThread.target.object; 
	end
	return nil;
end;

function msfUnit:addOnTargetingEventHandler(_event_handler) self.onTargeting[_event_handler] = _event_handler;
end;

function msfUnit:removeOnTargetingEventHandler(_event_handler) self.onTargeting[_event_handler] = nil;
end;

function msfUnit:setInterceptorTask(_value) self.interceptor = _value; end

function msfUnit:isInterceptor() 
	local _dcsw_unit = self:getDCSWorldUnit(); 
	if not (_dcsw_unit and _dcsw_unit:inAir()) then return false; end
	return self.interceptor; 
end;

function msfUnit:hasTargetForIntercepting()
	if self.interceptingThread then return true; end
	return false;
end;

function msfUnit:endInterceptTask() self.interceptingThread = nil;
end;

function msfUnit:startInterceptTask(_air_defence_groupment, _target)
	if self.interceptingThread then 
		env.info('MSF: could not start intercepting thread for ' .. tostring(self.unitName) .. ', cause other thread has already begun', false);
		return; 
	end
	self.interceptingThread = { airDefenceGroupment = _air_defence_groupment, target = _target, interceptor = self, updateIndex = 1 };
	timer.scheduleFunction(function(_intercept_thread)
			if not _intercept_thread.interceptor:getTargetForIntercepting() then return nil; end
			if not _intercept_thread.interceptor then env.error('MSF Error: intercepting thread lost access to interceptor unit', false);
				return nil; 
			end 
			if not (_intercept_thread.airDefenceGroupment and _intercept_thread.airDefenceGroupment:isOnline()
				and _intercept_thread.airDefenceGroupment:isActivated()) then _intercept_thread.interceptor:endInterceptTask();
				return nil;
			end
			local _dcsw_unit, _dcsw_target = _intercept_thread.interceptor:getDCSWorldUnit(), _intercept_thread.target.object;
			
			if not (_dcsw_unit and _dcsw_unit:inAir()) then _intercept_thread.interceptor:endInterceptTask(); return nil; end
			if not (_dcsw_target and _dcsw_target:isExist()) then
				for _handler_name, _handler in pairs(self.onTargeting) do
					local _status, _error = pcall(function() 
							_handler(_intercept_thread.interceptor, { 
								eventType = msfUnit.InterceptorEventType.NO_TARGET,
								airDefenceGroupment = _intercept_thread.airDefenceGroupment });
							return 0;
						end);
					if not _status then env.error('MSF Error: error on attempt to handle NO_TARGET event for interceptor.\n'
							.. 'Error: ' .. tostring(_error) .. '\n'
							.. 'EventHandler name: ' .. tostring(_handler_name) .. '\n', false);
						mission.controller.sendMesssageForAll('MSF Error: error on attempt to handle NO_TARGET event for interceptor.\n'
							.. 'Error: ' .. tostring(_error) .. '\n'
							.. 'EventHandler name: ' .. tostring(_handler_name) .. '\n');
					end
				end
				_intercept_thread.interceptor:endInterceptTask();
				return nil;
			end
			if not _intercept_thread.airDefenceGroupment:isTargetKnown(_intercept_thread.target.id) then 
				for _handler_name, _handler in pairs(self.onTargeting) do
					local _status, _error = pcall(function() 
							_handler(_intercept_thread.interceptor, { 
								eventType = msfUnit.InterceptorEventType.LOST_TARGET,
								airDefenceGroupment = _intercept_thread.airDefenceGroupment });
							return 0;
						end);
					if not _status then env.error('MSF Error: error on attempt to handle LOST_TARGET event for interceptor.\n'
							.. 'Error: ' .. tostring(_error) .. '\n'
							.. 'EventHandler name: ' .. tostring(_handler_name) .. '\n', false);
						mission.controller.sendMesssageForAll('MSF Error: error on attempt to handle LOST_TARGET event for interceptor.\n'
							.. 'Error: ' .. tostring(_error) .. '\n'
							.. 'EventHandler name: ' .. tostring(_handler_name) .. '\n');
					end
				end
				_intercept_thread.interceptor:endInterceptTask();
				return nil;
			end
			
			local _range, _heading_diff, _height_diff, _closing_speed = mission.utils.getGuidanceAtAir(_dcsw_unit, _dcsw_target);
			local _event_type, _event_type_name = msfUnit.InterceptorEventType.UPDATE_TARGET, 'UPDATE_TARGET';
			
			if 1 == _intercept_thread.updateIndex then _event_type, _event_type_name = msfUnit.InterceptorEventType.NEW_TARGET, 'NEW_TARGET'; end

			for _handler_name, _handler in pairs(self.onTargeting) do
				local _status, _error = pcall(function() 
					_handler(_intercept_thread.interceptor, { 
							eventType = _event_type,
							airDefenceGroupment = _intercept_thread.airDefenceGroupment,
							distance = _range, 
							headingDifference = _heading_diff,
							heightDifference = _height_diff,
							closingSpeed = _closing_speed });
						return 0;
					end);
				if not _status then 
					env.error('MSF Error: error on attempt to handle ' .. _event_type_name .. ' event for interceptor.\n'
						.. 'Error: ' .. tostring(_error) .. '\n'
						.. 'EventHandler name: ' .. tostring(_handler_name) .. '\n', false);
					mission.controller.sendMesssageForAll('MSF Error: error on attempt to handle ' .. _event_type_name .. ' event for interceptor.\n'
						.. 'Error: ' .. tostring(_error) .. '\n'
						.. 'EventHandler name: ' .. tostring(_handler_name) .. '\n');
				end
			end
			
			_intercept_thread.updateIndex = _intercept_thread.updateIndex + 1;
			return timer.getTime() + 16;
		end, self.interceptingThread, timer.getTime() + 0.5); 
end;

mission.controller["selectCargoPlaceTemplate"] = function(_data)
	_data.unit:setSelectedCargoTemplateName(_data.cargoTemplateName);
end;

mission.controller["resetCargoPlaceTemplate"] = function(_data)
	_data.unit:resetSelectedCargoTemplateName();
end;

function msfUnit:update()	
	local _dcsw_unit = self:getDCSWorldUnit();
	if not _dcsw_unit then return 0; end
				
	local _unit_desc = _dcsw_unit:getDesc();
				
	-- поиск доступных для перевозки групп для вертолетов при зависании:
	if Unit.Category.HELICOPTER ~= _unit_desc.category then 
		return 0;
	end
	
	local _unit_point = _dcsw_unit:getPoint();
	local _point_height = land.getHeight({ x = _unit_point.x, y = _unit_point.z } );
	local _unit_speed = self:getSpeedValue();
	local _hover_thread = self:getHoverThread();
				
	if math.abs(_unit_point.y - _point_height) < 4 and _unit_speed < 2 then 
		if _hover_thread.hoverStatus then 
			if _hover_thread.hoverTime > 6 and not _hover_thread.hoverEventRaised then 
				_hover_thread.hoverEventRaised = true;
								
				-- raise hover event:
				self:hover();
								
				-- pick up units:
				local _unit_cargo = self:getCargo();
				local _surface_type_under_unit = land.getSurfaceType({ x = _unit_point.x, y = _unit_point.z } );
								
				if self:hasTransportTask() and not _unit_cargo.units 
					and (land.SurfaceType.LAND == _surface_type_under_unit or land.SurfaceType.ROAD == _surface_type_under_unit
						or land.SurfaceType.RUNWAY == _surface_type_under_unit) then
						
					if _unit_point ~= nil then
						for _cargo_place_index, _cargo_place in ipairs(mission.model.cargoPlaces) do
							local _dcsw_unit_coalition = _dcsw_unit:getCoalition();
							local _distance = mission.utils.getDistance(_unit_point,_cargo_place.point);
							
							if _dcsw_unit_coalition == _cargo_place.coalition and _distance < 150 then -- если коалиция юнита соответствует зоне погрузки
								local _suit_templates_to_add = {};
								self:resetSelectedCargoTemplateName();
								
								for  _cargo_template_index, _cargo_template in ipairs(_cargo_place.cargoTemplates) do
									local _dcsw_unit_type = _dcsw_unit:getTypeName();
									if _dcsw_unit_type == _cargo_template.unitType then -- если тип юнита соответствует шаблону
										table.insert(_suit_templates_to_add, { name = _cargo_template.name, displayName = _cargo_template.displayName;  });
									end
								end
								if #_suit_templates_to_add > 0 then
									local _cargo_place_submenu = nil;
									self:resetSelectedCargoTemplateName();
									if self:isExistRadioMenuItem('msfCargoPlace_' .. 'X') then
										self:removeRadioCommandByName('msfCargoPlace_' .. 'X');
									end
									local _cargo_place_name = string.sub(_cargo_place.name, 1, (string.find(_cargo_place.name, "PGRZ") - 1));
									_cargo_place_submenu = self:addRadioCommand('msfCargoPlace_' .. 'X', _cargo_place_name, true, nil, nil, nil);
									for _cargo_template_to_add_index, _cargo_template_to_add in ipairs(_suit_templates_to_add) do
										if self:isExistRadioMenuItem('msfCargoPlace_' .. 'X' .. '_' .. _cargo_template_to_add.name) then
											self:removeRadioCommandByName('msfCargoPlace_' .. 'X' .. '_' .. _cargo_template_to_add.name);
										end
										self:addRadioCommand('msfCargoPlace_' .. 'X' .. '_' .. _cargo_template_to_add.name, 
											_cargo_template_to_add.displayName, false, 
											mission.controller.selectCargoPlaceTemplate, 
											{ unit = self, cargoTemplateName = _cargo_template_to_add.name }, 
											_cargo_place_submenu);
									end
									if self:isExistRadioMenuItem('msfCargoPlace_' .. 'X' .. '_reset') then
										self:removeRadioCommandByName('msfCargoPlace_' .. 'X' .. '_reset');
									end
									self:addRadioCommand('msfCargoPlace_' .. 'X' .. '_reset', 
										'Отменить выбор типа груза', false,	mission.controller.resetCargoPlaceTemplate,
										{ unit = self }, _cargo_place_submenu);
								end
							end
						end
					end
						
					local _pickup_group_submenu = nil;
					local _unit_type = _dcsw_unit:getTypeName();
									
					if not self:isExistRadioMenuItem('msfCargo_' .. self.unitName .. '_pickUpGroup') then
						_pickup_group_submenu = self:addRadioCommand('msfCargo_' .. self.unitName .. '_pickUpGroup', 'Подобрать группу', true, nil, nil, nil);
					else
						_pickup_group_submenu = self:getRadioMenuItemByName('msfCargo_' .. self.unitName .. '_pickUpGroup');
					end
							
					local _found_groups = {};
					local _group_candidate = { groupObject = nil, groupName = '', groupUnit = nil };
					-- local _debug_point = _unit_point;
									
					world.searchObjects(Object.Category.UNIT, { id = world.VolumeType.SPHERE, params = { -- dont forgot back to SPHERE after fix!!!!
							point = _unit_point,
							radius = 500} }, 
						function(_object, _data)											
							if _object:isExist() then 	
								local _object_descriptor = _object:getDesc();
								if Unit.Category.GROUND_UNIT == _object_descriptor.category then
													
									local _found_ground_unit = _object;
									local _found_ground_group = _found_ground_unit:getGroup();
									local _found_group_name = _found_ground_group:getName();
													
									if nil ~= mission.model.groups[_found_group_name] and nil == _data.foundGroups[_found_group_name] then 
										if  mission.model.groups[_found_group_name]:doesTransportSuit(_data.tarnsportType) then 
															
											local _object_candidate_point = _object:getPoint();
											local _dcsw_transport_unit = _data.transportUnit:getDCSWorldUnit();
											local _dcsw_transport_unit_point = _dcsw_transport_unit:getPoint();
											local _distnace = mission.utils.getDistance(_object_candidate_point, _dcsw_transport_unit_point);
															
											if 0 == _data.minDistance then
												_data.minDistance = _distnace;
												_data.groupCandidate.groupObject = _found_ground_group;
												_data.groupCandidate.groupName = _found_group_name;
												_data.groupCandidate.groupUnit = _found_ground_unit;
											else
												if _data.minDistance > _distnace then 
													_data.minDistance = _distnace;
													_data.groupCandidate.groupObject = _found_ground_group;
													_data.groupCandidate.groupName = _found_group_name;
													_data.groupCandidate.groupUnit = _found_ground_unit;
												end
											end
														
											-- если транспорт подходит, то можно добавить радио комманду на перевозку:
															
											-- !!! - delete after update 1.2.5: temporary fix for clients crashes on attempt to delete more then two items at once
											--[[if not _data.transportUnit:isExistRadioMenuItem('msfCargo_toPickup' .. _found_group_name) then
												env.info('MSF Info: added radio command: msfCargo_toPickup'
													.. _found_group_name .. ' to unit: ' 
													.. _data.transportUnit:getUnitName(), false);
												_data.transportUnit:addRadioCommand('msfCargo_toPickup' .. _found_group_name, _found_group_name, false, 
													mission.controller.requestToPickupGroup, { 
														pickupingGroup = mission.model.groups[_found_group_name], 
														pickupingUnit = _found_ground_unit,  
														transport = _data.transportUnit }, 
														_data.pickupGroupsSubMenu);
														_data.foundGroups[_found_group_name] = _found_group_name;
												end --]]--
											_data.foundGroups[_found_group_name] = _found_group_name;
										end -- does transport suit
									end
								end
							end
											
							local _count_found_groups = 0;
							for _found_group_key, _found_group in pairs(_data.foundGroups) do
								_count_found_groups = _count_found_groups + 1;
							end
											
							-- !!! - delete after update 1.2.5: temporary fix for clients crashes on attempt to delete more then two items as once
							-- if _count_found_groups > 4 then return false; end -- uncomment after 1.2.5 update
							return true;
						end, { 
						foundGroups = _found_groups, 
						pickupGroupsSubMenu = _pickup_group_submenu, 
						transportUnit = self, 
						tarnsportType = _unit_type,
						minDistance = 0,
						groupCandidate = _group_candidate
					});
										
					if _group_candidate.groupObject then
						if not self:isExistRadioMenuItem('msfCargo_toPickup' .. _group_candidate.groupName) then
							self:addRadioCommand('msfCargo_toPickup' .. _group_candidate.groupName, _group_candidate.groupName, false, 
							mission.controller.requestToPickupGroup, { 
								pickupingGroup = mission.model.groups[_group_candidate.groupName], 
								pickupingUnit = _group_candidate.groupUnit,  
								transport = self 
							}, 
							_pickup_group_submenu);
						end
					else
						if self:isExistRadioMenuItem('msfCargo_' .. self.unitName .. '_pickUpGroup') then 
							self:removeRadioCommandByName('msfCargo_' .. self.unitName .. '_pickUpGroup');
						end
					end
					--env.info('MSF Debug: object searching complite', false);
				end	
			end
			_hover_thread.hoverTime = _hover_thread.hoverTime + 2;
		else
			_hover_thread.hoverStatus = true;
			_hover_thread.hoverEventRaised = false;
			_hover_thread.hoverTime = 0;
		end
	else
		if _hover_thread.hoverStatus and _hover_thread.hoverEventRaised then 
			-- raise event: end hover:
			self:endHover();
			self:resetSelectedCargoTemplateName();
			
			-- удалить радиоменю пункта подбора, если пилот ушел с режима висения:
			if self:isExistRadioMenuItem('msfCargoPlace_' .. 'X') then
				self:removeRadioCommandByName('msfCargoPlace_' .. 'X');
			end
						
			-- удалить радиоменю подбора группы, если пилот ушел с режима висения:
			if self:isExistRadioMenuItem('msfCargo_' .. self.unitName .. '_pickUpGroup') then 
				self:removeRadioCommandByName('msfCargo_' .. self.unitName .. '_pickUpGroup');
			end
		end 
		_hover_thread.hoverStatus = false;
		_hover_thread.hoverEventRaised = false;
		_hover_thread.hoverTime = 0;
	end -- check height and speed end
				
	return 0;
end;

function msfUnit:getHoverThread()
	return self.hoverThread;
end

function msfUnit:setTransportTask(_value)
	self.isTransport = _value;
	
	local _self_dcsw_unit = self:getDCSWorldUnit();
	
	if not _self_dcsw_unit then return; end
	local _dcsw_transport_group = _self_dcsw_unit:getGroup();
	
	if not (_dcsw_transport_group and _dcsw_transport_group:isExist()) then return; end
	if not (Group.Category.GROUND == _dcsw_transport_group:getCategory()) then return; end 
	local _dcsw_transport_group_name = _dcsw_transport_group:getName();
	local _dcsw_units = _dcsw_transport_group:getUnits();	
	local _unit_index = 0;
	local _count_transport_unit = 0;
	
	while(function()
			if _unit_index > #_dcsw_units then return false; end
			if _unit_index < 1 then return true; end
			if not (_dcsw_units[_unit_index] and _dcsw_units[_unit_index]:isExist()) then return true; end
			if not mission.model.units[_dcsw_units[_unit_index]:getName()] then return true; end
				
			if mission.model.units[_dcsw_units[_unit_index]:getName()]:hasTransportTask() then _count_transport_unit = _count_transport_unit + 1;
			end
				
		end)() do
		_unit_index = _unit_index + 1;
	end
	if _count_transport_unit > 4 then return; end
	if self.cargo.units then return; end
	
	if self:isExistRadioMenuItem('msfCargo_' .. self.unitName .. '_pickUpGroup', 'Подобрать группу') then
		self:removeRadioCommandByName('msfCargo_' .. self.unitName .. '_pickUpGroup', 'Подобрать группу');
	end
	
	if not self:isExistRadioMenuItem('msfCargo_' .. _dcsw_transport_group_name .. '_' .. self.unitName .. '_linkGroupsToPickUp') then
		self:addRadioCommand('msfCargo_' .. _dcsw_transport_group_name .. '_' .. self.unitName .. '_linkGroupsToPickUp',
			_self_dcsw_unit:getTypeName() .. ' (' .. self.unitName .. '): связаться с пехотой', false, mission.controller.getGroupsToPickUp, self, nil);
	end
end;

function msfUnit:hasTransportTask()
	return self.isTransport;
end;

function msfUnit:setSelectedCargoTemplateName(_cargo_template_name) self.selectedCargoTemplateName = _cargo_template_name; end;
function msfUnit:getSelectedCargoTemplateName() return self.selectedCargoTemplateName; end;
function msfUnit:resetSelectedCargoTemplateName() self.selectedCargoTemplateName = ''; end;

function msfUnit:getRadioMenuItemByName(_command_name)
	for _index, _command in ipairs(self.radioCommands) do
		if _command.name == _command_name then return _command; end
	end
	return nil;
end;

function msfUnit:isExistRadioMenuItem(_command_name)
	for _index, _command in ipairs(self.radioCommands) do
		if _command.name == _command_name then return true; end
	end
	return false;
end;

function msfUnit:refreshRadioMenu()
	for _command_key, _command in pairs(self.radioCommands) do
		if _command.executed then self:removeRadioCommand(_command.id); end
	end
end;

function msfUnit:removeRadioCommand(_command_id)
	local _dcsw_group_status, _dcsw_group_id = pcall(function()
			local _self_dcsw_unit = self:getDCSWorldUnit();
			if _self_dcsw_unit then 
				local _group = _self_dcsw_unit:getGroup();
				local _group_id = _group:getID();
				return _group_id;
			end
			return 0;
		end);
	
	if _dcsw_group_status and nil ~= _dcsw_group_id and '0' ~= (_dcsw_group_id .. '') then
		local _command_to_remove = nil;
		local _index_to_remove = 0;
		for _index, _command in ipairs(self.radioCommands) do
			if _command.id == _command_id then 
				_command_to_remove = _command;
				_index_to_remove = _index;
				break;
			end
		end
		if _index_to_remove == 0 then return; end
		
		if _command_to_remove.isSubMenu then  -- remove all children items at first
			for _command_key, _command in pairs(self.radioCommands) do
				if _command.subMenu and _command.subMenu.id == _command_to_remove.id then self:removeRadioCommand(_command.id);
				end
			end
			missionCommands.removeItemForGroup(_dcsw_group_id, _command_to_remove.path);
			table.remove(self.radioCommands, _index_to_remove);
		else
			missionCommands.removeItemForGroup(_dcsw_group_id, _command_to_remove.path);
			table.remove(self.radioCommands, _index_to_remove);
		end
	else
		if nil ~= _dcsw_group_id and ('' .. _dcsw_group_id) ~= '0' then 
			env.info('MSF Error: msfUnit:removeRadioCommand: could not get group Id for unit ' .. self.unitName, false);
			env.info('MSF Error: msfUnit:removeRadioCommand: ' .. _dcsw_group_id, false);
		end
		if (nil ~= _dcsw_group_id and ('' .. _dcsw_group_id) == '0') or nil == _dcsw_group_id then 
			env.info('MSF Error: msfUnit:removeRadioCommand: could not get group Id for unit ' .. self.unitName, false);
		end
	end
end;

function msfUnit:removeRadioCommandByName(_command_name)
	local _command_id_to_remove = 0;
	for _index, _command in ipairs(self.radioCommands) do
		if _command.name == _command_name then 
			_command_id_to_remove = _command.id; break;
		end
	end
	self:removeRadioCommand(_command_id_to_remove);
end;

function msfUnit:addRadioCommand(_name, _display_name, _is_sub_item, _on_command, _on_command_data, _sub_menu)
	local _dcsw_group_status, _dcsw_group_id = pcall(function()
			local _self_dcsw_unit = self:getDCSWorldUnit();
			if _self_dcsw_unit then 
				local _group = _self_dcsw_unit:getGroup();
				local _group_id = _group:getID();
				return _group_id;
			end
			return 0;
		end);
	
	if _dcsw_group_status and nil ~= _dcsw_group_id and '0' ~= (_dcsw_group_id .. '') then
		local _sub_path = nil;
		local _max_command_index = 0;
		local _new_command_id = 1;
		
		if #self.radioCommands > 0 then 
			_max_command_index = #self.radioCommands;
			_new_command_id = (self.radioCommands[_max_command_index].id + 1);
		end
		
		if _sub_menu then _sub_path = _sub_menu.path; end;
		
		if not _is_sub_item then
			local _path = missionCommands.addCommandForGroup(_dcsw_group_id, _display_name, _sub_path, _on_command, _on_command_data);
			table.insert(self.radioCommands, { id = _new_command_id, name = _name, path = _path, executed = false, isSubMenu = false, subMenu = _sub_menu });
		else
			local _path = missionCommands.addSubMenuForGroup(_dcsw_group_id, _display_name, _sub_path);
			table.insert(self.radioCommands, { id = _new_command_id, name = _name, path = _path, executed = false, isSubMenu = true, subMenu = _sub_menu });
		end
		_max_command_index = #self.radioCommands;
		
		return self.radioCommands[_max_command_index];
	else
		if nil ~= _dcsw_group_id and ('' .. _dcsw_group_id) ~= '0' then 
			env.error('MSF Error: msfUnit:addRadioCommand: could not get group Id for unit ' .. self.unitName, false);
			env.error('MSF Error: msfUnit:addRadioCommand: ' .. _dcsw_group_id, false);
			return nil;
		end
		if (nil ~= _dcsw_group_id and ('' .. _dcsw_group_id) == '0') or nil == _dcsw_group_id then 
			env.error('MSF Error: msfUnit:addRadioCommand: could not get group Id for unit ' .. self.unitName, false);
			return nil;
		end
	end
	return nil;
end;

function msfUnit:clearCargo() 
	self.cargo = {}; 
	self:resetSelectedCargoTemplateName();
	if self:isExistRadioMenuItem('uloadCargoRadioCommand') then self:removeRadioCommandByName('uloadCargoRadioCommand');
	end
end;

function msfUnit:getCargo() return self.cargo; end;

function msfUnit:requestToPickUp(_pickuping_group)
	for _onRequestToPickUpKey, _onRequestToPickUp in pairs(self.onRequestToPickUp) do
		_onRequestToPickUp(self, { pickupingGroup = _pickuping_group });
	end
end;

function msfUnit:hover()
	for _onHoverKey, _onHover in pairs(self.onHover) do
		_onHover(self, {});
	end
end;

function msfUnit:endHover()
	for _onEndHoverKey, _onEndHover in pairs(self.onEndHover) do
		_onEndHover(self, {});
	end
end;

function msfUnit:loadCargo(_cargo) 
	self.cargo = _cargo;
	self.isTransport = true;	
	
	if self.cargo.conditions and 0 == self.cargo.conditions.command and not self:isExistRadioMenuItem('uloadCargoRadioCommand') then 
		local _self_dcsw_unit = self:getDCSWorldUnit();
		if _self_dcsw_unit then 
			local _dcsw_transport_group = _self_dcsw_unit:getGroup();
			
			if _dcsw_transport_group and _dcsw_transport_group:isExist() then 
				if Group.Category.GROUND == _dcsw_transport_group:getCategory() then
					-- перенесено из msfCargoPlace:update()
					-- if Group.Category.GROUND == _group:getCategory() then 
					if self:isExistRadioMenuItem('msfCargo_' .. self:getUnitName() .. '_pickUpGroup') then 
						self:removeRadioCommandByName('msfCargo_' .. self:getUnitName() .. '_pickUpGroup');
					end	
					if self:isExistRadioMenuItem('msfCargo_' .. _dcsw_transport_group:getName() .. '_' .. self:getUnitName() 
						.. '_linkGroupsToPickUp') then 
						self:removeRadioCommandByName('msfCargo_' .. _dcsw_transport_group:getName() .. '_' .. self:getUnitName() 
							.. '_linkGroupsToPickUp');
					end
					-- end
					--
					self:addRadioCommand('uloadCargoRadioCommand', 'Группе ' 
							.. self.cargo.groupName .. ' покинть машину ' .. _self_dcsw_unit:getTypeName() .. ' (' .. self.unitName .. ')', 
						false, msfOnPrepareToUnloadRadioCommand, { unit = self, cargo = _cargo; }, nil);
				else
					self:addRadioCommand('uloadCargoRadioCommand', 
						'Приготовиться к высадке', false, msfOnPrepareToUnloadRadioCommand, { unit = self, cargo = _cargo; }, nil);
				end
			end
		end
	end
	for _onLoadCargoKey, _onLoadCargo in pairs(self.onLoadCargo) do
		_onLoadCargo(self, { cargo = _cargo });
	end
end;

function msfUnit:tryUnloadCargo(_delay) 
	local _dcsw_unit = self:getDCSWorldUnit();
	
	if not _dcsw_unit then 
		return nil; 
	end;
	if not self.cargo.units then 
		return nil; 
	end
	if not self.cargo.conditions then 
		return nil; 
	end 
	if not self.cargo.cargoPlace then
		return nil; 
	end
	
	local _unit_group = _dcsw_unit:getGroup();
	
	if 0 == self.cargo.conditions.command then 
		if self.unloadThread.unloadingStatus then 
			self.unloadThread.unloadingStatus = false;
			self.unloadThread.unloadingTime = 0;
			trigger.action.outTextForGroup(_unit_group:getID(), 'Выполнение выгрузки прервано по комманде пользователя.', _delay);
		end
		return nil;
	end
	
	local _unit_point = _dcsw_unit:getPoint();
	local _distance = math.sqrt(((_unit_point.x - self.cargo.cargoPlace.point.x)^2) + ((_unit_point.z - self.cargo.cargoPlace.point.z)^2));
	
	if _distance < (self.cargo.cargoPlace.radius + 150) then 
		if self.unloadThread.unloadingStatus then 
			self.unloadThread.unloadingStatus = false;
			self.unloadThread.unloadingTime = 0;
		end
		return nil; 
	end
	
	local _surface_type = land.getSurfaceType({ x = _unit_point.x, y = _unit_point.z });
	
	if land.SurfaceType.SHALLOW_WATER == _surface_type or land.SurfaceType.WATER == _surface_type then 
		if self.unloadThread.unloadingStatus then 
			self.unloadThread.unloadingStatus = false;
			self.unloadThread.unloadingTime = 0;
			trigger.action.outTextForGroup(_unit_group:getID(), 'Выполнение выгрузки прервано по причине не допустимой для высадки поверхности.', _delay);
		end
		return nil; 
	end
	
	local _point_height = _unit_point.y;
	if self.cargo.conditions.altitudeLessThen then 
		_point_height = land.getHeight({ x = _unit_point.x, y = _unit_point.z });
		
		local _height = _unit_point.y - _point_height;
		if _height > self.cargo.conditions.altitudeLessThen then
			if self.unloadThread.unloadingStatus then 
				self.unloadThread.unloadingStatus = false;
				self.unloadThread.unloadingTime = 0;
				trigger.action.outTextForGroup(_unit_group:getID(), 'Выполнение выгрузки прервано. Высота не удерживается', _delay);
			end
			return nil;
		end
	end
	
	if self.cargo.conditions.speedLessThen then 
		local _speed = self:getSpeedValue();
		if _speed > self.cargo.conditions.speedLessThen then
			if self.unloadThread.unloadingStatus then 
				self.unloadThread.unloadingStatus = false;
				self.unloadThread.unloadingTime = 0;
				trigger.action.outTextForGroup(_unit_group:getID(), 'Выполнение выгрузки прервано. Скорость выше допустимой.', _delay);
			end
			return nil;
		end
	end
	
	if self.unloadThread.unloadingStatus then 
		if self.unloadThread.unloadingTime < self.cargo.unloadTime then 	
			local _remaining_time = 0;
			if self.cargo.unloadTime > 0 then 
				if self.cargo.unloadTime > self.unloadThread.unloadingTime then 
					_remaining_time = self.cargo.unloadTime - self.unloadThread.unloadingTime;
				end
			end
			trigger.action.outTextForGroup(_unit_group:getID(), 'Продолжается выполнения высадки десанта. Осталось ' .. _remaining_time .. ' секунд.', _delay);
			self.unloadThread.unloadingTime = self.unloadThread.unloadingTime + _delay;
		else
			local _cargo = self.cargo; 
			
			self.cargo = {};
			self.unloadThread.unloadingStatus = false;
			self.unloadThread.unloadingTime = 0;
			
			-- _unit_point
			local _droping_units = {};
			local _group_name = _cargo.groupName;
			
			for _index = 1, #_cargo.units do
				local _x = (_unit_point.x + _cargo.units[_index].pointOffset.x); 
				local _y = (_unit_point.z + _cargo.units[_index].pointOffset.y);				
				
				table.insert(_droping_units, { 
					unitType = _cargo.units[_index].unitType, 
					unitName = _group_name .. '_' .. _index, 
					playerCanDrive = _cargo.units[_index].playerCanDrive, 
					point = { x = _x, y = _y  } });
			end
			
			local _group_descriptor = mission.utils.getGroupDescriptor(_group_name, 
				{ x = _unit_point.x, y = (_point_height + 0.1), z = _unit_point.z }, _droping_units);
			
			local _created_msf_group = mission.model:createGroup(Group.Category.GROUND, _cargo.countryId, _group_descriptor);
			local _created_dcsw_group = _created_msf_group:getDCSWorldGroup();
			
			if nil == _created_dcsw_group then 
				error('MSF Error: could not create new dcsw group');
			end
			if not _created_dcsw_group:isExist() then 
				error('MSF Error: new dcsw group does not exist');
			end
			
			_created_msf_group:setGroupTemplate(_cargo);
			_created_msf_group:addAvalibleTransport(_dcsw_unit:getTypeName());
			
			if Group.Category.GROUND == _unit_group:getCategory() then
				if not self:isExistRadioMenuItem('msfCargo_' .. _unit_group:getName() .. '_' .. self.unitName .. '_linkGroupsToPickUp') then
					self:addRadioCommand('msfCargo_' .. _unit_group:getName() .. '_' .. self.unitName .. '_linkGroupsToPickUp',
						_dcsw_unit:getTypeName() .. ' (' .. self.unitName .. '): связаться с пехотой', false, mission.controller.getGroupsToPickUp, self, nil);
				end
			end
			
			trigger.action.outTextForGroup(_unit_group:getID(), 'Высадка десанта успешно завершена.', _delay);
			
			for _onUnloadCargoKey, _onUnloadCargo in pairs(self.onUnloadCargo) do
				_onUnloadCargo(self, { cargo = _cargo, group = _created_msf_group });
			end
			self:resetSelectedCargoTemplateName();
			return _cargo;
		end
	else
		self.unloadThread.unloadingStatus = true;
		self.unloadThread.unloadingTime = 0;
		trigger.action.outTextForGroup(_unit_group:getID(), 'Начало выполнения высадки десанта.', _delay);		
	end
	return nil; 
end;

function msfUnit:addOnPickUpEventHandler(eventHandler)
	self.onRequestToPickUp[eventHandler] = eventHandler;
end;

function msfUnit:removeOnPickUpEventHandler(eventHandler)
	self.onRequestToPickUp[eventHandler] = nil;
end;

function msfUnit:addOnHoverEventHandler(eventHandler)
	self.onHover[eventHandler] = eventHandler;
end;

function msfUnit:addOnEndHoverEventHandler(eventHandler)
	self.onEndHover[eventHandler] = eventHandler;
end;

function msfUnit:removeOnHoverEventHandler(eventHandler)
	self.onHover[eventHandler] = nil;
end;

function msfUnit:removeOnEndHoverEventHandler(eventHandler)
	self.onHover[eventHandler] = nil;
end;

function msfUnit:addOnUnloadCargoEventHandler(eventHandler)
	self.onUnloadCargo[eventHandler] = eventHandler;
end;

function msfUnit:removeOnUnloadCargoEventHandler(eventHandler)
	self.onUnloadCargo[eventHandler] = nil;
end;

function msfUnit:addOnLoadCargoEventHandler(eventHandler)
	self.onLoadCargo[eventHandler] = eventHandler;
end;

function msfUnit:removeOnLoadCargoEventHandler(eventHandler)
	self.onLoadCargo[eventHandler] = nil;
end;

function msfUnit:getUnitType()
	local _self_dcsw_unit = self:getDCSWorldUnit();
	if not _self_dcsw_unit then return 'тип не определен'; end
	
	local _type_name = _self_dcsw_unit:getTypeName();	
	return _type_name;
end;

function msfUnit:getDammage()
	local _status, _health = pcall(function() return { initial = self.dcsWorldUnit:getLife0(), current = self.dcsWorldUnit:getLife() }; end);
	
	if not _status then return 1; end;
	return (1 - ((_health.current - 1) / _health.initial));
end;

function msfUnit:hasDammage()
	local _status, _health = pcall(function() return { initial = self.dcsWorldUnit:getLife0(), current = self.dcsWorldUnit:getLife() }; end);
	
	if not _status then return false; end;
	if _health.initial > (_health.current - 1) then return true; end;
	return false;
end

function msfUnit:isAlive()
	local _status, _health = pcall(function() return self.dcsWorldUnit:getLife(); end);

	if not _status then return false; end;
	if _health > 1 then return true; end
	return false;
end;

function msfUnit:getUnitID()
	local _self_dcsw_unit = self:getDCSWorldUnit();
	if not _self_dcsw_unit then return '0'; end
	
	local _dcsw_unit_id = _self_dcsw_unit:getID();
	return _dcsw_unit_id;
end;

function msfUnit:getSpeedValue()
	local _self_dcsw_unit = self:getDCSWorldUnit();
	if not _self_dcsw_unit then return 0; end
	
	local _velocity = _self_dcsw_unit:getVelocity();	
	local _speed_value = math.sqrt((_velocity.x^2) + (_velocity.y^2) + (_velocity.z^2));
	
	return _speed_value;
end;

function msfUnit:getAltitude()
	local _self_dcsw_unit = self:getDCSWorldUnit();
	if not _self_dcsw_unit then return 0; end
	
	local _unit_point = _self_dcsw_unit:getPoint();	
	return _unit_point.y;
end

function msfUnit:isUnitInsideZone(_zone)
	if (false == self.detectIntersectionZones) then 
		return false;
	end
	
	local _status, _unit_position = pcall(function() return self.dcsWorldUnit:getPosition(); end);
	
	if not _status then 
		local _dcsw_unit = Unit.getByName(self.unitName);
		_status, _unit_position = pcall(function() return self.dcsWorldUnit:getPosition(); end);
		
		if _status then
			self.dcsWorldUnit = _dcsw_unit;
		else
			return false;
		end
	end
	
	local _unit_distance = math.sqrt(((_unit_position.p.x - _zone.point.x)^2) + ((_unit_position.p.z - _zone.point.z)^2));
	
	if (_unit_distance < _zone.radius) then 
	 	return true;
	end
	
	return false;
end;

function msfUnit:enableDetectionIntersectionOfZones(_is_enbale)
	self.detectIntersectionZones = _is_enbale;
end;

function msfUnit:getDCSWorldUnit() 
	if not (self.dcsWorldUnit and self.dcsWorldUnit:isExist()) then 
		self.dcsWorldUnit = Unit.getByName(self.unitName); 
		if not (self.dcsWorldUnit and self.dcsWorldUnit:isExist()) then return nil; end
	end 
	return self.dcsWorldUnit; 
end;

function msfUnit:setDCSWorldUnit(_unit) self.dcsWorldUnit = _unit; end;

function msfUnit:getUnitName() return self.unitName; end;

function msfUnit:addOnEnemySpotted(eventHandler)
	self.onEnemySpotted[eventHandler] = eventHandler;
end;

function msfUnit:removeOnEnemySpotted(eventHandler)
	self.onEnemySpotted[eventHandler] = nil;
end;

function msfUnit:addOnLostVisualContact(eventHandler)
	self.onLostVisualContact[eventHandler] = eventHandler;
end;

function msfUnit:removeOnLostVisualContact(eventHandler)
	self.onLostVisualContact[eventHandler] = nil;
end;

function msfUnit:addOnTakeOffEventHandler(eventHandler)
	self.onTakeOff[eventHandler] = eventHandler;
end;

function msfUnit:removeOnTakeOffEventHandler(eventHandler)
	self.onTakeOff[eventHandler] = nil;
end;

function msfUnit:addOnPlayerEnterUnit(eventHandler)
	self.onPlayerEnterUnit[eventHandler] = eventHandler;
end;

function msfUnit:removeOnPlayerEnterUnit(eventHandler)
	self.onPlayerEnterUnit[eventHandler] = nil;
end;

function msfUnit:addOnPlayerLeaveUnit(eventHandler)
	self.onPlayerLeaveUnit[eventHandler] = eventHandler;
end;

function msfUnit:removeOnPlayerLeaveUnit(eventHandler)
	self.onPlayerLeaveUnit[eventHandler] = nil;
end;

function msfUnit:addOnHitEventHandler(eventHandler)
	self.onHit[eventHandler] = eventHandler;
end;

function msfUnit:removeOnHitEventHandler(eventHandler)
	self.onHit[eventHandler] = nil;
end;

function msfUnit:addOnShotEventHandler(eventHandler)
	self.onShot[eventHandler] = eventHandler;
end;

function msfUnit:removeOnShotEventHandler(eventHandler)
	self.onShot[eventHandler] = nil;
end;

function msfUnit:addOnDeadEventHandler(eventHandler)
	self.onDead[eventHandler] = eventHandler;
end;

function msfUnit:removeOnDeadEventHandler(eventHandler)
	self.onDead[eventHandler] = nil;
end;

function msfUnit:addOnHittingEventHandler(eventHandler)
	self.onHitting[eventHandler] = eventHandler;
end;

function msfUnit:removeOnHittingEventHandler(eventHandler)
	self.onHitting[eventHandler] = nil;
end;

function msfUnit:addOnBirthEventHandler(eventHandler)
	self.onBirth[eventHandler] = eventHandler;
end;

function msfUnit:removeOnBirthEventHandler(eventHandler)
	self.onBirth[eventHandler] = nil;
end;

function msfUnit:addOnEngineStartUpEventHandler(eventHandler)
	self.onEngineStartUp[eventHandler] = eventHandler;
end;

function msfUnit:removeOnEngineStartUpEventHandler(eventHandler)
	self.onEngineStartUp[eventHandler] = nil;
end;

function msfUnit:engineStartUp(eventArgs)
	for _onEngineStartUpEventHandlerKey, _onEngineStartUpEventHandler in pairs(self.onEngineStartUp) do
		_onEngineStartUpEventHandler(self, eventArgs);
	end
end;

function msfUnit:birth(eventArgs)
	for _onBirthEventHandlerKey, _onBirthEventHandler in pairs(self.onBirth) do
		_onBirthEventHandler(self, eventArgs);
	end
end;

function msfUnit:hit(eventArgs)
	for _onHitEventHandlerKey, _onHitEventHandler in pairs(self.onHit) do
		_onHitEventHandler(self, eventArgs);
	end
end;

function msfUnit:shot(eventArgs)
	for _onShotEventHandlerKey, _onShotEventHandler in pairs(self.onShot) do
		_onShotEventHandler(self, eventArgs);
	end
end;

function msfUnit:dead(eventArgs)
	for _onDeadEventHandlerKey, _onDeadEventHandler in pairs(self.onDead) do
		_onDeadEventHandler(self, eventArgs);
	end
end;

function msfUnit:playerEnterUnit(eventArgs)
	for _onPlayerEnterUnitEventHandlerKey, _onPlayerEnterUnitEventHandler in pairs(self.onPlayerEnterUnit) do
		_onPlayerEnterUnitEventHandler(self, eventArgs);
	end
end;

function msfUnit:playerLeaveUnit(eventArgs)
	for _onPlayerLeaveUnitEventHandlerKey, _onPlayerLeaveUnitEventHandler in pairs(self.onPlayerLeaveUnit) do
		_onPlayerLeaveUnitEventHandler(self, eventArgs);
	end
end;

function msfUnit:takeOff(eventArgs)
	for _onTakeOffEventHandlerKey, _onTakeOffEventHandler in pairs(self.onTakeOff) do
		_onTakeOffEventHandler(self, eventArgs);
	end
end;

function msfUnit:hitting(eventArgs)
	for _onHittingEventHandlerKey, _onHittingEventHandler in pairs(self.onHitting) do
		_onHittingEventHandler(self, eventArgs);
	end
end;

function msfUnit:spot(eventArgs)
	for _onEnemySpottedEventHandler, _onEnemySpottedEventHandlerKey in pairs(self.onEnemySpotted) do
		_onEnemySpottedEventHandler(self, eventArgs);
	end
end;

function msfUnit:lostVisualContact(eventArgs)
	for _onLostVisualContactKey, _onLostVisualContact in pairs(self.onLostVisualContact) do
		_onLostVisualContact(self, eventArgs);
	end
end;

-- air defence groupments:

msfAirDefenceUnitRole = {
	EWR = 0,
	LN = 1,
	AMBUSH = 2,
	COMMAND_POST = 3,
	STAND_ALONE = 4,
	OBSERVATION_POST = 5,
	COMMUNICATION_POST = 6,
	WRECKED_UNIT = 7,
	USER_CONTROLLED_UNIT = 8
};

msfAirDefenceTargetStatus = {
	NO_TARGET = 0,
	TARGET = 1,
	NEW_TARGET = 2
};

msfAirDefenceUnit = {
	["PriorityForIntercepting"] = {
		HIGHEST = 1,
		HIGH = 2,
		MEDIUM = 3,
		LOW = 4
	}, 
	["ThreatFlags"] = { 
		NO_THREAT = 0, 
		THREAT = 1,   			-- 1
		FIGTHER = 2,  			-- 2
		EXIST_TARGET = 4, 		-- 3
		CLOSE_TARGET = 8,		-- 4
		SEAD = 16,				-- 5
		AWACS = 32,				-- 6 
		BOMBER = 64,			-- 7
		STRATEGIC = 128,		-- 8
		CAS = 256,				-- 9
		HELI = 512,				-- 10 
		TRANSPORT = 1024		-- 11
	},
	["getThreatType"] = function(_dcsw_unit)
		local _threat_type, _does_suit_for_intercepting = msfAirDefenceUnit.ThreatFlags.NO_THREAT, false;
		local _priority_for_intercepting, _close_groups = msfAirDefenceUnit.PriorityForIntercepting.LOW, {};
		if not (_dcsw_unit and _dcsw_unit:isExist() and _dcsw_unit:inAir()) then 
			return _threat_type, _does_suit_for_intercepting, _priority_for_intercepting, _close_groups;
		end
		local _unit_desc, _exist_target, _category = _dcsw_unit:getDesc(), false, _dcsw_unit:getCategory();
		
		if Object.Category.UNIT ~= _category then return _threat_type, _does_suit_for_intercepting, _priority_for_intercepting, _close_groups; end
		if _unit_desc.attributes and _unit_desc.attributes['Fighters'] then 			
			_threat_type, _exist_target = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.THREAT, msfAirDefenceUnit.ThreatFlags.FIGTHER), false;
			world.searchObjects(Object.Category.UNIT, {
					id = world.VolumeType.SPHERE, params = { point = _dcsw_unit:getPoint(), radius = 5000 } },
				function(_object, _data)
					if _object:inAir() and _object:getCoalition() ~= _data.fighter:getCoalition() then 
						_exist_target = true;
						_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.CLOSE_TARGET);
						_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.EXIST_TARGET); 
						_priority_for_intercepting = msfAirDefenceUnit.PriorityForIntercepting.HIHG;
						_does_suit_for_intercepting = true;
						
						local _target_group, _object_name = _object:getGroup(), _object:getName();
						local _target_group_name = _target_group:getName();
						if mission.model.groups[_target_group_name] and not _close_groups[_target_group_name] then 
							if mission.model.units[_object_name] and mission.model.units[_object_name]:isInterceptor() then 
								if mission.model.units[_object_name]:hasTargetForIntercepting() then
									local _target_for_intercepting = mission.model.units[_object_name]:getTargetForIntercepting();
									local _target_for_intercepting_id, _object_id = _target_for_intercepting:getID(), _object:getID();
									
									if _target_for_intercepting_id ~= _object_id then 
										local _fighter_heading = mission.utils.getHeading(_data.fighter:getPosition());
										local _distance_to_target, _heading_to_target = mission.utils.getGuidanceAt(_data.fighter, _object);
										if math.abs(_heading_to_target - _fighter_heading) < 40 then 
											_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
											mission.model.units[_object_name]:endInterceptTask();
										end 
									end
								else
									_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
								end
							else 
								_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
							end
						end
					end
					return true;
				end, { fighter = _dcsw_unit  });
			if not _exist_target then
				world.searchObjects(Object.Category.UNIT, {
						id = world.VolumeType.SPHERE, params = { point = _dcsw_unit:getPoint(), radius = 25000 } },
					function(_object, _data)
						if _object:inAir() and _object:getCoalition() ~= _data.fighter:getCoalition() then 
							_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.EXIST_TARGET); 
							_priority_for_intercepting = msfAirDefenceUnit.PriorityForIntercepting.MEDIUM;
							_does_suit_for_intercepting = true;
							
							local _target_group, _object_name = _object:getGroup(), _object:getName();
							local _target_group_name = _target_group:getName();
							if mission.model.groups[_target_group_name] and not _close_groups[_target_group_name] then 
								if mission.model.units[_object_name] and mission.model.units[_object_name]:isInterceptor() then 
									if mission.model.units[_object_name]:hasTargetForIntercepting() then
										local _target_for_intercepting = mission.model.units[_object_name]:getTargetForIntercepting();
										local _target_for_intercepting_id, _object_id = _target_for_intercepting:getID(), _object:getID();
										
										if _target_for_intercepting_id ~= _object_id then 
											local _fighter_heading = mission.utils.getHeading(_data.fighter:getPosition());
											local _distance_to_target, _heading_to_target = mission.utils.getGuidanceAt(_data.fighter, _object);
											if math.abs(_heading_to_target - _fighter_heading) < 40 and _distance_to_target < 15000 then 
												_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
												mission.model.units[_object_name]:endInterceptTask();
											end 
										end
									else
										_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
									end
								else 
									_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
								end
							end
						end 
						return true;
					end, { fighter = _dcsw_unit });
			end
		end
		if 'F/A-18C' == _unit_desc.typeName or 'Su-25T' == _unit_desc.typeName 
			or 'Su-34' == _unit_desc.typeName or 'Su-24' == _unit_desc.typeName:sub(1, 5) 
			or 'Su-30' == _unit_desc.typeName or 'Tornado' == _unit_desc.typeName:sub(1, 7) then
			-- debug:
			-- mission.controller.sendMessageForAll(_dcsw_unit:getName() .. 'has attribute: SEAD');
			--
			_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.THREAT, msfAirDefenceUnit.ThreatFlags.SEAD);
			_priority_for_intercepting = msfAirDefenceUnit.PriorityForIntercepting.HIGH;
			_does_suit_for_intercepting = true;
		end
		if _unit_desc.attributes and _unit_desc.attributes['AWACS'] then 
			_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.THREAT, msfAirDefenceUnit.ThreatFlags.AWACS); 
		end
		if _unit_desc.attributes and _unit_desc.attributes['Bombers'] then 
			_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.THREAT, msfAirDefenceUnit.ThreatFlags.BOMBER);
			_does_suit_for_intercepting = true;
			_priority_for_intercepting = msfAirDefenceUnit.PriorityForIntercepting.HIGH;
			if _unit_desc.attributes['Strategic bombers'] then 
				_priority_for_intercepting = msfAirDefenceUnit.PriorityForIntercepting.HIGHEST;
				_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.STRATEGIC);
			end
		end
		if _unit_desc.attributes and _unit_desc.attributes['Battleplanes'] then
			local _object_types = {};
			
			table.insert(_object_types, Object.Category.UNIT);
			table.insert(_object_types, Object.Category.STATIC);
			
			_priority_for_intercepting = msfAirDefenceUnit.PriorityForIntercepting.MEDIUM;
			_does_suit_for_intercepting = true;
			_threat_type, _exist_target = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.THREAT, msfAirDefenceUnit.ThreatFlags.CAS), false;
			
			world.searchObjects(_object_types, {
					id = world.VolumeType.SPHERE, params = { point = _dcsw_unit:getPoint(), radius = 3000 } },
				function(_object, _data)
					if _object:getCoalition() ~= _data.battleplane:getCoalition() then 
						_exist_target = true;
						_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.CLOSE_TARGET);
						_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.EXIST_TARGET); 
						_priority_for_intercepting = msfAirDefenceUnit.PriorityForIntercepting.HIGHEST;
						
						if Object.Category.UNIT == _object:getCategory() then 
							local _target_group, _object_name = _object:getGroup(), _object:getName();
							local _target_group_name = _target_group:getName();
							if mission.model.groups[_target_group_name] and not _close_groups[_target_group_name] then 
								if mission.model.units[_object_name] and mission.model.units[_object_name]:isInterceptor() then 
									if mission.model.units[_object_name]:hasTargetForIntercepting() then
										local _target_for_intercepting = mission.model.units[_object_name]:getTargetForIntercepting();
										local _target_for_intercepting_id, _object_id = _target_for_intercepting:getID(), _object:getID();
										
										if _target_for_intercepting_id ~= _object_id then 
											local _fighter_heading = mission.utils.getHeading(_data.battleplane:getPosition());
											local _distance_to_target, _heading_to_target = mission.utils.getGuidanceAt(_data.battleplane, _object);
											if math.abs(_heading_to_target - _fighter_heading) < 40 then 
												_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
												mission.model.units[_object_name]:endInterceptTask();
											end 
										end
									else
										_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
									end
								else 
									_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
								end
							end
						end
					end 
					return true;
				end, { battleplane = _dcsw_unit });
			if not _exist_target then
				world.searchObjects(_object_types, {
						id = world.VolumeType.SPHERE, params = { point = _dcsw_unit:getPoint(), radius = 16000 } },
					function(_object, _data)
						if _object:getCoalition() ~= _data.battleplane:getCoalition() then 
							_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.EXIST_TARGET); 
							_priority_for_intercepting = msfAirDefenceUnit.PriorityForIntercepting.HIGH;
							
							if Object.Category.UNIT == _object:getCategory() then 
								local _target_group, _object_name = _object:getGroup(), _object:getName();
								local _target_group_name = _target_group:getName();
								if mission.model.groups[_target_group_name] and not _close_groups[_target_group_name] then 
									if mission.model.units[_object_name] and mission.model.units[_object_name]:isInterceptor() then 
										if mission.model.units[_object_name]:hasTargetForIntercepting() then
											local _target_for_intercepting = mission.model.units[_object_name]:getTargetForIntercepting();
											local _target_for_intercepting_id, _object_id = _target_for_intercepting:getID(), _object:getID();
											
											if _target_for_intercepting_id ~= _object_id then 
												local _fighter_heading = mission.utils.getHeading(_data.battleplane:getPosition());
												local _distance_to_target, _heading_to_target = mission.utils.getGuidanceAt(_data.battleplane, _object);
												if math.abs(_heading_to_target - _fighter_heading) < 40 and _distance_to_target < 5000 then 
													_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
													mission.model.units[_object_name]:endInterceptTask();
												end 
											end
										else
											_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
										end
									else 
										_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
									end
								end
							end
						end 
						return true;
					end, { battleplane = _dcsw_unit });
			end
		end
		if _unit_desc.attributes and _unit_desc.attributes['Helicopters'] then
			local _object_types = {};
			
			table.insert(_object_types, Object.Category.UNIT);
			table.insert(_object_types, Object.Category.STATIC);
			
			_does_suit_for_intercepting = true;
			_priority_for_intercepting = msfAirDefenceUnit.PriorityForIntercepting.LOW;
			_threat_type, _exist_target = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.THREAT, msfAirDefenceUnit.ThreatFlags.HELI), false;
			
			if _unit_desc.attributes['Attack helicopters'] then 
				_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.CAS);
				world.searchObjects(_object_types, {
						id = world.VolumeType.SPHERE, params = { point = _dcsw_unit:getPoint(), radius = 3000 } },
					function(_object, _data)
						if _object:getCoalition() ~= _data.heli:getCoalition() then 
							_exist_target = true;
							_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.CLOSE_TARGET);
							_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.EXIST_TARGET); 
							_priority_for_intercepting = msfAirDefenceUnit.PriorityForIntercepting.HIGH;
							
							if Object.Category.UNIT == _object:getCategory() then 
								local _target_group, _object_name = _object:getGroup(), _object:getName();
								local _target_group_name = _target_group:getName();
								if mission.model.groups[_target_group_name] and not _close_groups[_target_group_name] then 
									if not (mission.model.units[_object_name] and mission.model.units[_object_name]:isInterceptor()) then 
										_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
									end
								end
							end
						end
						return true;
					end, { heli = _dcsw_unit });
				if not _exist_target then
					world.searchObjects(_object_types, {
							id = world.VolumeType.SPHERE, params = { point = _dcsw_unit:getPoint(), radius = 11000 } },
						function(_object, _data)
							if _object:getCoalition() ~= _data.heli:getCoalition() then 
								_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.EXIST_TARGET);
								_priority_for_intercepting = msfAirDefenceUnit.PriorityForIntercepting.MEDIUM;
								if Object.Category.UNIT == _object:getCategory() then 
									local _target_group, _object_name = _object:getGroup(), _object:getName();
									local _target_group_name = _target_group:getName();
									if mission.model.groups[_target_group_name] and not _close_groups[_target_group_name] then 
										if not (mission.model.units[_object_name] and mission.model.units[_object_name]:isInterceptor()) then 
											_close_groups[_target_group_name] = mission.model.groups[_target_group_name];
										end
									end
								end
							end
							return true;
						end, { heli = _dcsw_unit });
				end
			end
			if _unit_desc.attributes and _unit_desc.attributes['Transport helicopters'] then 
				_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.TRANSPORT);
				_does_suit_for_intercepting = true;
				_priority_for_intercepting = msfAirDefenceUnit.PriorityForIntercepting.LOW;
			end
		end
		if _unit_desc.attributes and (_unit_desc.attributes['Transports'] or _unit_desc.attributes['Tankers']) then 
			_threat_type = _msf_or(_threat_type, msfAirDefenceUnit.ThreatFlags.THREAT, msfAirDefenceUnit.ThreatFlags.TRANSPORT);
		end
		return _threat_type, _does_suit_for_intercepting, _priority_for_intercepting, _close_groups; 
	end
};

function msfAirDefenceUnit:construct(parameters) 
	local _dcsw_unit = assert(parameters.airDefenceUnit:getDCSWorldUnit(), 
		'MSF Error: parameters.airDefenceUnit:getDCSWorldUnit() returned NIL on attempt to add air defence unit: ' .. tostring(parameters.displayName));
	
	self.coalition = _dcsw_unit:getCoalition();
	self.unit = parameters.airDefenceUnit;
	self.role = parameters.airDefenceRole;
	self.name = parameters.displayName;
	self.communicationUnit = parameters.communicationUnit;
	self.ambushZoneRadius = parameters.ambushZoneRadius;
	self.engageZoneRadius = parameters.engageZoneRadius;
	self.onLostConnection = Factory.create(msfEvent, {});
	-- self.onCeaseEngaging = Factory.create(msfEvent, {});
	-- self.isEngaging = false;
	self.activated = true;
	self.targetsOutOfRangeTime = 0;  -- счетчик времени отсутствия целей в зоне поражения
	self.targetsOutOfRangeMaxTime = parameters.maxActiveTimeIfNoTargets; -- максимальное время активности при отсутствии целей в зоне поражения
	self.wholeGroup = parameters.wholeGroup;
	self.lastActivationTime = 0;
end;

-- проверка находится ли цель в зоне ответствености ЗРК
function msfAirDefenceUnit:isTargetInRange(_target)
	local _dcsw_unit = self.unit:getDCSWorldUnit();
	
	if not (_dcsw_unit and _dcsw_unit:isExist()) then return false; end
	if not (_target.object and _target.object:isExist()) then return false; end
	
	local _distance = mission.utils.getDistance(_dcsw_unit:getPoint(), _target.object:getPoint());
	
	if _distance < self.engageZoneRadius then return true; end
	return false;
end;

function msfAirDefenceUnit:getCoalition() return self.coalition; end

function msfAirDefenceUnit:tryToPassTarget(_target, _threat_type) 
	if self.activated then return false; end 
	
	local _dcsw_unit = self.unit:getDCSWorldUnit();
	local _distance = mission.utils.getDistance(_dcsw_unit:getPoint(), _target.object:getPoint());

	-- debug: 
	-- mission.controller.sendMessageForAll('Try pass target for ' .. self.name .. '( current distance: ' 
	--	.. string.format("%.3f", _distance) .. ', engage radius: ' .. self.engageZoneRadius 
	--	.. ', ambush radius: ' .. self.ambushZoneRadius .. ' )');
	--
	
	if msfAirDefenceUnit.ThreatFlags.NO_THREAT == _threat_type then return false; end
	if self.role == msfAirDefenceUnitRole.LN then
		if 0 == _msf_and(_threat_type, msfAirDefenceUnit.ThreatFlags.EXIST_TARGET, msfAirDefenceUnit.ThreatFlags.BOMBER, msfAirDefenceUnit.ThreatFlags.AWACS)
			and 0 ~= _msf_xor(_threat_type, msfAirDefenceUnit.ThreatFlags.THREAT, msfAirDefenceUnit.ThreatFlags.HELI, msfAirDefenceUnit.ThreatFlags.TRANSPORT)
			then if _distance < self.ambushZoneRadius then self:setUnitActivate(true); return true; 
			else return false; end
		else
			if _distance < self.engageZoneRadius then self:setUnitActivate(true); return true; 
			else return false; end
		end
	end
	if self.role == msfAirDefenceUnitRole.AMBUSH then 
		if 0 == _msf_and(_threat_type, msfAirDefenceUnit.ThreatFlags.CLOSE_TARGET, msfAirDefenceUnit.ThreatFlags.AWACS) then 
			if _distance < self.ambushZoneRadius then self:setUnitActivate(true); return true;
			else return false; end
		else
			if _distance < self.engageZoneRadius then self:setUnitActivate(true); return true; 
			else return false; end
		end
	end
	return false;
end;

function msfAirDefenceUnit:tryToPassTargetForPlayer(_target, _threat_type)
	if self.activated then return false; end 
	
	local _dcsw_unit = self.unit:getDCSWorldUnit();
	local _distance = mission.utils.getDistance(_dcsw_unit:getPoint(), _target.object:getPoint());
	
	if msfAirDefenceUnit.ThreatFlags.NO_THREAT == _threat_type then return false; end
	
	if 0 == _msf_and(_threat_type, 
		msfAirDefenceUnit.ThreatFlags.EXIST_TARGET, msfAirDefenceUnit.ThreatFlags.BOMBER, msfAirDefenceUnit.ThreatFlags.AWACS) and 
		0 ~= _msf_xor(_threat_type, 
			msfAirDefenceUnit.ThreatFlags.THREAT, msfAirDefenceUnit.ThreatFlags.HELI, msfAirDefenceUnit.ThreatFlags.TRANSPORT) then 
		
		if _distance < self.ambushZoneRadius then self.activated = true; return true; else return false; end
	else 
		if _distance < self.engageZoneRadius then self.activated = true; return true; else return false; end
	end
	return false;
end;

function msfAirDefenceUnit:isActivated() return self.activated; end

-- возврат false говрит о неудачно попытке update-а
-- если update верул false, то юнит либо уничтожен, либо потерял связь - дальнейшее использование
-- в состве группировки не возможно
-- второе, возвращаемое значение = true соответствует прекращению юнитом атаки 
function msfAirDefenceUnit:update()
	if not self:isOnline() then return false, false; end 
	if not self.activated then return true, false; end -- чтобы вернуть правильные условия для ceaseEngaging(), если убрать 
													   -- эта функция вернет true, true
	
	if self:setUnitActivate(false) then return true, true; end -- деактивация не выполнится, если есть цели (проверка внутри)
	return true, false;
end;

-- возвращаемый результат аналогичен методу update
function msfAirDefenceUnit:updateUserControlledUnit()
	if not self:isOnline() then return false, false; end
	if not self.activated then return true, false; end
	
	local _dcsw_unit = self.unit:getDCSWorldUnit();
	if not _dcsw_unit then return false, false; end
	
	local _military_service_area, _enemy_units_data = {
		id = world.VolumeType.SPHERE,
		params = { point = _dcsw_unit:getPoint(), radius = self.engageZoneRadius; }
	}, 
	{
		coalition = _dcsw_unit:getCoalition(), 
		targetsExist = false;
	};
	
	world.searchObjects(Object.Category.UNIT, _military_service_area, function(_object, _data)
			local _object_coalition = _object:getCoalition();  
			if _object_coalition ~= _data.coalition and _object:inAir() then 
				_data.targetsExist = true;
				return false; 
			end
			return true;
		end, _enemy_units_data);
	
	-- debug:
	--mission.controller.sendMessageForAll(self.name .. ' (user unit) updating: ' 
	--	.. 'count targets: ' .. tostring(_enemy_units_data.targetsExist) .. ', targets out of range time: ' .. tostring(self.targetsOutOfRangeTime));
	-- end
	
	if _enemy_units_data.targetsExist then return true, false;
	else
		self.targetsOutOfRangeTime = self.targetsOutOfRangeTime + 2;
		if self.targetsOutOfRangeTime < self.targetsOutOfRangeMaxTime then return true, false; 
		else self.targetsOutOfRangeTime = 0; self.activated = false; return true, true; end
	end
	return true, false;
end;

function msfAirDefenceUnit:isTargetInSensorRange(_target)
	local _dcsw_unit = self.unit:getDCSWorldUnit();
	if not (_dcsw_unit and _dcsw_unit:isExist()) then return false; end
	if not (_target.object and _target.object:isExist()) then return false; end
	
	if not _dcsw_unit:hasSensors(Unit.SensorType.RADAR) then return false; end
	
	local _sensors = _dcsw_unit:getSensors();
	if not _sensors[Unit.SensorType.RADAR] then return false; end
	
	local _unit_point, _target_point = _dcsw_unit:getPoint(), _target.object:getPoint();
	if not land.isVisible({ x = _unit_point.x, y = _unit_point.y + 5, z = _unit_point.z }, _target_point) then return false; end
	
	local _distance = mission.utils.getDistance(_unit_point, _target_point);
	
	for _sensor_index, _sensor in ipairs(_sensors[Unit.SensorType.RADAR]) do
		local _detection_range = mission.utils.getDetectionRangeForASRadar(_target.object, _sensor, _dcsw_unit); 
		
		-- debug: 
		-- mission.controller.sendMessageForAll('MSF Debug: detection range between ' .. _target.object:getName() 
		--	.. ' and ' .. _dcsw_unit:getName() .. ' is ' .. _detection_range .. ', real distance: ' .. _distance);
		-- 
		
		if _distance < _detection_range then return true; end
	end
	return false;
end;

function msfAirDefenceUnit:getDisplayName() return self.name; end;

--[[ function msfAirDefenceUnit:addOnCeaseEngagingEventHandler(eventHandler)
	self.onCeaseEngaging[eventHandler] = eventHandler;
end;

function msfAirDefenceUnit:removeOnCeaseEngagingEventHandler(eventHandler)
	self.onCeaseEngaging[eventHandler] = nil;
end; --]]

-- function msfAirDefenceUnit:isUnitEngaging() return self.isEngaging; end;

-- function msfAirDefenceUnit:engaging()
-- 	self.isEngaging = true;
-- end;

--[[ function msfAirDefenceUnit:ceaseEngaging()
	self.isEngaging = false;
	for _onCeaseEngagingEventHandlerKey, _onCeaseEngagingEventHandler in pairs(self.onCeaseEngaging) do
		_onCeaseEngagingEventHandler(self, {});
	end
end; --]]

function msfAirDefenceUnit:addOnLostConnectionEventHandler(eventHandler)
	table.insert(self.onLostConnection, eventHandler);
end;

function msfAirDefenceUnit:lostConnection()
	self:setUnitActivate(true);
	for _, _event_handler in ipairs(self.onLostConnection) do 
		_event_handler(self, {});
	end
	self.role = msfAirDefenceUnitRole.STAND_ALONE; -- изменение роли происходит только, после отработки eventHandler-ов
end;

function msfAirDefenceUnit:isOnline()
	local _dcsw_unit = self.unit:getDCSWorldUnit();
	local _dcsw_communication = self.communicationUnit:getDCSWorldUnit();
	
	if not (_dcsw_unit and _dcsw_unit:isExist()) then self.role = msfAirDefenceUnitRole.WRECKED_UNIT; return false; end
	if not self.unit:isAlive() then self.role = msfAirDefenceUnitRole.WRECKED_UNIT; return false; end
	if not (_dcsw_communication and _dcsw_communication:isExist()) then self:lostConnection(); return false; end
	if not self.communicationUnit:isAlive() then self:lostConnection(); return false; end
	if mission.utils.getLogic(self.communicationUnit:getDammage()) then  self:lostConnection(); return false; end
	return true;
end;

function msfAirDefenceUnit:getUnit() return self.unit; end;
function msfAirDefenceUnit:getCommunicationUnit() return self.communicationUnit; end;
function msfAirDefenceUnit:setCommunicationUnit(_new_communication_unit) 
	self.communicationUnit = _new_communication_unit;
end;
function msfAirDefenceUnit:getRole() return self.role; end;
function msfAirDefenceUnit:setRole(_air_defence_unit_role)
	self.role = _air_defence_unit_role;
end;

function msfAirDefenceUnit:setUnitActivate(_value)
	-- debug:
	--error('Try to set actiovation for ' .. self.name .. ' to value: ' .. tostring(_value));
	--[[local _debug_report = '';
	if self.name == '3-я батарея ЗРК Бук 2-ого дивизиона 643-ого зрп' and _value ~= self.activated then
		_debug_report = _debug_report .. 'Try to set actiovation for ' .. self.name .. ' to value: ' .. tostring(_value);
		mission.controller.sendMessageForAll('Try to set actiovation for ' .. self.name .. ' to value: ' .. tostring(_value));
		if not _value and not self.activated then error('Try to set actiovation for ' .. self.name .. ' to value: ' .. tostring(_value)
			.. ', instead it has self.activated value ' .. tostring(self.activated)); end
	end--]]
	--
	if not _value and not self.activated then return true; end
	
	local _dcsw_unit = self.unit:getDCSWorldUnit();
	if not _dcsw_unit then 
		-- debug:
		--[[ if self.name == '3-я батарея ЗРК Бук 2-ого дивизиона 643-ого зрп' then
			_debug_report = _debug_report .. '\n\r Could not get dcsw unit of ' .. self.name;
			mission.controller.sendMessageForAll(_debug_report);
		end --]]
		-- 
		return false; 
	end
	
	-- не выключать AI, если включили не более чем 10 сек назад
	if not _value and 0 ~= self.lastActivationTime and (timer.getTime() - self.lastActivationTime) < 10 then 
		-- debug:
		--[[ if self.name == '3-я батарея ЗРК Бук 2-ого дивизиона 643-ого зрп' then
			_debug_report = _debug_report .. '\n\r too little time after activation: ' .. tostring(_value) .. ', last activation time: '
				.. tostring(self.lastActivationTime) .. ', current time: ' .. timer.getTime()
				.. ', time diff: ' .. (timer.getTime() - self.lastActivationTime);
			 mission.controller.sendMessageForAll(_debug_report);
		end --]]
		--
		return false; 
	end
	
	-- не выключать AI для LN и AMBUSH, если они заняты целью
	if not _value and (msfAirDefenceUnitRole.LN == self.role or msfAirDefenceUnitRole.AMBUSH == self.role) then 
		local _targets = self:getTargets();
		local _target_exists = false;
		local _target_index = 0;
		if #_targets > 0 then while(function()
				if _target_index == 0 then return true; end
				if _target_index > #_targets then return false; end
				local _detected, _visible, _type, _distance, _position, _velocity = self:isTargetDetected(_targets[_target_index].object);
				if not (_targets[_target_index].object and _targets[_target_index].object:isExist()) then return true; end 
				local _category = _targets[_target_index].object:getCategory();
				if Object.Category.UNIT ~= _category then return true; end
				if _visible and _distance < self.engageZoneRadius then self.targetsOutOfRangeTime = 0; _target_exists = true; return false; end
			end)() do
				_target_index = _target_index + 1;
			end
			if _target_exists then 
				-- debug:
				--[[ if self.name == '3-я батарея ЗРК Бук 2-ого дивизиона 643-ого зрп' and _value ~= self.activated then
					_debug_report = _debug_report .. '\n\r target exists: ' .. _targets[1].object:getName();
					mission.controller.sendMessageForAll(_debug_report);
				end--]]
				--
				return false;
			else self.targetsOutOfRangeTime = self.targetsOutOfRangeTime + 2; end -- 2 - frame time
			if self.targetsOutOfRangeTime < self.targetsOutOfRangeMaxTime then 
				-- debug:
				--[[ if self.name == '3-я батарея ЗРК Бук 2-ого дивизиона 643-ого зрп' and _value ~= self.activated then
					_debug_report = _debug_report .. '\n\r out of range time: ' .. self.targetsOutOfRangeTime
						.. ' less then: ' .. self.targetsOutOfRangeMaxTime;
					mission.controller.sendMessageForAll(_debug_report);
				end --]]
				--
				return false; 
			end
		-- else return false; - что это за х-ня??? 
		end
	end
	
	local _controller = nil; 
	if self.wholeGroup then _controller = _dcsw_unit:getController(); 
	else local _dcsw_group = _dcsw_unit:getGroup(); _controller = _dcsw_group:getController(); end
	
	if _value and not self.activated then self.lastActivationTime = timer.getTime(); end
	if not _value and self.activated then self.lastActivationTime = 0; end
	
	-- debug:
	--error('Try to set actiovation for ' .. self.name .. ' to value: ' .. tostring(_value));
	--[[ if self.name == '3-я батарея ЗРК Бук 2-ого дивизиона 643-ого зрп' then
		mission.controller.sendMessageForAll('SUCCESS to set actiovation for ' .. self.name .. ' to value: ' .. tostring(_value));
		--if not _value then error('Try to set actiovation for ' .. self.name .. ' to value: ' .. tostring(_value)); end
	end --]]
	--
	
	_controller:setOnOff(_value);
	self.activated = _value;
	
	return true;
end;

function msfAirDefenceUnit:getTargets()
	local _dcsw_unit = self.unit:getDCSWorldUnit();
	if not (_dcsw_unit and _dcsw_unit:isExist() and self.unit:isAlive()) then return {}; end
	local _controller = _dcsw_unit:getController();
	
	return _controller:getDetectedTargets(nil);
end;

function msfAirDefenceUnit:isTargetDetected(_target_object) -- detected, visible, type, distance, position, velocity
	local _dcsw_unit = self.unit:getDCSWorldUnit();
	if not (_dcsw_unit and _dcsw_unit:isExist()) then return false, false, false, 0, { x = 0, y = 0, z = 0}, { x = 0, y = 0, z = 0}; end
	if not (_target_object and _target_object:isExist() and _target_object:inAir()) then return false, false, false, 0, { x = 0, y = 0, z = 0 }, { x = 0, y = 0, z = 0}; end
	local _controller = _dcsw_unit:getController();
	local _detected, _visible, _, _type, _distance, _last_position, _last_velocity = _controller:isTargetDetected(_target_object, nil);
	
	if not _detected then return false, false, false, 0, { x = 0, y = 0, z = 0}, { x = 0, y = 0, z = 0}; end
	if not _visible then return true, false, _type, _distance, _last_position, _last_velocity; end
	
	_last_position, _last_velocity = _target_object:getPoint(), _target_object:getVelocity();
	_distance = mission.utils.getDistance(_dcsw_unit:getPoint(), _last_position); 
	
	return true, true, _type, _distance, _last_position, _last_velocity;
end

msfAirDefenceGroupment = {
};

function msfAirDefenceGroupment:construct(parameters)
	self.units = {};
	self.targets = {};  -- id, object, status, isTypeKnown, typeName, isVisisble, distance, point, velocity
	self.subGroupments = {};
	self.name = parameters.name;
	self.displayName = parameters.displayName;
	self.coalition = nil;
	self.parentGroupment = parameters.parent;
	self.onLostConnection = Factory.create(msfEvent, {});
	-- self.onLostUnit = Factory.create(msfEvent, {});
	self.onRequestClosestBandit = Factory.create(msfEvent, {});
	self.onEngage = Factory.create(msfEvent, {});
	self.onCeaseEngaging = Factory.create(msfEvent, {});
	self.onActivate = Factory.create(msfEvent, {});
	self.onDeactivate = Factory.create(msfEvent, {});
	self.onTargetDetected = Factory.create(msfEvent, {});
	self.activated = parameters.isActivated;
	self.deactivateGroupmentIfNoTargets = parameters.deactivateIfNoTargets;
	self.destroyed = false;
	self.isLostConnection = false;
end;

function msfAirDefenceGroupment:isDestroyed() return self.destroyed; end;
function msfAirDefenceGroupment:getCoalition() return self.coalition; end;

function msfAirDefenceGroupment:requestClosestBandit(_requester)
	local _closest_target, _ = self:getClosestBanditFor(_requester);
	
	for _, _event_handler in pairs(self.onRequestClosestBandit) do
		_event_handler(self, { requester = _requester, bandit = _closest_target });
	end
end;

function msfAirDefenceGroupment:getClosestBanditFor(_requester)
	if self.destroyed and self.isLostConnection then return nil, 0; end
	if not self.activated then return nil, 0; end
	if not (_requester and _requester:isExist()) then return nil, 0; end
	
	local _requester_point, _closet_target_index, _min_distance, _closest_target = _requester:getPoint(), 0, 0, nil;
	
	for _target_index, _target in ipairs(self.targets) do
		if _target.object and _target.object:isExist() then 
			local _target_point = _target.object:getPoint();
			local _distance = mission.utils.getDistance(_requester_point, _target_point);
			if _min_distance > 0 then if _min_distance > _distance then _min_distance, _closet_target_index = _distance, _target_index; end
			else _min_distance, _closet_target_index = _distance, _target_index; end
		end
	end
	if _closet_target_index > 0 then _closest_target = self.targets[_closet_target_index].object; 
	end
	for _, _subgroupment in pairs(self.subGroupments) do
		local _subgroupment_closest_target, _subgroupment_min_distance = _subgroupment:getClosestBanditFor(_requester);
		if 0 ~= _subgroupment_min_distance and  nil ~= _subgroupment_closest_target and _subgroupment_min_distance < _min_distance then 
			_closest_target, _min_distance = _subgroupment_closest_target, _subgroupment_min_distance;
		end
	end
	return _closest_target, _min_distance;
end;

function msfAirDefenceGroupment:update()
	local _ewr_units, _command_posts, _ln_units, _communications, _ambush_units, _observer_and_ewr_units = {}, {}, {}, {}, {}, {};
	local _user_controlled_units = {};
	local _is_hq_alive, _is_groupment_online, _no_targets, _is_ewr_alive  = false, false, false, false; 
	
	-- debug:
	-- if 'georgia_air_defence_brigade' == self.name then 
	--	mission.controller.sendMessageForAll('MSF Debug: updating ' .. self.displayName .. ', groupment destroyed: ' .. tostring(self.destroyed) ..
	--		', lost connection: ' .. tostring(self.isLostConnection) .. ', activated: ' .. tostring(self.activated));
	-- end
	--
	
	if not self.destroyed then
		-- если группировка не считается уничтоженной:
		-- получение массивов всех типов юнитов группировки:
		for _air_defence_unit_key, _air_defence_unit in pairs(self.units) do
			local _air_defence_unit_role = _air_defence_unit:getRole();
			if msfAirDefenceUnitRole.EWR == _air_defence_unit_role then table.insert(_ewr_units, _air_defence_unit); end
			-- обзор ведут посты наблюдения и радары, если они активированны:
			if msfAirDefenceUnitRole.OBSERVATION_POST == _air_defence_unit_role or (msfAirDefenceUnitRole.EWR == _air_defence_unit_role and self.activated) then 
				table.insert(_observer_and_ewr_units, _air_defence_unit); 
			end
			if msfAirDefenceUnitRole.COMMUNICATION_POST == _air_defence_unit_role then table.insert(_communications, _air_defence_unit); end
			if msfAirDefenceUnitRole.COMMAND_POST == _air_defence_unit_role then table.insert(_command_posts, _air_defence_unit); end
			if msfAirDefenceUnitRole.AMBUSH == _air_defence_unit_role then 
				table.insert(_ambush_units, _air_defence_unit); 
				local _, _cease_fire = _air_defence_unit:update();
				if _cease_fire then self:ceaseEngaging(_air_defence_unit); end
			end
			if msfAirDefenceUnitRole.LN == _air_defence_unit_role then 
				table.insert(_ln_units, _air_defence_unit); 
				local _, _cease_fire = _air_defence_unit:update(); 
				if _cease_fire then self:ceaseEngaging(_air_defence_unit); end
			end
			if msfAirDefenceUnitRole.USER_CONTROLLED_UNIT == _air_defence_unit_role then 
				table.insert(_user_controlled_units, _air_defence_unit);
				local _, _cease_fire = _air_defence_unit:updateUserControlledUnit();
				if _cease_fire then self:ceaseEngaging(_air_defence_unit); end 
			end
		end
		-- debug:
		-- if 'georgia_air_defence_brigade' == self.name then 
		--	mission.controller.sendMessageForAll('MSF Debug: updating ' .. self.displayName .. ', count EWR: ' .. #_ewr_units 
		--		.. ', observations and EWR: ' .. #_observer_and_ewr_units 
		--		.. ', communication post: ' .. #_communications
		--		.. ', command posts: ' .. #_command_posts .. ', ambushes: ' .. #_ambush_units .. ', LN: ' .. #_ln_units);
		-- end
		--
		-- проверка коммандных постов:
		local _hq_index = 0;
		while(function()
			if _hq_index == 0 then return true; end
			if _hq_index > #_command_posts then return false; end
			-- завершить проверку по первому  найденному командному посту на связи: установить признак _is_hq_alive
			if _command_posts[_hq_index]:isOnline() then _is_hq_alive = true; return false; end;	
			end)() do
			_hq_index = _hq_index + 1;
		end
		-- если коммандование в состоянии управлять группировкой (_is_hq_alive)
		if _is_hq_alive then 
			-- проверка состояния обзорных РЛС:
			for _ewr_index, _ewr in ipairs(_ewr_units) do if _ewr:isOnline() then _is_ewr_alive = true; break; end
			end
		end
		-- если уничтожено командование или уничтожены все обзорные РЛС
		if not _is_hq_alive or not _is_ewr_alive then 
			self.destroyed = true;
			-- командный центр уничтожен или потерял связь с группировкой
			-- все переходят на автономную работу: 
			for _ln_index, _ln in ipairs(_ln_units) do _ln:setUnitActivate(true); _ln:setRole(msfAirDefenceUnitRole.STAND_ALONE);
			end
			for _ambush_index, _ambush in ipairs(_ambush_units) do _ln:setUnitActivate(true); _ambush:setRole(msfAirDefenceUnitRole.STAND_ALONE);
			end
			for _ewr_index, _ewr in ipairs(_ewr_units) do _ewr:setUnitActivate(true); _ewr:setRole(msfAirDefenceUnitRole.STAND_ALONE);
			end
			for _subgroupment_key, _subgroupment in pairs(self.subGroupments) do _subgroupment:activate(); _subgroupment:setAutoDeactivation(false); end
		end
		-- проверка связи с другими группировками:
		if not self.isLostConnection then 
			local _communication_index = 0;
			while(function()
					if _communication_index == 0 then return true; end
					if _communication_index > #_communications then return false; end
					if _communications[_communication_index]:isOnline() then _is_groupment_online = true; return false; end;	
				end)() do
				_communication_index = _communication_index + 1;
			end
			if not _is_groupment_online then 
				self.isLostConnection = true; 
				self:lostConnection();
				for _subgroupment_key, _subgroupment in pairs(self.subGroupments) do _subgroupment:activate(); _subgroupment:setAutoDeactivation(false); end
			end
		end
	end
	-- обновление группировок в подчинениии:
	for _subgroupment_key, _subgroupment in pairs(self.subGroupments) do _subgroupment:update(); end
	-- если текущая группровка уничтожена или не активированна завершить её обновление
	if self.destroyed or not self.activated then return; end
	
	-- если все нормально со свзью и командованием:
	-- ищем цели:
	for _ewr_index, _ewr in ipairs(_observer_and_ewr_units) do
		local _targets = _ewr:getTargets();
		local _target_index = 0;
		if #_targets > 0 then while(function()
				if _target_index == 0 then return true; end
				if _target_index > #_targets then return false; end
				
				-- debug:
				-- env.info('MSF Debug: №1' .. _ewr:getDisplayName() .. ' has visual contact with target: ' 
				--	.. tostring(_targets[_target_index].visible) .. '. Target name: ' .. _targets[_target_index].object:getName(), false);
				-- 
				
				if not _targets[_target_index].visible then return true; end -- NO_TARGET;
				local _detected, _visible, _type, _distance, _position, _velocity = _ewr:isTargetDetected(_targets[_target_index].object);
				
				-- debug:
				-- env.info('MSF Debug: №2' .. _ewr:getDisplayName() .. ' has visual contact with target: ' 
				--	.. tostring(_visible) .. '. Target name: ' .. _targets[_target_index].object:getName(), false);
				-- 
				
				if not _visible then return true; end -- NO_TARGET
				if not self:isTargetKnown(_targets[_target_index].object:getName()) then -- добавить новую цель
					local _type_name = 'unknown';
					if _type then _type_name = _targets[_target_index].object:getTypeName(); end
					local _new_target = {
						id = _targets[_target_index].object:getName(),
						object = _targets[_target_index].object,
						status = msfAirDefenceTargetStatus.NEW_TARGET,
						isTypeKnown = _type,
						typeName = _type_name,
						isVisible = _visible,
						distance = _distance, 
						point = _position,
						velocity = _velocity,
						assignedWith = nil -- air defence unit работающий по цели
					};
					table.insert(self.targets, _new_target);
					self:targetDetected(_ewr, _new_target); -- вызов события на обнаружение новой цели
				else -- отметить цель как актуальную
					self:setTargetStatus(_targets[_target_index].object:getName(), msfAirDefenceTargetStatus.TARGET);
				end
			end)() do
				_target_index = _target_index + 1;
			end
		end
	end
	
	-- деактивация группировки, если нет целей и указано поведение деактивироваться при отсутствии целей:
	-- debug:
	--if #self.targets > 0 then mission.controller.sendMessageForAll('MSF Debug: groupment: ' .. self.displayName ..  ' count targets: ' .. #self.targets); end
	-- 
	
	-- отмечание целей на удаление: 
	local _targets_to_delete = {};
	for _target_index, _target in ipairs(self.targets) do
		if _target.status == msfAirDefenceTargetStatus.NO_TARGET then 
			table.insert(_targets_to_delete, _target_index); 
		end
	end
	for _target_to_delete_index, _target_to_delete in ipairs(_targets_to_delete) do table.remove(self.targets, _target_to_delete);
	end
	if #self.targets < 1 and self.deactivateGroupmentIfNoTargets then self:deactivate(); end
	
	-- выбор средств для атаки:
	do
		local _target_index = 0;
		local _targets_for_intercepting, _all_close_groups = {}, {};
		
		while(function()
				if 0 == _target_index then return true; end
				if _target_index > #self.targets then return false; end
				
				local _target = self.targets[_target_index];
				local _units_limit = 0;
				local _threat_type, _does_suit_for_intercepting, _priority_for_intercepting, _close_groups 
					= msfAirDefenceUnit.getThreatType(_target.object);
				
				if msfAirDefenceUnit.ThreatFlags.NO_THREAT == _threat_type then return true;
				else
					if 0 ~= _msf_and(_threat_type, msfAirDefenceUnit.ThreatFlags.THREAT) then _units_limit = 1;
						if 0 ~= _msf_and(_threat_type, msfAirDefenceUnit.ThreatFlags.AWACS, msfAirDefenceUnit.ThreatFlags.STRATEGIC
							, msfAirDefenceUnit.ThreatFlags.CLOSE_TARGET) then _units_limit = -1; -- infinity units to attack 
						else
							if 0 ~= _msf_and(_threat_type, msfAirDefenceUnit.ThreatFlags.EXIST_TARGET) then _units_limit = 4;  
							else _units_limit = 2; end
							if 0 == _msf_xor(_threat_type, msfAirDefenceUnit.ThreatFlags.THREAT, msfAirDefenceUnit.ThreatFlags.HELI,
								msfAirDefenceUnit.ThreatFlags.TRANSPORT) then _units_limit = 4; 
							end
							if 0 ~= _msf_and(_threat_type, msfAirDefenceUnit.ThreatFlags.SEAD) then _units_limit = 1; end  
						end
					end
				end
				
				-- сохраение групп под угрозой атаки: 
				for _close_group_key, _close_group in pairs(_close_groups) do
					if not _all_close_groups[_close_group_key] then 
						_all_close_groups[_close_group_key] = { group = _close_group, threat = _target.object }; 
					end
				end
				
				-- сохранение целей для перехвата:
				if _does_suit_for_intercepting then 
					table.insert(_targets_for_intercepting, { target = _target, priority = _priority_for_intercepting });
				end
				
				-- собственные средства:
				-- для управляемых игроком ЗРК назначение цели происходит даже, если цель в работе у ЗРК под управлением AI
				local _count_units_for_attack = 0;
				for _, _user_controlled_unit in ipairs(_user_controlled_units) do
					if _count_units_for_attack < _units_limit or -1 == _units_limit then
						if _user_controlled_unit:tryToPassTargetForPlayer(_target, _threat_type) then 
							_count_units_for_attack = _count_units_for_attack + 1;
							self:engage(_user_controlled_unit, _target.object, _threat_type);
						end
					else return true; end
				end
				
				-- если цель в работе у юнита ЗРК под управлением AI прекратить дальнейший выбор средств её атаки
				if _target.assignedWith then if _target.assignedWith:isTargetInRange(_target) then return true;
					else _target.assignedWith = nil; end 
				end
				
				for _ambush_index, _ambush in ipairs(_ambush_units) do
					if _count_units_for_attack < _units_limit or -1 == _units_limit then 
						if _ambush:tryToPassTarget(_target, _threat_type) then 
							_count_units_for_attack = _count_units_for_attack + 1;
							_target.assignedWith = _ambush;
							self:engage(_ambush, _target.object, _threat_type);
						end
					else return true; end
				end
				for _ln_index, _ln in ipairs(_ln_units) do 
					if _count_units_for_attack < _units_limit or -1 == _units_limit then 
						if _ln:tryToPassTarget(_target, _threat_type) then 
							_count_units_for_attack = _count_units_for_attack + 1; 
							_target.assignedWith = _ln;
							self:engage(_ln, _target.object, _threat_type);
						end
					else return true; end
				end
				-- подчиненные группировки (если выполнение дошло до этой точки, то группировке не хватает собственных сил):
				for _subgroupment_key, _subgroupment in pairs(self.subGroupments) do 
					if _subgroupment:tryToPassTarget(_target) then return true; end 
				end
				end)() do
			_target_index = _target_index + 1;
		end
		
		table.sort(_targets_for_intercepting, function(_target1, _target2)
				if _target1.priority < _target2.priority then return true; end
				return false;
			end);
		
		-- раздача слонов населению:
		while(function()
				local _target_for_intercepting = table.remove(_targets_for_intercepting);
				if not _target_for_intercepting then return false; end
				if not (_target_for_intercepting.target.object and _target_for_intercepting.target.object:isExist()) then return true; end
				
				for _, _msf_unit in pairs(mission.model.units) do
					local _dcsw_interceptor_unit = _msf_unit:getDCSWorldUnit();
					if _dcsw_interceptor_unit and _msf_unit:isInterceptor() and not _msf_unit:hasTargetForIntercepting() then 
						local _distance_to_target = mission.utils.getDistance(_dcsw_interceptor_unit:getPoint(), _target_for_intercepting.target.object:getPoint());
						if _distance_to_target < 50000 then _msf_unit:startInterceptTask(self, _target_for_intercepting.target); end
					end
				end
			end)() do
		end
		
		-- предепреждение об угрозах:
		for _, _close_group in pairs(_all_close_groups) do _close_group.group:detectThreat(self, _close_group.threat); 
		end
	end -- end targets handling
	
	-- сброс статуса целей:
	for _target_index, _target in ipairs(self.targets) do self:setTargetStatus(_target.id, msfAirDefenceTargetStatus.NO_TARGET);
	end
end;

-- передача цели на споровождение, подчиненнной группировке:
function msfAirDefenceGroupment:tryToPassTarget(_target)
	local _observer_units, _ewr_units = {}, {};
	
	-- debug:
	-- mission.controller.sendMessageForAll('MSF Debug: attempt to pass target to: ' .. self.displayName
	--	.. ', target name: ' .. _target.object:getName());
	--
	
	if self.activated then return false; end
	if not self.destroyed and not self.isLostConnection then 
		for _air_defence_unit_key, _air_defence_unit in pairs(self.units) do
			local _air_defence_unit_role = _air_defence_unit:getRole();
			if msfAirDefenceUnitRole.EWR == _air_defence_unit_role then 
				table.insert(_ewr_units, _air_defence_unit); 
			end
			if msfAirDefenceUnitRole.OBSERVATION_POST == _air_defence_unit_role then 
				table.insert(_observer_units, _air_defence_unit);
			end
		end
	else return false; end
	for _ewr_index, _ewr in ipairs(_ewr_units) do
		if _ewr:isTargetInSensorRange(_target) then self:activate(); return true; end
	end
	for _observer_index, _observer in ipairs(_observer_units) do
		local _, _visible, _, _, _, _ = _observer:isTargetDetected(_target.object);
		if _visible then self:activate(); return true; end
	end
	return false;
end;

function msfAirDefenceGroupment:targetDetected(_air_defenece_unit, _target)
	for _onTargetDetectedEventHandlerKey, _onTargetDetectedEventHandler in pairs(self.onTargetDetected) do
		_onTargetDetectedEventHandler(self, { airDefenceUnit = _air_defenece_unit, target = _target });
	end
end;

function msfAirDefenceGroupment:addOnTargetDetectedEventHandler(eventHandler)
	self.onTargetDetected[eventHandler] = eventHandler;
end;

function msfAirDefenceGroupment:removeOnTargetDetectedEventHandler(eventHandler)
	self.onTargetDetected[eventHandler] = nil;
end;

function msfAirDefenceGroupment:setTargetStatus(_target_id, _status) -- 0 - no target, 1 - target, 2 - new target 
	for _target_index, _target in ipairs(self.targets) do
		if _target.id == _target_id then _target.status = _status;  end 
	end
end;

--[[ function msfAirDefenceGroupment:addOnLostUnitEventHandler(eventHandler)
	self.onLostUnit[eventHandler] = eventHandler;
end;

function msfAirDefenceGroupment:removeOnLostUnitEventHandler(eventHandler)
	self.onLostUnit[eventHandler] = nil;
end;

function msfAirDefenceGroupment:lostUnit(_lost_unit)
	for _onLostUnitEventHandlerKey, _onLostUnitEventHandler in pairs(self.onLostUnit) do 
		_onLostUnitEventHandler(self, { lostUnit = _lost_unit }); 
	end
end; --]]

function msfAirDefenceGroupment:setAutoDeactivation(_deactivate_if_no_tragets)
	self.deactivateGroupmentIfNoTargets = _deactivate_if_no_tragets;
end;

function msfAirDefenceGroupment:deactivateIfNoTargets() 
	return self.deactivateGroupmentIfNoTargets;
end;


-- активация группировки означает включение только обзорных радаров
-- включение РЛС пусковых установок идет только от командного центра группировки
function msfAirDefenceGroupment:activate()
	self.activated = true;
	for _air_defence_unit_key, _air_defence_unit in pairs(self.units) do
		if msfAirDefenceUnitRole.EWR == _air_defence_unit:getRole() then _air_defence_unit:setUnitActivate(true); end
	end
	for _onActivateEventHandlerKey, _onActivateEventHandler in pairs(self.onActivate) do
		_onActivateEventHandler(self, {});
	end
end;

function msfAirDefenceGroupment:addOnActivateEventHandler(eventHandler)
	self.onActivate[eventHandler] = eventHandler;
end;

function msfAirDefenceGroupment:removeOnActivateEventHandler(eventHandler)
	self.onActivate[eventHandler] = nil;
end;

function msfAirDefenceGroupment:deactivate() 
	local _can_not_deactivate = false; 
	
	self.targets = {};
	for _air_defence_unit_key, _air_defence_unit in pairs(self.units) do
		local _air_defence_unit_role = _air_defence_unit:getRole();
		if msfAirDefenceUnitRole.EWR == _air_defence_unit_role or msfAirDefenceUnitRole.AMBUSH == _air_defence_unit_role
			or msfAirDefenceUnitRole.LN == _air_defence_unit_role then if not _air_defence_unit:setUnitActivate(false) then _can_not_deactivate = true; end
		end
	end
	if not _can_not_deactivate then self.activated = false; 
	else return; end
	
	for _onDeactivateEventHandlerKey, _onDeactivateEventHandler in pairs(self.onDeactivate) do
		_onDeactivateEventHandler(self, {});
	end
end;

function msfAirDefenceGroupment:addOnDeactivateEventHandler(eventHandler)
	self.onDeactivate[eventHandler] = eventHandler;
end;

function msfAirDefenceGroupment:removeOnDeactivateEventHandler(eventHandler)
	self.onDeactivate[eventHandler] = nil;
end;

function msfAirDefenceGroupment:isActivated() return self.activated; end;

function msfAirDefenceGroupment:addOnEngageEventHandler(eventHandler)
	self.onEngage[eventHandler] = eventHandler;
end;

function msfAirDefenceGroupment:addOnCeaseEngagingEventHandler(eventHandler)
	self.onCeaseEngaging[eventHandler] = eventHandler;
end;

function msfAirDefenceGroupment:removeOnEngageEventHandler(eventHandler)
	self.onEngage[eventHandler] = nil;
end;

function msfAirDefenceGroupment:removeOnCeaseEngagingEventHandler(eventHandler)
	self.onCeaseEngaging[eventHandler] = nil;
end;

function msfAirDefenceGroupment:engage(_air_defence_unit, _target, _threat_type)	
	for _, _onEngageEventHandler in pairs(self.onEngage) do
		_onEngageEventHandler(self, { airDefenceUnit = _air_defence_unit, target = _target, threatType = _threat_type });
	end
end;

function msfAirDefenceGroupment:ceaseEngaging(_air_defence_unit)
	for _onCeaseEngagingEventHandlerKey, _onCeaseEngagingEventHandler in pairs(self.onCeaseEngaging) do
		_onCeaseEngagingEventHandler(self, { airDefenceUnit = _air_defence_unit });
	end
end;

function msfAirDefenceGroupment:addOnLostConnectionEventHandler(eventHandler)
	table.insert(self.onLostConnection, eventHandler);
end;

function msfAirDefenceGroupment:lostConnection()
	for _onLostConnectionEventHandlerIndex, _onLostConnectionEventHandler in ipairs(self.onLostConnection) do
		_onLostConnectionEventHandler(self, {});
	end
end;

function msfAirDefenceGroupment:addOnRequestClosestBanditEventHandler(eventHandler)
	self.onRequestClosestBandit[eventHandler] = eventHandler;
end;

function msfAirDefenceGroupment:removeOnRequestClosestBanditEventHandler(eventHandler)
	self.onRequestClosestBandit[eventHandler] = nil;
end;

function msfAirDefenceGroupment:getName() return self.name; end;
function msfAirDefenceGroupment:getDisplayName() return self.displayName; end;
function msfAirDefenceGroupment:getSubGroupments() return self.subGroupments; end;
function msfAirDefenceGroupment:getAirDefenceUnits() return self.units; end;
function msfAirDefenceGroupment:getParentGroupment() return self.parentGroupment; end;
function msfAirDefenceGroupment:setParentGroupment(_groupment)
	if self.parentGroupment then 
		self.parentGroupment.subGroupments[self.name] = nil;
	end
	_groupment:addSubGroupment(self.name, self.displayName);
end;

function msfAirDefenceGroupment:addSubGroupment(_name, _display_name, _is_activated, _auto_deactivation)
	self.subGroupments[_name] = Factory.create(msfAirDefenceGroupment, { 
		name = _name
		, displayName = _display_name
		, parent = self
		, isActivated = _is_activated 
		, deactivateIfNoTargets = _auto_deactivation });
end;

function msfAirDefenceGroupment:removeSubGroupment(_name)
	self.subGroupments[_name] = nil;
end;

function msfAirDefenceGroupment:addAirDefenceUnit(_unit, _air_defence_unit_role, 
	_ambush_zone_radius, _engage_zone_radius, _communication_unit, _display_name, _deactivate_delay, _as_whole_group)
	
	if nil == _as_whole_group then _as_whole_group = false; end 
	self.units[_unit:getUnitName()] = Factory.create(msfAirDefenceUnit, {
			airDefenceUnit = _unit,
			airDefenceRole = _air_defence_unit_role,
			displayName = _display_name,
			communicationUnit = _communication_unit,
			ambushZoneRadius = _ambush_zone_radius,
			engageZoneRadius = _engage_zone_radius,
			maxActiveTimeIfNoTargets = _deactivate_delay,
			wholeGroup = _as_whole_group
		});
	
	local _dcsw_unit = _unit:getDCSWorldUnit();
	if not _dcsw_unit then return; end
	
	if not self.coalition then self.coalition = _dcsw_unit:getCoalition();
	end
	
	if msfAirDefenceUnitRole.USER_CONTROLLED_UNIT == _air_defence_unit_role then 
		local _controller = _dcsw_unit:getController();
		_controller:setOption(AI.Option.Ground.id.ALARM_STATE, AI.Option.Ground.val.ALARM_STATE.GREEN);
		_controller:setOption(AI.Option.Ground.id.ROE, AI.Option.Ground.val.ROE.RETURN_FIRE);
	end
end;

function msfAirDefenceGroupment:removeAirDefenceUnit(_unit_name)
	self.units[_unit_name]:setRole(msfAirDefenceUnitRole.WRECKED_UNIT);
end;

-- проверка связи с ближайщим командным пунктом
function msfAirDefenceGroupment:isOnline()
	return (not self.isLostConnection and not self.destroyed);
end;

-- проверка: обнаружен ли юнит обзорными РЛС 
-- или средствами визуального наблюдения.
function msfAirDefenceGroupment:isTargetKnown(_found_target_id)
	for _target_index, _target in ipairs(self.targets) do		
		if _target.id == _found_target_id then return true; 
		end 
	end
	return false;
end;

-- trigger zone:

msfZone = {
};

function msfZone:construct(parameters)
	self.enteredUnits = {};
	self.dcsWorldTriggerZone = parameters.newTriggerZone;
	self.onUnitEnter = Factory.create(msfEvent, {});
	self.onUnitLeave = Factory.create(msfEvent, {});
end;

function msfZone:getUnitsInZone() return self.enteredUnits; end;
function msfZone:getDCSWorldTriggerZone() return self.dcsWorldTriggerZone; end;

function msfZone:refreshUnitsInZone()
	for _unit_index, _unit in pairs(self.enteredUnits) do
		if (false == _unit:isUnitInsideZone(self.dcsWorldTriggerZone)) then
			table.remove(self.enteredUnits, _unit_index);
		end
	end
end;

function msfZone:removeUnitFormZone(_removing_unit)
	local _removing_unit_name = _removing_unit:getUnitName();

	for _unit_index, _unit in pairs(self.enteredUnits) do
		if (_removing_unit_name == _unit:getUnitName()) then
			table.remove(self.enteredUnits, _unit_index);
			break;
		end
	end
end;

function msfZone:isUnitInsideZone(_checking_unit)	
	local _checking_unit_name = _checking_unit:getUnitName();

	for _unit_index, _unit in pairs(self.enteredUnits) do
		if (_checking_unit_name == _unit:getUnitName()) then 
			return true;
		end;
	end
	return false;
end;

function msfZone:unitEnter(_entered_unit)
	table.insert(self.enteredUnits, _entered_unit);
	for _onUnitEnterEventHandlerKey, _onUnitEnterEventHandler in pairs(self.onUnitEnter) do
		_onUnitEnterEventHandler(self, { enteredUnit = _entered_unit });
	end
end;

function msfZone:unitLeave(_left_unit)
	self:removeUnitFormZone(_left_unit);
	for _onUnitLeaveEventHandlerKey, _onUnitLeaveEventHandler in pairs(self.onUnitLeave) do
		_onUnitLeaveEventHandler(self, { leftUnit = _left_unit });
	end
end;

function msfZone:addOnUnitEnterEventHandler(eventHandler)
	self.onUnitEnter[eventHandler] = eventHandler;
end;

function msfZone:removeOnUnitEnterEventHandler(eventHandler)
	self.onUnitEnter[eventHandler] = nil;
end;

function msfZone:addOnUnitLeaveEventHandler(eventHandler)
	self.onUnitLeave[eventHandler] = eventHandler;
end

function msfZone:removeOnUnitLeaveEventHandler(eventHandler)
	self.onUnitLeave[eventHandler] = nil;
end;

-- way points: 
msfWayPoint = {};

function msfWayPoint:construct(parameters) 
	self.ETA = parameters.ETA;
	self.point = { x = parameters.point.x, y = parameters.point.y };
	self.onReach = Factory.create(msfEvent, {});
	self.onMoveOn = Factory.create(msfEvent, {});
	self.name = parameters.name;
	self.isThisCurrent = false;
	self.group = parameters.group;
	self.uniqueId = parameters.id;
	self.wpType = 0; -- only for aircarfts
	self.turnMethod = 0; -- only for aircrafts
end;

function msfWayPoint:getType() return self.wpType; end;
function msfWayPoint:setType(_type) self.wpType = _type; end;

function msfWayPoint:getTurnMethod() return self.turnMethod; end;
function msfWayPoint:setTurnMethod(_turn_method) self.turnMethod = _turn_method; end;

function msfWayPoint:getUniqueID()
	return self.uniqueId;
end;

function msfWayPoint:getPoint()
	return self.point;
end;

function msfWayPoint:getName() 
	return self.name;
end;

function msfWayPoint:setETA(_ETA)
	self.ETA = _ETA;
end;

function msfWayPoint:setCurrent(_is_current)
	self.isThisCurrent = _is_current;
end;

function msfWayPoint:isCurrent()
	return self.isThisCurrent;
end;

function msfWayPoint:addOnMoveOnEventHandler(eventHandler)
	self.onMoveOn[eventHandler] = eventHandler;
end;

function msfWayPoint:removeOnMoveOnEventHandler(eventHandler)
	self.onMoveOn[eventHandler] = nil;
end;

function msfWayPoint:addOnReachEventHandler(eventHandler)
	self.onReach[eventHandler] = eventHandler;
end;

function msfWayPoint:removeOnReachEventHandler(eventHandler)
	self.onReach[eventHandler] = nil;
end;

function msfWayPoint:isOk()
	if (self.ETA > (timer.getTime() + 4) and self.point.x > 0 and self.point.y > 0) then 
		return true;
	end
	return false;
end;

function msfWayPoint:moveOn()
	local _eventArgs = { timeOfMoveOn = timer.getTime(), movingGroup = self.group }
	
	for _onMoveOnEventHandlerKey, _onMoveOnEventHandler in pairs(self.onMoveOn) do 
		_onMoveOnEventHandler(self, _eventArgs);
	end
end;

function msfWayPoint:reach()
	local _eventArgs = { timeOfArrival = timer.getTime(), arrivedGroup = self.group }
	
	for _index = 1, #self.group.waypoints do
		if true == self.group.waypoints[_index]:isCurrent() then
			self.group.waypoints[_index]:setCurrent(false); break;
		end
	end
	self.isThisCurrent = true;
	
	for _onReachEventHandlerKey, _onReachEventHandler in pairs(self.onReach) do 
		_onReachEventHandler(self, _eventArgs);
	end
end;

-- groups:
msfGroup = {
	["tasks"] = {},
	["orders"] = {}
};

function msfGroup:init(eventArgs) 
	for _onInitEventHandlerKey, _onInitEventHandler in pairs(self.onInit) do
		_onInitEventHandler(self, eventArgs);
	end
end;

function msfGroup:hit(eventArgs)
	for _onHitEventHandlerKey, _onHitEventHandler in pairs(self.onHit) do
		_onHitEventHandler(self, eventArgs);
	end
end;

function msfGroup:readyToShot(eventArgs)
	for _onReadyToShotEventHandlerKey, _onReadyToShotEventHandler in pairs(self.onReadyToShot) do
		_onReadyToShotEventHandler(self, eventArgs);
	end
end;

function msfGroup:prepareToShot(eventArgs)
	for _onPrepareToShotEventHandlerKey, _onPrepareToShotEventHandler in pairs(self.onPrepareToShot) do
		_onPrepareToShotEventHandler(self, eventArgs);
	end
end;

function msfGroup:addTask(taskName, _functionToCall, _argument, _time)
	self.tasks[taskName] = Factory.create(msfTask, { functionToCall = _functionToCall, functionArgument = _argument, startTime = _time });
end;

function msfGroup:construct(parameters) 
	self.groupName = parameters.newGroupName;
	self.dcsWorldGroup = parameters.newGroup;
	self.onInit = Factory.create(msfEvent, {});
	self.onHit = Factory.create(msfEvent, {});
	self.onShot = Factory.create(msfEvent, {});
	self.onDead = Factory.create(msfEvent, {});
	self.onTakeOff = Factory.create(msfEvent, {});
	self.onPrepareToShot = Factory.create(msfEvent, {}); -- only for artillery
	self.onReadyToShot = Factory.create(msfEvent, {}); -- only for artillery 
	self.onPlayerEnterUnit = Factory.create(msfEvent, {});
	self.onPlayerLeaveUnit = Factory.create(msfEvent, {});
	self.onDetectThreat = Factory.create(msfEvent, {});
	self.groupHasStopped = false;
	self.targets = {};  -- only for artillery
	self.artilleryObservers = {}; -- only for artillery
	self.maxEngageDistance = 0; -- only for artillery
	self.targetMarkedBySmoke = false; -- only for artillery
	self.messagesQueue = {};
	self.waypoints = {};
	self.airbase = 0; -- only for aircrafts
	self.airbase2 = 0; -- only for aircrafts
	self.avalibleTransportList = {};
	self.groupTemplate = {};
	--self.radioCommands = {}; -- command id, path, executed or not, subItem or not, subMenu
	self.cargoPlaceRadioCommand = nil;
	self.audioMessageReciverStart = -1;
	self.audioMessageLength = 0;
	self.lastThreatDetectionTime = 0;
end;

function msfGroup:addOnDetectThreatEventHandler(_event_handler) self.onDetectThreat[_event_handler] = _event_handler;
end;

function msfGroup:removeOnDetectThreatEventHandler(_event_handler) self.onDetectThreat[_event_handler] = nil;
end;

function msfGroup:detectThreat(_detector, _threat)
	if math.abs(timer.getTime() - self.lastThreatDetectionTime) < 45 then return; end
	
	self.lastThreatDetectionTime = timer.getTime();
	
	for _handler_name, _handler in pairs(self.onDetectThreat) do
		local _status, _error = pcall(function() _handler(self, { detector = _detector, threat = _threat }); return 0; end);
		
		if not _status then 
			env.error('MSF Error: error on attemp to handle threat detection.\n'
				.. 'Error: ' .. tostring(_error) .. '\n'
				.. 'EventHandler name: ' .. tostring(_handler_name), false);
			mission.controller.sendMessageForAll('MSF Error: error on attemp to handle threat detection.\n'
				.. 'Error: ' .. tostring(_error) .. '\n'
				.. 'EventHandler name: ' .. tostring(_handler_name));
		end
	end
end;

function msfGroup:setRadioCommunicationsWithAirDefence(_value)
	if _value then 
		local _add_air_defence_groupment_radio = function(_groupment, _msf_unit, _radio_menu, _function_to_add)
				local _groupment_radio = _msf_unit:addRadioCommand('msf_radio_comm_with_ad' .. _groupment:getName(), 
					_groupment:getDisplayName(), true, nil, nil, _radio_menu); 

				if not _groupment_radio then 
					env.error('MSF Error: could not set on radio communications with air defence for ' .. self.groupName 
						.. ', cause could not add item for ' .. _groupment:getName(), false);
					return;
				end
				
				local _, _subgroupments_radio = _msf_unit:addRadioCommand('msf_radio_comm_with_ad' .. _groupment:getName() .. '_get_closest_ac', 
					'Ближайший Л/A противника', false, function(_data) 
							_data.groupment:requestClosestBandit(_data.requester); 
						end, { groupment = _groupment, requester = _msf_unit:getDCSWorldUnit() }, _groupment_radio), 
					_msf_unit:addRadioCommand('msf_radio_comm_with_ad' .. _groupment:getName() .. '_subgroupments', 
						'Подчиненные группировки', true, nil, nil, _groupment_radio); 
				
				for _, _subgroupment in pairs(_groupment.subGroupments) do 
					_function_to_add(_subgroupment, _msf_unit, _subgroupments_radio, _function_to_add);
				end
			end;
	
		local _dcsw_group = self:getDCSWorldGroup();
		if not _dcsw_group then return; end
		
		local _dcsw_units, _unit, _air_defence_radio, _category = _dcsw_group:getUnits(), nil, nil, _dcsw_group:getCategory();
		local _group_coalition = _dcsw_group:getCoalition();
		if #_dcsw_units < 1 then return; end

		for _, _dcsw_unit in ipairs(_dcsw_units) do 
			if mission.model.units[_dcsw_unit:getName()] then 
				_unit = mission.model.units[_dcsw_unit:getName()];
				if _unit:isExistRadioMenuItem('msf_radio_communcations_with_ad') then 
					_unit:removeRadioCommandByName('msf_radio_communcations_with_ad');
				end
			end
		end
		if not _unit then 
			mission.model:addUnit(_dcsw_units[1]:getName());
			if not mission.model.units[_dcsw_units[1]:getName()] then 
				env.error('MSF Error: could not set on radio communications with air defence for ' .. self.groupName 
					.. ', cause units of group are not controlled by MSF', false);
				return;
			else _unit = mission.model.units[_dcsw_units[1]:getName()]; end
		end		
		_air_defence_radio = _unit:addRadioCommand('msf_radio_communcations_with_ad', 'ПВО', true, nil, nil, nil); 

		if not _air_defence_radio then 
			env.error('MSF Error: could not set on radio communications with air defence for ' .. self.groupName 
				.. ', cause radio menu item msf_radio_communcations_with_ad does not exist.', false);
			return;
		end
		for _, _air_defence_groupment in pairs(mission.model.airDefenceGroupments) do 
			if _group_coalition == _air_defence_groupment:getCoalition() then
				_add_air_defence_groupment_radio(_air_defence_groupment, _unit, _air_defence_radio, _add_air_defence_groupment_radio);
			end
		end
		_add_air_defence_groupment_radio = nil;
	end
	if not _value then error('MSF Error: set Off radio communications with air defence has not implemented yet');
	end
end;

function msfGroup:startAudioReciving(_message_length)
	self.audioMessageReciverStart, self.audioMessageLength = timer.getTime(), _message_length;
end;

function msfGroup:audioReciverRealse()
	self.audioMessageReciverStart, self.audioMessageLength = -1, 0;
end;

function msfGroup:isAudioReciverLocked()
	if -1 ~= self.audioMessageReciverStart then 
		if (self.audioMessageReciverStart + self.audioMessageLength) > timer.getTime() then return true; end
	end
	self.audioMessageReciverStart, self.audioMessageLength = -1, 0;
	return false;
end;


function msfGroup:outSound(...)
	local _dcsw_group = self:getDCSWorldGroup();
	if not _dcsw_group then return; end
	
	local _group_id = _dcsw_group:getID(); 
	local _time_lag = 0.5;
	
	for _sound_key_index = 1, arg.n do
		if mission.model.sounds[arg[_sound_key_index]] then timer.scheduleFunction(function(_data)
					trigger.action.outSoundForGroup(_data.groupId, _data.path); return nil;
				end, { path = mission.model.sounds[arg[_sound_key_index]].path, groupId = _group_id }, timer.getTime() + _time_lag);
			_time_lag = _time_lag + mission.model.sounds[arg[_sound_key_index]].length;
		end
	end
end;

function msfGroup:getGroupName() return self.groupName; end;
function msfGroup:getCargoPlacesRadio() return self.cargoPlaceRadioCommand; end;
function msfGroup:setCargoPlacesRadio(_radio_command) self.cargoPlaceRadioCommand = _radio_command; end;

function msfGroup:renewCargoPlacesList()
	local _self_dcsw_group = self:getDCSWorldGroup();
	if not _self_dcsw_group then 
		env.info('MSF INFO: could not execute msfGroup:renewCargoPlacesList() for ' .. assert(self.groupName, 'self.groupName has nil value') 
			.. ', cause group does not exist', false); 
		return; 
	end
	if not (Group.Category.GROUND == _self_dcsw_group:getCategory()) then return; end
	
	local _group_units = _self_dcsw_group:getUnits();
		
	for _unit_index, _dcsw_unit in ipairs(_group_units) do
		local _msf_unit = mission.model.units[_dcsw_unit:getName()];
		if _msf_unit then if _msf_unit:hasTransportTask() then
				_msf_unit:resetSelectedCargoTemplateName(); -- удалить информацию о выборе шаблона груза
				
				local _cargo_places_list = nil;
				local _self_cargo_places_list = nil;
				
				_cargo_places_list = self.cargoPlaceRadioCommand;
				
				if not _cargo_places_list then
					local _path = missionCommands.addSubMenuForGroup(_self_dcsw_group:getID(), 'Пункты подбора десанта', nil);
					_cargo_places_list = { id = -1, 
						name = 'msfCargo_' .. self.groupName .. '_cargoPlacesList', path = _path, executed = false, isSubMenu = true, subMenu = nil };
						self.cargoPlaceRadioCommand = _cargo_places_list;
				end
				if not _cargo_places_list then env.error('MSF Error: msfUnit:setTransportTask() could not get\add radio command ' 
					.. 'msfCargo_' .. self.groupName .. '_cargoPlacesList', false); return; 
				end
				
				if not _msf_unit:isExistRadioMenuItem('msfCargo_' .. self.groupName .. '_cargoPlacesListFor_' .. _msf_unit:getUnitName()) then
					_self_cargo_places_list = _msf_unit:addRadioCommand('msfCargo_' .. self.groupName .. '_cargoPlacesListFor_' .. _msf_unit:getUnitName(), 
						'Пункты подбора десанта для ' .. _dcsw_unit:getTypeName() .. ' (' .. _msf_unit:getUnitName() .. ')', true, _msfOnResetCargoPlaceTemplate, {}, _cargo_places_list);
				else
					_self_cargo_places_list = _msf_unit:getRadioMenuItemByName('msfCargo_' .. self.groupName .. '_cargoPlacesListFor_' .. _msf_unit:getUnitName());
				end
				if not _self_cargo_places_list then env.error('MSF Error: msfUnit:setTransportTask() could not get\add radio command ' 
					.. 'msfCargo_' .. self.groupName .. '_cargoPlacesListFor_' .. _msf_unit:getUnitName(), false); return; 
				end
				
				for _cargo_place_index, _cargo_place in ipairs(mission.model.cargoPlaces) do
					local _initiator_unit_coalition = _dcsw_unit:getCoalition(); 
					if _initiator_unit_coalition == _cargo_place.coalition then -- если коалиция юнита соответствует зоне погрузки
						local _suit_templates_to_add = {};
						for  _cargo_template_index, _cargo_template in ipairs(_cargo_place.cargoTemplates) do
							local _initiator_unit_type = _dcsw_unit:getTypeName();
							if _initiator_unit_type == _cargo_template.unitType then -- если тип юнита соответствует шаблону
								table.insert(_suit_templates_to_add, { name = _cargo_template.name, displayName = _cargo_template.displayName;  });
							end
						end
						if #_suit_templates_to_add > 0 then 
							local _cargo_place_submenu = nil;
							if not _msf_unit:isExistRadioMenuItem('msfCargoPlace_' .. _cargo_place_index) then 
								_cargo_place_submenu = _msf_unit:addRadioCommand('msfCargoPlace_' .. _cargo_place_index, _cargo_place.name, 
									true, _msfOnResetCargoPlaceTemplate, {}, _self_cargo_places_list);
							else
								_cargo_place_submenu = _msf_unit:getRadioMenuItemByName('msfCargoPlace_' .. _cargo_place_index);
							end
							for _cargo_template_to_add_index, _cargo_template_to_add in ipairs(_suit_templates_to_add) do
								if not _msf_unit:isExistRadioMenuItem('msfCargoPlace_' .. _cargo_place_index .. '_' .. _cargo_template_to_add.name) then 
									_msf_unit:addRadioCommand('msfCargoPlace_' .. _cargo_place_index .. '_' .. _cargo_template_to_add.name, 
										_cargo_template_to_add.displayName, false, 
										_msfOnSelectCargoPlaceTemplate, { unit = _msf_unit, cargoTemplateName = _cargo_template_to_add.name }, 
										_cargo_place_submenu);
								end
							end
							if not _msf_unit:isExistRadioMenuItem('msfCargoPlace_' .. _cargo_place_index .. '_reset') then
								_msf_unit:addRadioCommand('msfCargoPlace_' .. _cargo_place_index .. '_reset', 
									'Отменить выбор типа десанта', false, _msfOnResetCargoPlaceTemplate, { unit = _unit }, _cargo_place_submenu);
							end
						end
					end
				end
			end -- has unit transport task
		end -- is unit under MSF control
	end
end;

function msfGroup:merge(_group)
	local _self_dcsw_group, _dcsw_group = self:getDCSWorldGroup(), _group:getDCSWorldGroup();
	if not _self_dcsw_group then 
		env.info('MSF INFO: could not execute msfGroup:merge() for ' .. assert(self.groupName, 'self.groupName has nil value') 
			.. ', cause group does not exist', false); 
		return; 
	end
	if not _dcsw_group then 
		env.info('MSF INFO: could not execute msfGroup:merge() for ' .. assert(self.groupName, 'self.groupName has nil value') 
			.. ', cause argument #1 has wrong value.', false);
		return;
	end
	
	local _first_part_units = _self_dcsw_group:getUnits();
	local _sacond_part_units = _dcsw_group:getUnits();
	
	local _new_group_units = {};
	local _units_to_delete = {};
	local _new_group_point = { x = 0, y = 0, z = 0 };
	local _new_group_country = 0;
	
	for _unit_index, _unit in ipairs(_first_part_units) do
		local _unit_point = _unit:getPoint();
		_new_group_point.x, _new_group_point.y, _new_group_point.z = _unit_point.x,_unit_point.y, _unit_point.z; 
		_new_group_country = _unit:getCountry();
		table.insert(_new_group_units, { unitType = _units[_unit_index]:getTypeName(), 
			unitName = _units[_unit_index]:getName(), playerCanDrive = false, point = { x = _unit_point.x, y = _unit_point.z  } });
		table.insert(_units_to_delete, _unit);
	end
	for _unit_index, _unit in ipairs(_sacond_part_units) do
		local _unit_point = _unit:getPoint();
		table.insert(_new_group_units, { unitType = _units[_unit_index]:getTypeName(), 
			unitName = _units[_unit_index]:getName(), playerCanDrive = false, point = { x = _unit_point.x, y = _unit_point.z  } });
		table.insert(_units_to_delete, _unit);
	end
	for _unit_index, _unit in ipairs(_units_to_delete) do _unit:destroy(); end
	
	local _group_descriptor = mission.utils.getGroupDescriptor(self.groupName, _new_group_point, _new_group_units);
	
	self.dcsWorldGroup = coalition.addGroup(_new_group_country, Group.Category.GROUND, _descriptor);
	
	mission.model.groups[_group:getGroupName()] = nil;
end;

function msfGroup:split(...)
	local _self_dcsw_group = self:getDCSWorldGroup();
	if not _self_dcsw_group then 
		env.info('MSF INFO: could not execute msfGroup:split() for ' .. assert(self.groupName, 'self.groupName has nil value') 
			.. ', cause group does not exist', false); 
		return; 
	end
	
	local _units = _self_dcsw_group:getUnits();
	if #_units < 1 then return; end
	
	local _new_group_units = {};
	local _units_templates = {};
	local _units_to_delete = {};
	local _new_group_index = 1;
	local _new_group_name = '';
	local _new_group_point = { x = 0, y = 0, z = 0 };
	local _new_group_country = 0;
	
	while(function()
			if _new_group_index > 1000 then return false; end -- end cycle couse of too many attempts.
				
			_new_group_name = self.groupName .. '_' .. _new_group_index;
				
			local _test_exist_group = Group.getByName(_new_group_name);
			if nil ~= _test_exist_group then return true;  
			else return false; end
		end)() do
		_new_group_index = _new_group_index + 1;
	end
	for i = 1, arg.n do
        local _unit_index = arg[i];
		if (#_units) > _unit_index then 
			local _unit_point = _units[_unit_index]:getPoint();
			_new_group_point.x, _new_group_point.y, _new_group_point.z = _unit_point.x, _unit_point.y + 0.1, _unit_point.z;
			_new_group_country = _units[_unit_index]:getCountry();
			table.insert(_new_group_units, { unitType = _units[_unit_index]:getTypeName(), unitName = _units[_unit_index]:getName(), playerCanDrive = false, 
				point = { x = _unit_point.x, y = _unit_point.z  } });
			table.insert(_units_to_delete, _units[_unit_index]);
			table.insert(_units_templates, { unitType = _units[_unit_index]:getTypeName(), playerCanDrive = false, pointOffset = { x = 0, y = 0 } });
		end
    end
	for _index, _unit in ipairs(_units_to_delete) do _unit:destroy(); end
	
	local _group_descriptor = mission.utils.getGroupDescriptor(_new_group_name, _new_group_point, _new_group_units);
	local _created_msf_group = mission.model:createGroup(Group.Category.GROUND, _new_group_country, _group_descriptor);
	
	for _avalible_transport_index, _avalible_transport_name in ipairs(self.avalibleTransportList) do
		_created_msf_group:addAvalibleTransport(_avalible_transport_name);
	end
	
	if self.groupTemplate and self.groupTemplate.conditions then 
		local _unload_conditions_command = nil;
		if nil ~= self.groupTemplate.conditions.command then _unload_conditions_command = 0; end
		
		_created_msf_group:setGroupTemplate({ 
			units = _units_templates, 
			unloadTime = ((#_units_templates) * 2),
			groupName = _new_group_name,
			cargoPlace = { point = _new_group_point, radius = 2 },
			countryId = _new_group_country,
			conditions = { 
				altitudeLessThen = self.groupTemplate.conditions.altitudeLessThen,
				speedLessThen = self.groupTemplate.conditions.speedLessThen,
				command = _unload_conditions_command } });
	end
	return _created_msf_group;
end;

function msfGroup:takeCover()
	local _self_dcsw_group = self:getDCSWorldGroup();
	if not _self_dcsw_group then 
		env.info('MSF INFO: could not execute msfGroup:takeCover() for ' .. assert(self.groupName, 'self.groupName has nil value') 
			.. ', cause group does not exist', false);
		return;
	end
	if not (Group.Category.GROUND == _self_dcsw_group:getCategory()) then return; end
	local _units = _self_dcsw_group:getUnits();
	local _coalition = _self_dcsw_group:getCoalition();
	
	if #_units < 1 then return; end
	
	self:atEase();
	
	local _enviroment_data = { hideOuts = {}, unitEnemies = {}, hiddingGroups = {}, coalition = _coalition };
	local _hide_outs_categories = {};
	local _start_point = _units[1]:getPoint();
	
	table.insert(_hide_outs_categories, Object.Category.STATIC);
	table.insert(_hide_outs_categories, Object.Category.SCENERY);
	
	-- поиск укрыйтий:
	world.searchObjects(_hide_outs_categories, { 
		id = world.VolumeType.SPHERE, -- back to SPHERE (point = _start_point, radius = 500) after fix!!! 
		params = {
			point = _start_point,
			radius = 500} }, 
		function(_object, _data)
			local _object_desc = _object:getDesc();
			if _object_desc.box then table.insert(_data.hideOuts, { hideOut = _object, using = 0 });
			end
			return true;
		end
		, _enviroment_data);
	
	-- разбиение группы на пары:
	local _units_pairs = {};
	if #_units < 2 then table.insert(_units_pairs, self); 
	else
		local _internal_part, _factorial_part = math.modf((#_units) / 2);
		for _pair_index = 1, _internal_part do
			local _splitted_group = self:split(1, 2);
			table.insert(_units_pairs, _splitted_group);
		end
		if _factorial_part > 0 then table.insert(_units_pairs, self); end
	end
	_enviroment_data.hiddingGroups = _units_pairs;
	
	-- поиск юнитов противника: 
	world.searchObjects(Object.Category.UNIT, { 
		id = world.VolumeType.SPHERE, -- back to SPHERE (point = _start_point, radius = 10000) after fix!!!  
		params = {
			point = _start_point,
			radius = 10000} }, 
		function(_object, _data)
			local _enemy_position = _object:getPosition();
			if _data.coalition == _object:getCoalition() then return true; end
			
			for _msf_group_index, _msf_group in ipairs(_data.hiddingGroups) do
				local _dcsw_group = _msf_group:getDCSWorldGroup();
				local _dcsw_units = _dcsw_group:getUnits();
				if (#_dcsw_units) > 0 then
					local _distance = mission.utils.getDistance(_dcsw_units[1]:getPoint(), _enemy_position.p);
					local _unit_id = _dcsw_units[1]:getID();
					if not _data.unitEnemies[_unit_id] then _data.unitEnemies[_unit_id] = { closestEnemy = _object, distance = _distance, group = _msf_group };
					else if _data.unitEnemies[_unit_id].distance > _distance then 
						_data.unitEnemies[_unit_id].closestEnemy = _object; _data.unitEnemies[_unit_id].distance = _distance; end
					end
				end
			end
			return true;
		end
		, _enviroment_data);
	for _unit_id, _unit_enemy_pair in pairs(_enviroment_data.unitEnemies) do
		local _closest_distance = 0;
		local _closest_hideout_index = 0;
		local _dcsw_group = _unit_enemy_pair.group:getDCSWorldGroup();
		local _dcsw_units = _dcsw_group:getUnits();
		local _unit_position = _dcsw_units[1]:getPoint();
		
		local _hideout_index = 0;
		while(function()
				if 0 == _hideout_index then return true; end
				if _hideout_index > (#_enviroment_data.hideOuts) then return false; end
				local _current_hideout = _enviroment_data.hideOuts[_hideout_index];
				
				if _current_hideout.using > 0 then return true; end
				local _distance = mission.utils.getDistance(_unit_position, _current_hideout.hideOut:getPoint());
				
				if 0 == _closest_hideout_index then 
					_closest_hideout_index = _hideout_index;
					_closest_distance = _distance; 
					_current_hideout.using = _current_hideout.using + 1;
					return true;
				end
				
				if _closest_distance > _distance and 0 ~= _closest_hideout_index then 
					_enviroment_data.hideOuts[_closest_hideout_index].using = (_enviroment_data.hideOuts[_closest_hideout_index].using - 1);
					_closest_hideout_index = _hideout_index;
					_closest_distance = _distance; 
					_current_hideout.using = _current_hideout.using + 1;
					return true;
				end
			end)() do
			_hideout_index = _hideout_index + 1;
		end
		if _closest_hideout_index > 0 then -- если нашлось укрытие
			local _hideout_desc = _enviroment_data.hideOuts[_closest_hideout_index].hideOut:getDesc();
			local _hideout_position = _enviroment_data.hideOuts[_closest_hideout_index].hideOut:getPosition();
			local _hideout_min_point = mission.utils.transformPoint(_hideout_position, _hideout_desc.box.min);
			local _hideout_max_point = mission.utils.transformPoint(_hideout_position, _hideout_desc.box.max);
			local _enemy_unit_point = _unit_enemy_pair.closestEnemy:getPoint();
			local _unit_desc = _dcsw_units[1]:getDesc();
			
			local _x = 0;
			local _y = 0;
			if math.abs(_hideout_max_point.x - _enemy_unit_point.x) > math.abs(_hideout_min_point.x - _enemy_unit_point.x) then _x = _hideout_max_point.x;
			else _x = _hideout_min_point.x; end
			if math.abs(_hideout_max_point.z - _enemy_unit_point.z) > math.abs(_hideout_min_point.z - _enemy_unit_point.z) then _y = _hideout_max_point.z;
			else _y = _hideout_min_point.z; end
			
			local _cover_wp = _unit_enemy_pair.group:addAirWaypoint('take_cover', _x, _y);
			
			_cover_wp:addOnReachEventHandler(function(sender, eventArgs)
					eventArgs.arrivedGroup:danger();
				end);
			
			timer.scheduleFunction(function(_data)
					local _status, _result = pcall(function()
						_data.group:moveTo(_data.wpId, 10, _data.speed, AI.Task.VehicleFormation.OFF_ROAD, true);
					end);
					if not _status then mission.controller:sendMessageForAll('MSF Error: ' .. _result); evn.error('MSF Error: ' .. _result, false); end
					return nil;
				end, { group = _unit_enemy_pair.group, wpId = _cover_wp:getUniqueID(), speed = _unit_desc.speedMax }, timer.getTime() + 1);
		end -- if укрытие найдено для юнита 
	end -- цикл по группам 
	return _enviroment_data.hiddingGroups;
end;

function msfGroup:setGroupTemplate(_group_template)
	self.groupTemplate = _group_template;
end;
function msfGroup:getGroupTemplate()
	return self.groupTemplate;
end;

function msfGroup:doesTransportSuit(_transport_type_name)
	for _transport_type_index, _transport_type in ipairs(self.avalibleTransportList) do
		if _transport_type == _transport_type_name then return true; end
	end
	return false;
end;
function msfGroup:addAvalibleTransport(_transport_type_name)
	table.insert(self.avalibleTransportList, _transport_type_name);
end;
function msfGroup:removeAvalibleTransport(_transport_type_name)
	local _transport_index_to_remove = 0;
	for _transport_type_index, _transport_type in ipairs(self.avalibleTransportList) do
		if _transport_type == _transport_type_name then _transport_index_to_remove = _transport_type_index; break; end
	end
	if _transport_index_to_remove > 0 then 
		table.remove(self.avalibleTransportList, _transport_index_to_remove);
	end
end;

function msfGroup:takeOff(eventArgs)
	for _onTakeOffEventHandlerKey, _onTakeOffEventHandler in pairs(self.onTakeOff) do
		_onTakeOffEventHandler(self, eventArgs);
	end
end;

function msfGroup:setFirstAirbase(_airbase) self.airbase = _airbase; end;
function msfGroup:setSacondAirbase(_airbase) self.airbase2 = _airbase; end;

function msfGroup:getFirstAirbase() return self.airbase; end;
function msfGroup:getSacondAirbase() return slef.airbase2; end;

function msfGroup:addAirWaypoint(_name, _x, _y, _type, _turn_method)
	local _count_wp_items = #self.waypoints;
	local _unique_id = 0;
	
	if _count_wp_items > 0 then
		_unique_id = (1 + self.waypoints[_count_wp_items]:getUniqueID());
	else
		_unique_id = 1;
	end
	
	local _new_waypoint = Factory.create(msfWayPoint, { ETA = 4, name = _name, point = { x = _x, y = _y }, group = self , id = _unique_id });
	
	_new_waypoint:setType(_type);
	_new_waypoint:setTurnMethod(_turn_method);
	
	table.insert(self.waypoints, _new_waypoint);
	return _new_waypoint;
end;

function msfGroup:addWayPoint(_name, _x, _y)
	local _count_wp_items = #self.waypoints;
	local _unique_id = 0;
	
	if _count_wp_items > 0 then
		_unique_id = (1 + self.waypoints[_count_wp_items]:getUniqueID());
	else
		_unique_id = 1;
	end
	
	local _new_waypoint = Factory.create(msfWayPoint, { ETA = 4, name = _name, point = { x = _x, y = _y }, group = self , id = _unique_id });
	table.insert(self.waypoints, _new_waypoint);
	return _new_waypoint;
end;

function msfGroup:getWaypointByID(_unique_id)
	for _index = 1,#self.waypoints do
		local _wp_unique_id = self.waypoints[_index]:getUniqueID();
		if _wp_unique_id == _unique_id then return self.waypoints[_index]; end
	end
	return nil;
end;

function msfGroup:removeLastWayPoint()
	table.remove(self.waypoints);
end;

function msfGroup:removeFirstWayPoint()
	if #self.waypoints > 0 then 
		if not self.waypoints[1]:isCurrent() then
			table.remove(self.waypoints, 1);
		end
	end
end;

function msfGroup:removeWayPoint(_number)
	if #self.waypoints < _number then 
		env.info('MSF: msfGroup:removeWayPoint(): could not delete WP with _number: ' .. _number .. '. Couse it does not exist.', false);
		return; 
	end
	table.remove(self.waypoints, _number);
end;

function msfGroup:removeWayPointById(_wp_id)
	local _index_to_remove = 0;
	for _index = 1, #self.waypoints do
		if _wp_id == self.waypoints[_index]:getUniqueID() then 
			_index_to_remove = _index; break;
		end
	end
	if 0 ~= _index_to_remove then
		table.remove(self.waypoints, _index_to_remove);
	else
		env.info('MSF: could not delete WP with _wp_id: ' .. _wp_id .. '. Couse it does not exist.', false);
	end
end;

function msfGroup:moveToPoint(_point, _speed, _formation, _on_reach_event_handler, _on_move_on_event_handler)
	local _self_dcsw_group = self:getDCSWorldGroup();
	if not _self_dcsw_group then 
		env.info('MSF INFO: could not execute msfGroup:moveToPoint() for ' .. tostring(self.groupName) .. ', cause group does not exist', false);
		return false, nil;
	end
	
	local _waypoint = self:addWayPoint('new_wp_' .. string.format("%d", _point.x) .. '_' .. string.format("%d", _point.y), 
		_point.x, _point.y, 0);
				
	local _waypoint_id = _waypoint:getUniqueID();
		
	_waypoint:addOnMoveOnEventHandler(_on_move_on_event_handler);
	_waypoint:addOnReachEventHandler(_on_reach_event_handler);
	
	local _units = _self_dcsw_group:getUnits();
	
	if #_units < 1 then 
		env.info('MSF: msfGroup:moveToPoint(): _group ' .. self.groupName .. ' has been destroyed.', false);
		return false, nil;
	end
	
	local _category = _self_dcsw_group:getCategory();
	local _wp_point = _waypoint:getPoint();
	local _start_position = _units[1]:getPoint();
	
	if _category == Group.Category.GROUND then 
		local _on_reach_script = 'mission.controller.reachWayPoint(\"' .. self.groupName .. '\", ' .. _waypoint_id .. ');';
		local _on_move_on_script = 'mission.controller.moveOnToWayPoint(\"' .. self.groupName .. '\", ' .. _waypoint_id .. ');';
			
		local _on_move_on_task = {
			id = 'WrappedAction', 
			params = { action = { id = 'Script', params = { command = _on_move_on_script; } } }
		};
			
		local _on_reach_task = {
			id = 'WrappedAction', 
			params = { action = { id = 'Script', params = { command = _on_reach_script; } } }
		};
			
		local _task = {
			id = 'Mission', 
			params = { route = { 
				points = { 
					[1] = {
						action = _formation,
						x = (_start_position.x + 0.2), 
						y = (_start_position.z + 0.2), 
						speed = _speed, 
						ETA = (timer.getTime() + 4), 
						ETA_locked = false,
						speed_locked = true,
						name = 'start_point', 
						task = _on_move_on_task },
					[2] = {
						action = _formation,
						x = _wp_point.x, 
						y = _wp_point.y, 
						speed = _speed,
						ETA = (timer.getTime() + 20),
						ETA_locked = false,
						speed_locked = true,
						name = _waypoint:getName(), 
						task = _on_reach_task
					}
				} 
			} } 
		};
			
		local _controller = _self_dcsw_group:getController();
			
		local _task_status, _task_error = pcall(function() _controller:setTask(_task); return 0; end);
			
		if not _task_status then 
			mission.model.messages.push('MSF: msfGroup:moveToPoint() task error: ' .. _task_error);
			env.error('MSF: msfGroup:moveToPoint() task error: ' .. _task_error, false);
			return false, _waypoint;
		end
	else
		error('MSF: msfGroup:moveToPoint(): _group ' .. self.groupName .. ' has category ' 
			.. mission.utils.getEnumItemNameById(_category, Group.Category) .. '. The Function does not implemented for this.');
		return false, _waypoint;
	end
	
	return true, _waypoint;
end;

function msfGroup:moveToObject(_object, _speed, _formation, _on_reach_event_handler, _on_move_on_event_handler)
	local _self_dcsw_group = self:getDCSWorldGroup();
	if not _self_dcsw_group then 
		env.info('MSF INFO: could not execute msfGroup:moveToObject() for ' .. assert(self.groupName, 'self.groupName has nil value') 
			.. ', cause group does not exist', false);
		return false, nil;
	end
	local _object_status, _object_is_exist = pcall(function() local _result = _object:isExist(); return _result; end);
	
	local _waypoint_status, _waypoint = false, nil;
	
	if _object_status and _object_is_exist then 
		local _object_position = _object:getPoint(); 
		
		_waypoint_status, _waypoint = self:moveToPoint(
			{ x = _object_position.x, y = _object_position.z }, _speed, _formation, _on_reach_event_handler, _on_move_on_event_handler);
	else
		env.error('MSF: msfGroup:moveToObject(): _group ' .. self.groupName .. ' could not move to _object, couse it does not exist.', false);
		return false, _waypoint;
	end
	return _waypoint_status, _waypoint;
end;

function msfGroup:rtb(_ETA, _speed, _lock_speed, _altitude)
	local _self_dcsw_group = self:getDCSWorldGroup();
	if not _self_dcsw_group then 
		env.info('MSF: msfGroup:rtb(): _group ' .. assert(self.groupName, 'self.groupName has nil value') .. ' does not exist', false);
		return false;
	end
	
	local _units = _self_dcsw_group:getUnits();
	
	if #_units < 1 then 
		env.info('MSF: msfGroup:rtb(): _group ' .. self.groupName .. ' has been destroyed.', false);
		return false;
	end
	
	local _category = _self_dcsw_group:getCategory();
	local _start_position = _units[1]:getPoint();
	
	if _category == Group.Category.AIRPLANE or _category == Group.Category.HELICOPTER then
		local _first_airbase_point = self.airbase:getPoint();
		local _sacond_airbase_point = self.airbase2:getPoint();
		
		local _rtb_airbase = self.airbase;
		
		local _distance_1 = math.sqrt(((_start_position.x - _first_airbase_point.x)^2) + ((_start_position.z - _first_airbase_point.z)^2));
		local _distance_2 = math.sqrt(((_start_position.x - _sacond_airbase_point.x)^2) + ((_start_position.z - _sacond_airbase_point.z)^2));
		
		if _distance_1 > _distance_2 then _rtb_airbase = self.airbase2; end
		
		local _helipad_id = nil;
		local _airfield_id = nil;
		
		local _rtb_airbase_desc = _rtb_airbase:getDesc();
		
		if _rtb_airbase_desc.category == Airbase.Category.AIRDROME or _rtb_airbase_desc.category == Airbase.Category.SHIP 
		then _airfield_id = _rtb_airbase:getID();
		else _helipad_id = _rtb_airbase:getID(); 
		end
		
		local _rtb_point = _rtb_airbase:getPoint();
		
		local _wp_speed = 10;
		local _wp_eta = (timer.getTime() + 20);
		local _wp_eta_lock = false;
		
		if _lock_speed and _speed > 0 then 
			_wp_speed = _speed;
		else
			if not _lock_speed and _ETA > 1 then 
				_wp_eta_lock = true;
				_wp_eta = (timer.getTime() + _ETA);
			end
		end
		
		local _task = {
			id = 'Mission', 
			params = { route = { 
				points = { 
					[1] = {
						type = AI.Task.WaypointType.TURNING_POINT,
						airdromeId = nil, 
						helipadId = nil,
						action = AI.Task.TurnMethod.FLY_OVER_POINT, 
						x = (_start_position.x + 0.2), 
						y = (_start_position.z + 0.2), 
						alt = _start_position.y,
						alt_type = AI.Task.AltitudeType.BARO,
						speed = 40,
						ETA = (timer.getTime() + 2),
						ETA_locked = true,
						speed_locked = false,
						name = 'start_point_land', 
						task = nil 
					},
					[2] = {
						type = AI.Task.WaypointType.LAND,
						airdromeId = _airfield_id, 
						helipadId = _helipad_id,
						action = AI.Task.TurnMethod.FIN_POINT,
						x = _rtb_point.x, 
						y = _rtb_point.z, 
						alt = _rtb_point.y,
						alt_type = AI.Task.AltitudeType.BARO,
						speed = _wp_speed,
						ETA = _wp_eta,
						ETA_locked = _wp_eta_lock,
						speed_locked = _lock_speed,
						name = 'land', 
						task = nil
					}
				} 
			} } 
		};
		
		local _controller = _self_dcsw_group:getController();
		
		local _task_status, _task_error = pcall(function() _controller:setTask(_task); return 0; end);
		
		if not _task_status then 
			mission.model.messages.push('MSF: msfGroup:rtb() task error: ' .. _task_error);
			env.error('MSF: msfGroup:rtb() task error: ' .. _task_error, false);
			return false;
		end
	else
		env.error('MSF: msfGroup:flyAhead(): _group ' .. self.groupName .. ' has category ' .. mission.utils.getEnumItemNameById(_category, Group.Category) 
			.. '. The Function does not implemented for this.', false);
		error('MSF: msfGroup:flyAhead(): _group ' .. self.groupName .. ' has category ' .. mission.utils.getEnumItemNameById(_category, Group.Category) 
			.. '. The Function does not implemented for this.');
		return false;
	end
	return false;
end;

function msfGroup:flyAhead(_ETA, _speed, _lock_speed, _altitude)
	local _waypoint = nil;
	local _waypoint_id = 0;
	
	if #self.waypoints < 1 then 
		env.info('MSF: msfGroup:flyAhead(): _group ' .. self.groupName .. ' can not fly. There is no way', false);
		return false;
	else
		_waypoint = self.waypoints[1];
		_waypoint_id  = _waypoint:getUniqueID();
	end
	
	for _index = 1, #self.waypoints do 
		if true == self.waypoints[_index]:isCurrent() then 
			if _index < #self.waypoints then
				_waypoint = self.waypoints[_index + 1];
				_waypoint_id = _waypoint:getUniqueID();
				break;
			else
				env.info('MSF: msfGroup:flyAhead(): _group ' .. self.groupName .. ' can not fly. There is no way', false);
				return false;
			end
		end
	end
	_waypoint:setETA(_ETA);
	
	local _self_dcsw_group = self:getDCSWorldGroup();
	if not _self_dcsw_group then 
		env.info('MSF: msfGroup:flyAhead(): _group ' .. assert(self.groupName, 'self.groupName has nil value') .. ' does not exist', false);
		return false;
	end
	local _units = _self_dcsw_group:getUnits();
	
	if #_units < 1 then 
		env.info('MSF: msfGroup:flyAhead(): _group ' .. self.groupName .. ' has been destroyed.', false);
		return false;
	end
	
	local _category = _self_dcsw_group:getCategory();
	local _wp_point = _waypoint:getPoint();
	local _start_position = _units[1]:getPoint();
	
	if _category == Group.Category.AIRPLANE or _category == Group.Category.HELICOPTER then
		local _on_reach_script = 'mission.controller.reachWayPoint(\"' .. self.groupName .. '\", ' .. _waypoint_id .. ');';
		local _on_move_on_script = 'mission.controller.moveOnToWayPoint(\"' .. self.groupName .. '\", ' .. _waypoint_id .. ');';
		
		local _on_move_on_task = {
			id = 'WrappedAction', 
			params = { action = { id = 'Script', params = { command = _on_move_on_script; } } }
		};
		
		local _on_reach_task = {
			id = 'WrappedAction', 
			params = { action = { id = 'Script', params = { command = _on_reach_script; } } }
		};
		
		local _wp_speed = 10;
		local _wp_eta = (timer.getTime() + 20);
		local _wp_eta_lock = false;
		
		if _lock_speed and _speed > 0 then 
			_wp_speed = _speed;
		else
			if not _lock_speed and _ETA > 1 then 
				_wp_eta_lock = true;
				_wp_eta = (timer.getTime() + _ETA);
			end
		end
		
		local _task = {
			id = 'Mission', 
			params = { route = { 
				points = { 
					[1] = {
						type = AI.Task.WaypointType.TURNING_POINT,
						airdromeId = nil, 
						helipadId = nil,
						action = AI.Task.TurnMethod.FLY_OVER_POINT, 
						x = (_start_position.x + 0.2), 
						y = (_start_position.z + 0.2), 
						alt = _start_position.y,
						alt_type = AI.Task.AltitudeType.BARO,
						speed = 40,
						ETA = (timer.getTime() + 2),
						ETA_locked = true,
						speed_locked = false,
						name = 'start_point_' .. _waypoint:getName(), 
						task = _on_move_on_task 
					},
					[2] = {
						type = _waypoint:getType(),
						airdromeId = nil, 
						helipadId = nil,
						action = _waypoint:getTurnMethod(),
						x = _wp_point.x, 
						y = _wp_point.y, 
						alt = _altitude,
						alt_type = AI.Task.AltitudeType.BARO,
						speed = _wp_speed,
						ETA = _wp_eta,
						ETA_locked = _wp_eta_lock,
						speed_locked = _lock_speed,
						name = _waypoint:getName(), 
						task = _on_reach_task
					}
				} 
			} } 
		};
		
		local _controller = _self_dcsw_group:getController();
		
		local _task_status, _task_error = pcall(function() _controller:setTask(_task); return 0; end);
		
		if not _task_status then 
			mission.model.messages.push('MSF: msfGroup:flyAhead() task error: ' .. _task_error);
			env.error('MSF: msfGroup:flyAhead() task error: ' .. _task_error, false);
			return false;
		end
		return true;
	else
		env.error('MSF: msfGroup:flyAhead(): _group ' .. self.groupName .. ' has category ' .. mission.utils.getEnumItemNameById(_category, Group.Category) 
			.. '. The Function does not implemented for this.', false);
		error('MSF: msfGroup:flyAhead(): _group ' .. self.groupName .. ' has category ' .. mission.utils.getEnumItemNameById(_category, Group.Category) 
			.. '. The Function does not implemented for this.');
		return false;
	end
	return false;
end;

function msfGroup:moveAhead(_ETA, _speed, _formation, _lock_speed)
	local _waypoint = nil;
	local _waypoint_id = 0;
	
	if #self.waypoints < 1 then 
		env.info('MSF: msfGroup:moveAhead(): _group ' .. self.groupName .. ' can not move. There is no way', false);
		return false;
	else
		_waypoint = self.waypoints[1];
		_waypoint_id  = _waypoint:getUniqueID();
	end
	
	for _index = 1, #self.waypoints do 
		if true == self.waypoints[_index]:isCurrent() then 
			if _index < #self.waypoints then
				_waypoint = self.waypoints[_index + 1];
				_waypoint_id = _waypoint:getUniqueID();
				break;
			else
				env.info('MSF: msfGroup:moveAhead(): _group ' .. self.groupName .. ' can not move. There is no way', false);
				return false;
			end
		end
	end
	_waypoint:setETA(_ETA);
	
	local _self_dcsw_group = self:getDCSWorldGroup();
	if not _self_dcsw_group then 
		env.info('MSF: msfGroup:moveAhead(): _group ' .. tostring(self.groupName) .. ' does not exist', false);
		return false;
	end
	local _units = _self_dcsw_group:getUnits();
	
	if #_units < 1 then 
		env.info('MSF: msfGroup:moveAhead(): _group ' .. self.groupName .. ' has been destroyed.', false);
		return false;
	end
	
	local _category = _self_dcsw_group:getCategory();
	local _wp_point = _waypoint:getPoint();
	local _start_position = _units[1]:getPoint();
	
	if _category == Group.Category.GROUND then 
		local _on_reach_script = 'mission.controller.reachWayPoint(\"' .. self.groupName .. '\", ' .. _waypoint_id .. ');';
		local _on_move_on_script = 'mission.controller.moveOnToWayPoint(\"' .. self.groupName .. '\", ' .. _waypoint_id .. ');';
		
		local _on_move_on_task = {
			id = 'WrappedAction', 
			params = { action = { id = 'Script', params = { command = _on_move_on_script; } } }
		};
		
		local _on_reach_task = {
			id = 'WrappedAction', 
			params = { action = { id = 'Script', params = { command = _on_reach_script; } } }
		};
		
		local _wp_speed = 10;
		local _wp_eta = (timer.getTime() + 20);
		local _wp_eta_lock = false;
		
		if _lock_speed and _speed > 0 then 
			_wp_speed = _speed;
		else
			if not _lock_speed and _ETA > 1 then 
				_wp_eta_lock = true;
				_wp_eta = (timer.getTime() + _ETA);
			end
		end
		
		local _task = {
			id = 'Mission', 
			params = { route = { 
				points = { 
					[1] = {
						action = _formation, 
						x = (_start_position.x + 0.2), 
						y = (_start_position.z + 0.2), 
						speed = 10,
						ETA = (timer.getTime() + 4),
						ETA_locked = true,
						speed_locked = false,
						name = 'start_point', 
						task = _on_move_on_task },
					[2] = {
						action = _formation,
						x = _wp_point.x, 
						y = _wp_point.y, 
						speed = _wp_speed,
						ETA = _wp_eta,
						ETA_locked = _wp_eta_lock,
						speed_locked = _lock_speed,
						name = _waypoint:getName(), 
						task = _on_reach_task
					}
				} 
			} } 
		};
		
		local _controller = _self_dcsw_group:getController();
		
		local _task_status, _task_error = pcall(function() _controller:setTask(_task); return 0; end);
		
		if not _task_status then 
			mission.model.messages.push('MSF: msfGroup:moveAhead() task error: ' .. _task_error);
			env.error('MSF: msfGroup:moveAhead() task error: ' .. _task_error, false);
			return false;
		end
	else
		env.error('MSF: msfGroup:moveAhead(): _group ' .. self.groupName .. ' has category ' .. mission.utils.getEnumItemNameById(_category, Group.Category) 
			.. '. The Function does not implemented for this.', false);
		error('MSF: msfGroup:moveAhead(): _group ' .. self.groupName .. ' has category ' .. mission.utils.getEnumItemNameById(_category, Group.Category) 
			.. '. The Function does not implemented for this.');
		return false;
	end
	return true;
end;

function msfGroup:moveTo(_wp_id, _ETA, _speed, _formation, _lock_speed)
	local _waypoint = nil;
	
	for _index = 1, #self.waypoints do 
		if _wp_id == self.waypoints[_index]:getUniqueID() then 
			_waypoint = self.waypoints[_index]; break;
		end
	end
	
	if nil == _waypoint then 
		mission.model.messages.push('MSF: msfGroup:moveTo(): _group ' .. self.groupName .. ' can not move. There is no way');
		env.info('MSF: msfGroup:moveTo(): _group ' .. self.groupName .. ' can not move. There is no way', false);
		return false;
	end
	
	_waypoint:setETA(_ETA);
	
	local _self_dcsw_group = self:getDCSWorldGroup();
	if not _self_dcsw_group then 
		env.info('MSF: msfGroup:moveTo(): _group ' .. assert(self.groupName, 'self.groupName has nil value') .. ' does not exist', false);
		return false;
	end
	local _units = _self_dcsw_group:getUnits();
	
	if #_units < 1 then 
		env.info('MSF: msfGroup:moveTo(): _group ' .. self.groupName .. ' has been destroyed.', false);
		return false;
	end
	
	local _category = _self_dcsw_group:getCategory();
	local _wp_point = _waypoint:getPoint();
	local _start_position = _units[1]:getPoint();
	
	if _category == Group.Category.GROUND then 
		local _on_reach_script = 'mission.controller.reachWayPoint(\"' .. self.groupName .. '\", ' .. _wp_id .. ');';
		local _on_move_on_script = 'mission.controller.moveOnToWayPoint(\"' .. self.groupName .. '\", ' .. _wp_id .. ');';
		
		local _on_move_on_task = {
			id = 'WrappedAction', 
			params = { action = { id = 'Script', params = { command = _on_move_on_script; } } }
		};
		
		local _on_reach_task = {
			id = 'WrappedAction', 
			params = { action = { id = 'Script', params = { command = _on_reach_script; } } }
		};
		
		local _wp_speed = 10;
		local _wp_eta = (timer.getTime() + 20);
		local _wp_eta_lock = false;
		
		if _lock_speed and _speed > 0 then 
			_wp_speed = _speed;
		else
			if not _lock_speed and _ETA > 1 then 
				_wp_eta_lock = true;
				_wp_eta = (timer.getTime() + _ETA);
			end
		end
		
		local _task = {
			id = 'Mission', 
			params = { route = { 
				points = { 
					[1] = {
						action = _formation, 
						x = (_start_position.x + 0.2), 
						y = (_start_position.z + 0.2), 
						speed = 10,
						ETA = (timer.getTime() + 4),
						ETA_locked = true,
						speed_locked = false,
						name = 'start_point', 
						task = _on_move_on_task },
					[2] = {
						action = _formation,
						x = _wp_point.x, 
						y = _wp_point.y, 
						speed = _wp_speed,
						ETA = _wp_eta,
						ETA_locked = _wp_eta_lock,
						speed_locked = _lock_speed,
						name = _waypoint:getName(), 
						task = _on_reach_task
					}
				} 
			} } 
		};
		
		local _controller = _self_dcsw_group:getController();
		local _task_status, _task_error = pcall(function() _controller:setTask(_task); return 0; end);
		
		if not _task_status then 
			mission.model.messages.push('MSF: msfGroup:moveTo() task error: ' .. _task_error);
			env.error('MSF: msfGroup:moveTo() task error: ' .. _task_error, false);
			return false;
		end
	else
		env.error('MSF: msfGroup:moveTo(): _group ' .. self.groupName .. ' has category ' .. mission.utils.getEnumItemNameById(_category, Group.Category) 
			.. '. The Function does not implemented for this.', false);
		error('MSF: msfGroup:moveTo(): _group ' .. self.groupName .. ' has category ' .. mission.utils.getEnumItemNameById(_category, Group.Category) 
			.. '. The Function does not implemented for this.');
		return false;
	end
	return true;
end;

function msfGroup:flyBack(_ETA, _speed, _lock_speed, _altitude)
	local _waypoint = nil;
	local _waypoint_id = 0;
	
	if #self.waypoints < 1 then 
		env.info('MSF: msfGroup:moveBack(): _group ' .. self.groupName .. ' can not move. There is no way', false);
		return false;
	else
		_waypoint = self.waypoints[1];
	end
	
	for _index = 1, #self.waypoints do 
		if true == self.waypoints[_index]:isCurrent() then 
			if _index > 2 then
				_waypoint = self.waypoints[_index - 1];
				break;
			else
				env.info('MSF: msfGroup:moveBack(): _group ' .. self.groupName .. ' can not move. There is no way', false);
				return false;
			end
		end
	end
	_waypoint:setETA(_ETA);
	_waypoint_id = _waypoint:getUniqueID();
	
	local _self_dcsw_group = self:getDCSWorldGroup();
	if not _self_dcsw_group then 
		env.info('MSF: msfGroup:flyBack(): _group ' .. assert(self.groupName, 'self.groupName has nil value') .. ' does not exist', false);
		return false;
	end
	local _units = _self_dcsw_group:getUnits();
	
	if #_units < 1 then 
		env.info('MSF: msfGroup:flyBack(): _group ' .. self.groupName .. ' has been destroyed.', false);
		return false;
	end
	
	local _category = _self_dcsw_group:getCategory();
	local _wp_point = _waypoint:getPoint();
	local _start_position = _units[1]:getPoint();
	
	if _category == Group.Category.AIRPLANE or _category == Group.Category.HELICOPTER then
		local _on_reach_script = 'mission.controller.reachWayPoint(\"' .. self.groupName .. '\", ' .. _waypoint_id .. ');';
		local _on_move_on_script = 'mission.controller.moveOnToWayPoint(\"' .. self.groupName .. '\", ' .. _waypoint_id .. ');';
		
		local _on_move_on_task = {
			id = 'WrappedAction', 
			params = { action = { id = 'Script', params = { command = _on_move_on_script; } } }
		};
		
		local _on_reach_task = {
			id = 'WrappedAction', 
			params = { action = { id = 'Script', params = { command = _on_reach_script; } } }
		};
		
		local _wp_speed = 10;
		local _wp_eta = (timer.getTime() + 20);
		local _wp_eta_lock = false;
		
		if _lock_speed and _speed > 0 then 
			_wp_speed = _speed;
		else
			if not _lock_speed and _ETA > 1 then 
				_wp_eta_lock = true;
				_wp_eta = (timer.getTime() + _ETA);
			end
		end
		
		local _task = {
			id = 'Mission', 
			params = { route = { 
				points = { 
					[1] = {
						type = AI.Task.WaypointType.TURNING_POINT,
						airdromeId = nil, 
						helipadId = nil,
						action = AI.Task.TurnMethod.FLY_OVER_POINT, 
						x = (_start_position.x + 0.2), 
						y = (_start_position.z + 0.2), 
						alt = _start_position.y,
						alt_type = AI.Task.AltitudeType.BARO,
						speed = 40,
						ETA = (timer.getTime() + 2),
						ETA_locked = true,
						speed_locked = false,
						name = 'start_point_' .. _waypoint:getName(), 
						task = _on_move_on_task 
					},
					[2] = {
						type = _waypoint:getType(),
						airdromeId = nil, 
						helipadId = nil,
						action = _waypoint:getTurnMethod(),
						x = _wp_point.x, 
						y = _wp_point.y, 
						alt = _altitude,
						alt_type = AI.Task.AltitudeType.BARO,
						speed = _wp_speed,
						ETA = _wp_eta,
						ETA_locked = _wp_eta_lock,
						speed_locked = _lock_speed,
						name = _waypoint:getName(), 
						task = _on_reach_task
					}
				} 
			} } 
		};
		
		local _controller = _self_dcsw_group:getController();
		local _task_status, _task_error = pcall(function() _controller:setTask(_task); return 0; end);
		
		if not _task_status then 
			mission.model.messages.push('MSF: msfGroup:flyAhead() task error: ' .. _task_error);
			env.error('MSF: msfGroup:flyAhead() task error: ' .. _task_error, false);
			return false;
		end
		return true;
	else
		env.error('MSF: msfGroup:flyAhead(): _group ' .. self.groupName .. ' has category ' .. mission.utils.getEnumItemNameById(_category, Group.Category) 
			.. '. The Function does not implemented for this.', false);
		error('MSF: msfGroup:flyAhead(): _group ' .. self.groupName .. ' has category ' .. mission.utils.getEnumItemNameById(_category, Group.Category) 
			.. '. The Function does not implemented for this.');
		return false;
	end
	return false;
end;

function msfGroup:moveBack(_ETA, _speed, _formation, _lock_speed)
	local _waypoint = nil;
	local _waypoint_id = 0;
	
	if #self.waypoints < 1 then 
		env.info('MSF: msfGroup:moveBack(): _group ' .. self.groupName .. ' can not move. There is no way', false);
		return false;
	else
		_waypoint = self.waypoints[1];
	end
	
	for _index = 1, #self.waypoints do 
		if true == self.waypoints[_index]:isCurrent() then 
			if _index > 2 then
				_waypoint = self.waypoints[_index - 1];
				break;
			else
				env.info('MSF: msfGroup:moveBack(): _group ' .. self.groupName .. ' can not move. There is no way', false);
				return false;
			end
		end
	end
	_waypoint:setETA(_ETA);
	_waypoint_id = _waypoint:getUniqueID();
	
	local _self_dcsw_group = self:getDCSWorldGroup();
	if not _self_dcsw_group then 
		env.info('MSF: msfGroup:moveBack(): _group ' .. assert(self.groupName, 'self.groupName has nil value') .. ' does not exist', false);
		return false;
	end
	local _units = _self_dcsw_group:getUnits();
	
	if #_units < 1 then 
		env.info('MSF: msfGroup:moveBack(): _group ' .. self.groupName .. ' has been destroyed.', false);
		return false;
	end
	
	local _category = _self_dcsw_group:getCategory(); 
	local _wp_point = _waypoint:getPoint();
	local _start_position = _units[1]:getPoint();
	
	if _category == Group.Category.GROUND then  
		local _on_reach_script = 'mission.controller.reachWayPoint(\"' .. self.groupName .. '\", ' .. _waypoint_id .. ');';
		local _on_move_on_script = 'mission.controller.moveOnToWayPoint(\"' .. self.groupName .. '\", ' .. _waypoint_id .. ');';
		
		local _on_move_on_task = {
			id = 'WrappedAction', 
			params = { action = { id = 'Script', params = { command = _on_move_on_script; } } }
		};
		
		local _on_reach_task = {
			id = 'WrappedAction', 
			params = { action = { id = 'Script', params = { command = _on_reach_script; } } }
		};
		
		local _wp_speed = 10;
		local _wp_eta = (timer.getTime() + 20);
		local _wp_eta_lock = false;
		
		if _lock_speed and _speed > 0 then 
			_wp_speed = _speed;
		else
			if not _lock_speed and _ETA > 1 then 
				_wp_eta_lock = true;
				_wp_eta = (timer.getTime() + _ETA);
			end
		end
		
		local _task = {
			id = 'Mission', 
			params = { route = { 
				points = { 
					[1] = {
						action = _formation, 
						x = (_start_position.x + 0.2), 
						y = (_start_position.z + 0.2), 
						speed = 10,
						ETA = (timer.getTime() + 4),
						ETA_locked = true,
						speed_locked = false,
						name = 'start_point', 
						task = _on_move_on_task },
					[2] = {
						action = _formation,
						x = _wp_point.x, 
						y = _wp_point.y, 
						speed = _wp_speed,
						ETA = _wp_eta,
						ETA_locked = _wp_eta_lock,
						speed_locked = _lock_speed,
						name = _waypoint:getName(), 
						task = _on_reach_task
					}
				} 
			} } 
		};
		
		local _controller = _self_dcsw_group:getController();
		local _task_status, _task_error = pcall(function() _controller:setTask(_task); return 0; end);
		
		if not _task_status then 
			mission.model.messages.push('MSF: msfGroup:moveBack() task error: ' .. _task_error);
			env.error('MSF: msfGroup:moveBack() task error: ' .. _task_error, false);
			return false;
		end
	else
		env.error('MSF: msfGroup:moveBack(): _group ' .. self.groupName .. ' has category ' .. mission.utils.getEnumItemNameById(_category, Group.Category) 
			.. '. The Function does not implemented for this.', false);
		error('MSF: msfGroup:moveBack(): _group ' .. self.groupName .. ' has category ' .. mission.utils.getEnumItemNameById(_category, Group.Category) 
			.. '. The Function does not implemented for this.');
		return false;
	end
	return true;
end;

function msfGroup:enableMarkTargetBySmoke(_enable)
	self.targetMarkedBySmoke = _enable;
end;

function msfGroup:isEnableMarkTargetsBySmoke()
	return self.targetMarkedBySmoke;
end;

function msfGroup:getDCSWorldGroup() 
	if not (self.dcsWorldGroup and self.dcsWorldGroup:isExist()) then 
		self.dcsWorldGroup = Group.getByName(self.groupName); 
		if not (self.dcsWorldGroup and self.dcsWorldGroup:isExist()) then return nil; end
	end
	return self.dcsWorldGroup; 
end;

function msfGroup:setDCSWorldGroup(_group) self.dcsWorldGroup = _group;  end;
function msfGroup:getGroupName() return self.groupName; end;

function msfGroup:recieveMessage()
	return table.remove(self.messagesQueue, 1);
end;

function msfGroup:addArtilleryObserver(_artillery_observer, _visibility_range)
	table.insert(self.artilleryObservers, { artilleryObserver = _artillery_observer, maxVisibilityRange = _visibility_range });
end;

function msfGroup:addTarget(_target) 
	table.insert(self.targets, { group = _target, point = {x = 1, y = 1}, isKnown = true });
end;

function msfGroup:requestTargetsForArtillery()
	
	for _artillery_observer_index = 1, #self.artilleryObservers do		
		
		local _is_observer_alive = self.artilleryObservers[_artillery_observer_index].artilleryObserver:isAlive();
		local _artillery_observer_dcsw_unit = self.artilleryObservers[_artillery_observer_index].artilleryObserver:getDCSWorldUnit();
		local _visibility_range = self.artilleryObservers[_artillery_observer_index].maxVisibilityRange;
		
		local _get_coalition_status, _artillery_observer_coalition = pcall(function()
			return _artillery_observer_dcsw_unit:getCoalition(); end);
			
		if not _get_coalition_status then return 0; end
		
		if coalition.side.RED == _artillery_observer_coalition and _is_observer_alive then 
			
			for _blueforce_country_index = 1, #env.mission.coalition.blue.country do
								
				local _blueforce_ground_groups = {};
				
				if env.mission.coalition.blue.country[_blueforce_country_index].vehicle then
					if env.mission.coalition.blue.country[_blueforce_country_index].vehicle.group then 
						_blueforce_ground_groups = env.mission.coalition.blue.country[_blueforce_country_index].vehicle.group;
					end
				end
				
				for _blueforce_group_index = 1, #_blueforce_ground_groups do				
					local _dcsw_group = Group.getByName(
						_blueforce_ground_groups[_blueforce_group_index].name);
						
					local _get_observer_position_status, _observer_position = pcall(function() 
						return _artillery_observer_dcsw_unit:getPosition(); end);
					
					if not _get_observer_position_status then return 0; end
					
					local _artillery_units = self.dcsWorldGroup:getUnits();
					
					if #_artillery_units < 1 then return 0; end

					local _artillery_position = _artillery_units[1]:getPosition();
					
					local _dcsw_group_units = _dcsw_group:getUnits();
					
					if #_dcsw_group_units < 1 then return 0; end
					
					local _dcsw_group_position = _dcsw_group_units[1]:getPosition();
					
					local _distance_to_observer = mission.utils.getDistance(_observer_position.p, _dcsw_group_position.p);
					local _distance_to_artillery = mission.utils.getDistance(_artillery_position.p, _dcsw_group_position.p);
					
					local _targets_to_remove = {};
					local _is_group_visibly = false; -- признак того, что группу не видит ни один из наводчиков
					local _is_known_group = false;
					
					for _target_index = 1, #self.targets do
						local _target_dcsw_group = self.targets[_target_index].group:getDCSWorldGroup();
						local _target_dcsw_group_id = _target_dcsw_group:getID();
						local _dcsw_group_id = _dcsw_group:getID();
						
						if _dcsw_group_id == _target_dcsw_group_id then 
							-- если такая группа уже есть:
							_is_known_group = true;
							-- проверяем в пределах ли она досигаемости
							if _distance_to_artillery < self.maxEngageDistance or self.maxEngageDistance == 0 then
								-- mission.model.messages.push('target: ' ..  _blueforce_ground_groups[_blueforce_group_index].name 
								--	.. ', distance to artillery: ' .. _distance_to_artillery .. ', max distance: ' .. self.maxEngageDistance); 
								-- если да, проверим видит ли её наводчик
								if _distance_to_observer < _visibility_range and land.isVisible(
									{ x = _observer_position.p.x, y = (_observer_position.p.y + 10), z = _observer_position.p.z}, 
									{ x = _dcsw_group_position.p.x, y = (_dcsw_group_position.p.y + 10), z = _dcsw_group_position.p.z }) then
									
									-- отметить, что информация о перемещеиях группы
									-- от наводчика поступает
									if not _is_group_visibly then 
										self.artilleryObservers[_artillery_observer_index].artilleryObserver:spot({ 
											group = self.targets[_target_index].group, isNew = false });
									end
									
									_is_group_visibly = true;
									self.targets[_target_index].isKnown = true;
								else
									-- отметить, что информация о перемещеиях группы
									-- от наводчика не поступает
									if not _is_group_visibly then 
										self.targets[_target_index].isKnown = false;
										
										self.artilleryObservers[_artillery_observer_index].artilleryObserver:lostVisualContact({ 
											group = self.targets[_target_index].group });
									end
								end
							else
								-- если нет, добавим в список для удаления:
								table.insert(_targets_to_remove, _target_index);
							end
						end
					end -- конец цикла перебора целей
					
					if not _is_known_group then
						-- не известная группа:
						-- проверяем в пределах ли она досигаемости
						if _distance_to_artillery < self.maxEngageDistance or self.maxEngageDistance == 0 then
							-- mission.model.messages.push('target: ' ..  _blueforce_ground_groups[_blueforce_group_index].name 
							--	.. ', distance to artillery: ' .. _distance_to_artillery .. ', max distance: ' .. self.maxEngageDistance); 
							-- если да, проверим видит ли её наводчик
							if _distance_to_observer < _visibility_range and land.isVisible(
								{ x = _observer_position.p.x, y = (_observer_position.p.y + 10), z = _observer_position.p.z}, 
								{ x = _dcsw_group_position.p.x, y = (_dcsw_group_position.p.y + 10), z = _dcsw_group_position.p.z }) then
								-- если видит: 
								local _dcsw_group_name = _dcsw_group:getName();
									
								if not mission.model.groups[_dcsw_group_name] then
									mission.model:addGroup(_dcsw_group_name);
								end
									
								-- добавить группу в списко целей:
								self:addTarget(mission.model.groups[_dcsw_group_name]);
									
								self.artilleryObservers[_artillery_observer_index].artilleryObserver:spot({ 
									group = mission.model.groups[_dcsw_group_name], isNew = true });
							end
						end
					end
					
					-- удаление целей вышедших из зоны досягаемости:
					for _target_to_remove_index = 1, #_targets_to_remove do
						table.remove(self.targets, _target_to_remove_index);
					end
				end
			end
		end
		if coalition.side.BLUE == _artillery_observer_coalition and _is_observer_alive then
			for _redforce_country_index = 1, #env.mission.coalition.red.country do
				
				local _redforce_ground_groups = {};
				
				if env.mission.coalition.red.country[_redforce_country_index].vehicle then
					if env.mission.coalition.red.country[_redforce_country_index].vehicle.group then 
						_redforce_ground_groups = env.mission.coalition.red.country[_redforce_country_index].vehicle.group;
					end
				end
				
				for _redforce_group_index = 1, #_redforce_ground_groups do
					local _dcsw_group = Group.getByName(
						_redforce_ground_groups[_redforce_group_index].name);
						
					local _get_observer_position_status, _observer_position = pcall(function() 
						return _artillery_observer_dcsw_unit:getPosition(); end);
					
					if not _get_observer_position_status then return 0; end;
					
					local _artillery_units = self.dcsWorldGroup:getUnits();
					
					if #_artillery_units < 1 then return 0; end;

					local _artillery_position = _artillery_units[1]:getPosition();
					
					local _dcsw_group_units = _dcsw_group:getUnits();
					
					if #_dcsw_group_units < 1 then return 0; end
					
					local _dcsw_group_position = _dcsw_group_units[1]:getPosition();
					
					local _distance_to_observer = mission.utils.getDistance(_observer_position.p, _dcsw_group_position.p);
					local _distance_to_artillery = mission.utils.getDistance(_artillery_position.p, _dcsw_group_position.p);
					
					local _targets_to_remove = {};
					local _is_group_visibly = false; -- признак того, чьл группу не видит ни один из наводчиков
					local _is_known_group = false;
					
					for _target_index = 1, #self.targets do
						local _target_dcsw_group = self.targets[_target_index].group:getDCSWorldGroup();
						local _target_dcsw_group_id = _target_dcsw_group:getID();
						local _dcsw_group_id = _dcsw_group:getID();
						
						if _dcsw_group_id == _target_dcsw_group_id then 
							-- если такая группа уже есть:
							_is_known_group = true;
							-- проверяем в пределах ли она досигаемости
							if _distance_to_artillery < self.maxEngageDistance or self.maxEngageDistance == 0 then
								-- если да, проверим видит ли её наводчик
								if _distance_to_observer < _visibility_range and land.isVisible(
									{ x = _observer_position.p.x, y = (_observer_position.p.y + 10), z = _observer_position.p.z}, 
									{ x = _dcsw_group_position.p.x, y = (_dcsw_group_position.p.y + 10), z = _dcsw_group_position.p.z }) then
									
									-- отметить, что информация о перемещеиях группы
									-- от наводчика поступает
									if not _is_group_visibly then 
										self.artilleryObservers[_artillery_observer_index].artilleryObserver:spot({ 
											group = self.targets[_target_index].group, isNew = false });
									end
									
									_is_group_visibly = true;
									self.targets[_target_index].isKnown = true;
								else
									-- отметить, что информация о перемещеиях группы
									-- от наводчика не поступает
									if not _is_group_visibly then 
										self.targets[_target_index].isKnown = false;
										
										self.artilleryObservers[_artillery_observer_index].artilleryObserver:lostVisualContact({ 
											group = self.targets[_target_index].group });
									end
								end
							else
								-- если нет, добавим в список для удаления:
								table.insert(_targets_to_remove, _target_index);
							end
						end
					end -- конец цикла перебора целей
					
					if not _is_known_group then
						-- не известная группа:
						-- проверяем в пределах ли она досигаемости
						if _distance_to_artillery < self.maxEngageDistance or self.maxEngageDistance == 0 then
							-- если да, проверим видит ли её наводчик
							if _distance_to_observer < _visibility_range and land.isVisible(
								{ x = _observer_position.p.x, y = (_observer_position.p.y + 10), z = _observer_position.p.z}, 
								{ x = _dcsw_group_position.p.x, y = (_dcsw_group_position.p.y + 10), z = _dcsw_group_position.p.z }) then
								-- если видит: 
								local _dcsw_group_name = _dcsw_group:getName();
									
								if not mission.model.groups[_dcsw_group_name] then
									mission.model:addGroup(_dcsw_group_name);
								end
									
								-- добавить группу в списко целей:
								self:addTarget(mission.model.groups[_dcsw_group_name]);
									
								self.artilleryObservers[_artillery_observer_index].artilleryObserver:spot({ 
									group = mission.model.groups[_dcsw_group_name], isNew = true });
							end
						end
					end
					
					-- удаление целей вышедших из зоны досягаемости:
					for _target_to_remove_index = 1, #_targets_to_remove do
						table.remove(self.targets, _target_to_remove_index);
					end
				end
			end
		end
	end -- конец цикла по арт наводчикам
	
	-- mission.model.messages.push('Отладка: Все цели получены. Число целей: ' .. #self.targets);
	
	local _total_count_targets = #self.targets;
	
	return _total_count_targets;
end;

function msfGroup:markTargetsBySmoke(_target)
	
	if nil ~= self.tasks.msf_MarkBySmoke then
		self.tasks.msf_MarkBySmoke:Stop();
		self.tasks.msf_MarkBySmoke = nil;
	end
	
	self.tasks.msf_MarkBySmoke = Factory.create(msfTask, 
		{ functionToCall = function (_task_arguments)
		
			local _dcsw_target_group = _task_arguments.target:getDCSWorldGroup();
			local _target_units = _dcsw_target_group:getUnits();
			local _count_target_units = #_target_units;
			
			local _smoke_color = trigger.smokeColor.White;
			
			local _artillery_units = self.dcsWorldGroup:getUnits();
			
			if #_artillery_units < 1 then return nil; end
			
			local _artillery_coalition = _artillery_units[1]:getCoalition();
			
			if coalition.side.BLUE == _artillery_coalition then 
				_smoke_color = trigger.smokeColor.Red;
			end
			if coalition.side.RED == _artillery_coalition then 
				_smoke_color = trigger.smokeColor.Blue;
			end
			
			if _count_target_units > 0 then 
				local _first_unit_position = _target_units[1]:getPosition();
				local _last_unit_position = _target_units[_count_target_units]:getPosition();
				
				trigger.action.smoke(_first_unit_position.p, _smoke_color);
				trigger.action.smoke(_last_unit_position.p, _smoke_color);
			end
			return nil;
		end,
		functionArgument = { target = _target }, 
		startTime = (timer.getTime() + 100000) });
end;

function msfGroup:requestArtilleryOnTargets(_targets, _range, _round_time, _start_time, _mark_by_smoke)
	
	if nil ~= self.tasks.msf_FireMission then 
		self.tasks.msf_FireMission:Stop();
		self.tasks.msf_FireMission = nil;
	end
	
	if not _range.maximalDistance then 
		_range.maximalDistance = 0;
	else
		self.maxEngageDistance = _range.maximalDistance;
	end;
	
	for _target_inex = 1, #_targets do
		table.insert(self.targets, { group = _targets[_target_inex], point = {x = 1, y = 1}, isKnown = true });
	end
	
	self.tasks.msf_FireMission = Factory.create(msfTask, 
		{ functionToCall = function (_task_arguments)
			self:prepareToShot({});
			
			local _engaging_units = self.dcsWorldGroup:getUnits();
			
			-- mission.model.messages.push('Отладка: артиллерия на связи.');
			
			if #_engaging_units < 1 then return nil; end
			
			local _count_enemies_groups = #self.targets;
			
			-- mission.model.messages.push('Отладка: число целей для атаки: ' .. _count_enemies_groups);
			
			if _count_enemies_groups > 0 then  
				local _engaged_group_index = math.random(_count_enemies_groups);
				local _engaged_group_status, _engaged_group = pcall(function() 
						return self.targets[_engaged_group_index].group:getDCSWorldGroup(); 
					end);
				
				if _engaged_group_status then 
				
					local _engaged_groups_units = _engaged_group:getUnits();
					
					if (#_engaged_groups_units > 0) then
						local _engaged_unit_index = math.random(#_engaged_groups_units);
						local _engaged_unit_position = _engaged_groups_units[_engaged_unit_index]:getPosition();
						local _radius = _task_arguments.maxRadius * math.random();
						
						if _radius < _task_arguments.minRadius then 
							_radius = _task_arguments.minRadius;
						end
						
						local _engaged_unit_position_x = 0;
						local _engaged_unit_position_y = 0;
						
						if self.targets[_engaged_group_index].isKnown then 
							_engaged_unit_position_x, _engaged_unit_position_y = _engaged_unit_position.p.x, _engaged_unit_position.p.z;
							
							self.targets[_engaged_group_index].point.x = _engaged_unit_position.p.x; 
							self.targets[_engaged_group_index].point.y = _engaged_unit_position.p.z;
						else
							_engaged_unit_position_x = self.targets[_engaged_group_index].point.x;
							_engaged_unit_position_y = self.targets[_engaged_group_index].point.y;
						end
						
						local _engaging_units_position = _engaging_units[1]:getPosition(); 
						
						local _distance = mission.utils.getDistance(_engaging_units_position.p, 
							{ x = _engaged_unit_position_x, z = _engaged_unit_position_y });
							
						if _task_arguments.maxDistance < 1 or _distance < _task_arguments.maxDistance then 
							local _fire_task  = { id = 'FireAtPoint', params = {
									point = { x = _engaged_unit_position_x, y = _engaged_unit_position_y },
									radius = _radius;
								}
							};
							
							local _horizontal_orientation = { 
								x = _engaging_units_position.x.x + _engaging_units_position.z.x, 
								y = _engaging_units_position.x.z + _engaging_units_position.z.z };
							
							local _horizontal_angle_rad = math.acos(
								(_horizontal_orientation.x * _engaged_unit_position_x + _horizontal_orientation.y * _engaged_unit_position_y) /
								(math.sqrt((_horizontal_orientation.x * _horizontal_orientation.x + _horizontal_orientation.y * _horizontal_orientation.y)
									+  (_engaged_unit_position_x * _engaged_unit_position_x + _engaged_unit_position_y * _engaged_unit_position_y))));
							
							local _horizontal_angle = _horizontal_angle_rad * 180 / math.pi;
							local _protractor_ext = 0;
							
							if _horizontal_angle < 180 then 
								_protractor_ext, _ = math.modf(_horizontal_angle / 6);
							else
								_protractor_ext, _ = math.modf(_horizontal_angle / 6);
								_protractor_ext = 60 - _protractor_ext;
							end
							
							local _protractor_int, _ = math.modf(((_horizontal_angle % 6) * 100) / 6);
							
							local _controller = self.dcsWorldGroup:getController();
							
							_controller:setTask(_fire_task);
							
							_distance, _ = math.modf(_distance); 
							
							self:readyToShot({ target = _engaged_group, 
								protractor = { ext = _protractor_ext, int = _protractor_int }, 
								distance = _distance / 1000 });
								
							if _mark_by_smoke then 
								self.targetMarkedBySmoke = true;
								self:markTargetsBySmoke(self.targets[_engaged_group_index].group);
							end
						else
							table.remove(self.targets, _engaged_group_index);
						end
					else
						table.remove(self.targets, _engaged_group_index);
					end
				end
				return (timer.getTime() + _task_arguments.roundTime);
			else
				local _cease_fire_task = { id = 'NoTask', params = {} };
				local _controller = self.dcsWorldGroup:getController();
				
				_controller:setTask(_cease_fire_task);
				
				return (timer.getTime() + _task_arguments.roundTime);
			end
		end, 
		functionArgument = { maxDistance = _range.maximalDistance, minRadius = _range.minimal, maxRadius = _range.maximal, roundTime = _round_time }, 
		startTime = (timer.getTime() + _start_time) });
end;

function msfGroup:isClearToMoveOn() 
	if (false == self.groupHasStopped) then
		return true;
	end;
	
	return false; 
end;

function msfGroup:holdOn()
	local _dcsw_group = self:getDCSWorldGroup();
	if not _dcsw_group then 
		env.info('could not call msfGroup:holdOn() for ' .. self.groupName .. ', cause group does not exist', false);
		return nil;
	end
	
	local _controller = _dcsw_group:getController();
	self.groupHasStopped = true;
	
	_controller:setTask({
		id = 'ControlledTask',
		params = { 
			task = { id = 'Hold', params = {} }, 
			stopCondition = { condition = 'mission.model.groups.' .. _dcsw_group:getName() .. ':isClearToMoveOn();' } 
		}
	});
end;

function msfGroup:holdOnForAWhile(_delay)
	local _dcsw_group = self:getDCSWorldGroup();
	if not _dcsw_group then 
		env.info('MSF Info: could not execute msfGroup:holdOnForAWhile() for ' .. self.groupName .. ', cause group does not exist', false);
		return nil;
	end
	local _controller = _dcsw_group:getController();
	
	self.groupHasStopped = true;
	_controller:setTask({
		id = 'ControlledTask',
		params = { 
			task = { id = 'Hold', params = {} }, 
			stopCondition = { 
				duration = _delay,
				condition = 'mission.model.groups.' .. _dcsw_group:getName() .. ':isClearToMoveOn();' }
		}
	});
end;

function msfGroup:moveOn()
	self.groupHasStopped = false;
end;

function msfGroup:setInvisible(_invisible)
	local _dcsw_group = self:getDCSWorldGroup();
	if not _dcsw_group then 
		env.info('MSF Info: could not execute msfGroup:setInvisible() for ' .. self.groupName .. ', cause group does not exist', false);
		return nil;
	end
	local _controller = _dcsw_group:getController();
	_controller:setCommand({ id = 'SetInvisible', params = { value = _invisible }});
end;

function msfGroup:setImmortal(_immortal)
	local _dcsw_group = self:getDCSWorldGroup();
	if not _dcsw_group then 
		env.info('MSF Info: could not execute msfGroup:setImmortal() for ' .. self.groupName .. ', cause group does not exist', false);
		return nil;
	end
	local _controller = _dcsw_group:getController();
	_controller:setCommand({ id = 'SetImmortal', params = { value = _immortal }});
end;

function msfGroup:atEase()
	local _dcsw_group = self:getDCSWorldGroup();
	if not _dcsw_group then 
		env.info('MSF Info: could not execute msfGroup:atEase() for ' .. self.groupName .. ', cause group does not exist', false);
		return nil;
	end
	local _controller = _dcsw_group:getController();
	_controller:setOption(9, 1);
end;

function msfGroup:stayAlert()
	local _dcsw_group = self:getDCSWorldGroup();
	if not _dcsw_group then 
		env.info('MSF Info: could not execute msfGroup:stayAlert() for ' .. self.groupName .. ', cause group does not exist', false);
		return nil;
	end
	local _controller = _dcsw_group:getController();
	_controller:setOption(9, 0);
end;

function msfGroup:danger()
	local _dcsw_group = self:getDCSWorldGroup();
	if not _dcsw_group then 
		env.info('MSF Info: could not execute msfGroup:danger() for ' .. self.groupName .. ', cause group does not exist', false);
		return nil;
	end
	local _controller = _dcsw_group:getController();
	_controller:setOption(9, 2);
end;

function msfGroup:holdFire()
	local _dcsw_group = self:getDCSWorldGroup();
	if not _dcsw_group then 
		env.info('MSF Info: could not execute msfGroup:holdFire() for ' .. self.groupName .. ', cause group does not exist', false);
		return nil;
	end
	local _controller = _dcsw_group:getController();
	_controller:setOption(0, 4);
end;

function msfGroup:shootBack()
	local _dcsw_group = self:getDCSWorldGroup();
	if not _dcsw_group then 
		env.info('MSF Info: could not execute msfGroup:shootBack() for ' .. self.groupName .. ', cause group does not exist', false);
		return nil;
	end
	local _controller = _dcsw_group:getController();
	_controller:setOption(0, 3);
end;

function msfGroup:engageAtWill()
	local _dcsw_group = self:getDCSWorldGroup();
	if not _dcsw_group then 
		env.info('MSF Info: could not execute msfGroup:engageAtWill() for ' .. self.groupName .. ', cause group does not exist', false);
		return nil;
	end
	local _controller = _dcsw_group:getController();
	_controller:setOption(0, 0);
end;

function msfGroup:addOnReadyToShot(eventHandler)
	self.onReadyToShot[eventHandler] = eventHandler;
end;

function msfGroup:removeOnReadyToShot(eventHandler)
	self.onReadyToShot[eventHandler] = nil;
end;

function msfGroup:addOnPrepareToShot(eventHandler) 
	self.onPrepareToShot[eventHandler] = eventHandler;
end;

function msfGroup:removeOnPrepareToShot(eventHandler)
	self.onPrepareToShot[eventHandler] = nil;
end;

function msfGroup:addOnPlayerEnterUnit(eventHandler)
	self.onPlayerEnterUnit[eventHandler] = eventHandler;
end;

function msfGroup:removeOnPlayerEnterUnit(eventHandler)
	self.onPlayerEnterUnit[eventHandler] = nil;
end;

function msfGroup:addOnPlayerLeaveUnit(eventHandler)
	self.onPlayerLeaveUnit[eventHandler] = eventHandler;
end;

function msfGroup:removeOnPlayerLeaveUnit(eventHandler)
	self.onPlayerLeaveUnit[eventHandler] = nil;
end;

function msfGroup:addOnInitEventHandler(eventHandler) 
	self.onInit[eventHandler] = eventHandler;
end;

function msfGroup:removeOnInitEventHandler(eventHandler)
	self.onInit[eventHandler] = nil;
end;

function msfGroup:addOnHitEventHandler(eventHandler)
	self.onHit[eventHandler] = eventHandler;
end;

function msfGroup:removeOnHitEventHandler(eventHandler)
	self.onHit[eventHandler] = nil;
end;

function msfGroup:addOnShotEventHandler(eventHandler)
	self.onShot[eventHandler] = eventHandler;
end;

function msfGroup:removeOnShotEventHandler(eventHandler)
	self.onShot[eventHandler] = nil;
end;

function msfGroup:addOnDeadEventHandler(eventHandler)
	self.onDead[eventHandler] = eventHandler;
end;

function msfGroup:removeOnDeadEventHandler(eventHandler)
	self.onDead[eventHandler] = nil;
end;

function msfGroup:addOnTakeOffEventHandler(eventHandler)
	self.onTakeOff[eventHandler] = eventHandler;
end;

function msfGroup:removeOnTakeOffEventHandler(eventHandler)
	self.onTakeOff[eventHandler] = nil;
end;

function msfGroup:shot(eventArgs)
	for _onShotEventHandlerKey, _onShotEventHandler in pairs(self.onShot) do
		_onShotEventHandler(self, eventArgs);
	end
end;

function msfGroup:dead(eventArgs)
	for _onDeadEventHandlerKey, _onDeadEventHandler in pairs(self.onDead) do
		_onDeadEventHandler(self, eventArgs);
	end
end;

function msfGroup:playerEnterUnit(eventArgs)
	for _onPlayerEnterUnitEventHandlerKey, _onPlayerEnterUnitEventHandler in pairs(self.onPlayerEnterUnit) do
		_onPlayerEnterUnitEventHandler(self, eventArgs);
	end
end;

function msfGroup:playerLeaveUnit(eventArgs)
	for _onPlayerLeaveUnitEventHandlerKey, _onPlayerLeaveUnitEventHandler in pairs(self.onPlayerLeaveUnit) do
		_onPlayerLeaveUnitEventHandler(self, eventArgs);
	end
end;

-- msfGargoPlace:
msfCargoPlace = {};

function msfCargoPlace:construct(parameters)
	self.point = { x = parameters.point.x, y = land.getHeight({ x = parameters.point.x, y = parameters.point.z }), z = parameters.point.z };
	self.radius = parameters.radius;
	self.coalition = parameters.coalition;
	self.name = parameters.name;
	-- self.countryId = parameters.countryId;
	self.cargoTemplates = {}; 
	self.loadingThreads = {};	-- unitId, time, _template
end;

function msfCargoPlace:addCargoTemplate(_name, _unit_type, _count, _group_template, _load_time, _unload_time, _unload_conditions, _country, _display_name)
	table.insert(self.cargoTemplates, {
		unitType = _unit_type, 
		count = _count, 
		template = _group_template, 
		loadTime =  _load_time, 
		unloadTime = _unload_time,
		conditions = _unload_conditions,
		countryId = _country,
		name = _name,
		displayName = _display_name
	});
end;

function msfCargoPlace:update(_delay)
	local _volume = { id = world.VolumeType.SPHERE, params = {  -- dont forgot back to SPHERE after fix!!!
		point = self.point,
		radius = self.radius} };
	local _found_units = {};
	
	world.searchObjects(Object.Category.UNIT, _volume, function(_object, _data) 
		local _msf_unit = nil;
		local _object_type = _object:getTypeName();
		local _unit_id = _object:getID();
		local _unit_name = _object:getName();
		
		table.insert(_data.foundUnits, _unit_id);
		
		for _unit_key, _unit in pairs(mission.model.units) do
			if _unit_id == _unit:getUnitID() then 
				_msf_unit = _unit; break; 
			end
		end
		if not _msf_unit then 
			mission.model:addUnit(_unit_name);
			_msf_unit = mission.model.units[_unit_name];
		end
		
		for _index = 1, #_data.templates do  
			if _object_type == _data.templates[_index].unitType and _data.templates[_index].count > 0 
				and _data.templates[_index].name == _msf_unit:getSelectedCargoTemplateName() then 
				local _unit_cargo = _msf_unit:getCargo();
				
				if not _unit_cargo.units then -- если юнит уже не загружен
					-- проверка хода загрузки: 
					local _unit_thread_index = 0;
					for _thread_index = 1, #_data.threads do
						if _unit_id == _data.threads[_thread_index].unitId then _unit_thread_index = _thread_index; break; end
					end
					
					-- проверка высоты:
					local _group = _object:getGroup();
					local _unit_point = _object:getPoint();
					
					if math.abs(_unit_point.y - _data.point.y) < 4.5 then
						if _unit_thread_index > 0 then
							local _loading_time = _data.threads[_unit_thread_index].loadingTime;
							if _loading_time < _data.templates[_index].loadTime then 
								local _remaining_time = 0;
								if _data.templates[_index].loadTime > _loading_time then
									_remaining_time = _data.templates[_index].loadTime - _loading_time;
								end
								trigger.action.outTextForGroup(_group:getID(), 'Выполняется погрузка: ' .. _data.templates[_index].displayName
									.. '. Осталось ' .. _remaining_time .. ' секунд.', _data.delay);
								_data.threads[_unit_thread_index].loadingTime = _loading_time + _data.delay;
							else
								trigger.action.outTextForGroup(_group:getID(), 'Погрузка: ' .. _data.templates[_index].displayName .. ' успешно завершена.', _data.delay);
								table.remove(_data.threads, _unit_thread_index);
								_msf_unit:loadCargo({ 
									units = _data.templates[_index].template, 
									unloadTime = _data.templates[_index].unloadTime,
									groupName = _data.templates[_index].name .. '_' .. _data.templates[_index].count,
									cargoPlace = { point = _data.point, radius = _data.radius },
									countryId = _data.templates[_index].countryId,
									conditions = { 
										altitudeLessThen = _data.templates[_index].conditions.altitudeLessThen,
										speedLessThen = _data.templates[_index].conditions.speedLessThen,
										command = _data.templates[_index].conditions.command } 
									});
								_data.templates[_index].count = (_data.templates[_index].count - 1);
								
								-- перенесено внутрь метода msfUnit:loadCargo()
								--[[if Group.Category.GROUND == _group:getCategory() then 
									if _msf_unit:isExistRadioMenuItem('msfCargo_' .. _msf_unit:getUnitName() .. '_pickUpGroup') then 
										_msf_unit:removeRadioCommandByName('msfCargo_' .. _msf_unit:getUnitName() .. '_pickUpGroup');
									end	
									if _msf_unit:isExistRadioMenuItem('msfCargo_' .. _group:getName() .. '_' .. _msf_unit:getUnitName() .. '_linkGroupsToPickUp') then 
										_msf_unit:removeRadioCommandByName('msfCargo_' .. _group:getName() .. '_' .. _msf_unit:getUnitName() .. '_linkGroupsToPickUp');
									end
								end --]]
							end
						else
							trigger.action.outTextForGroup(_group:getID(), 'Начало погрузки: ' .. _data.templates[_index].name, _data.delay);
							table.insert(_data.threads, { unitId = _unit_id, loadingTime = 0, template = _data.templates[_index], groupId = _group:getID() });
						end
					else
						if _unit_thread_index > 0 then 
							table.remove(_data.threads, _unit_thread_index);
							trigger.action.outTextForGroup(_group:getID(), 'Погрузка: ' .. _data.templates[_index].name .. ' прервана. Не удерживается высота.', _data.delay);
						end
					end
				end
				break;
			end
		end
		return true;
	end, { templates = self.cargoTemplates, threads = self.loadingThreads, point = self.point, radius = self.radius, delay = _delay, foundUnits = _found_units });
	
	local _thread_to_delete = {};
	
	for _index = 1, #self.loadingThreads do
		local _thread_active = false;
		for _unit_index = 1, #_found_units do 
			if _found_units[_unit_index] == self.loadingThreads[_index].unitId then
				_thread_active = true;
				break;
			end
		end
		if not _thread_active then 
			table.insert(_thread_to_delete, _index); 
			trigger.action.outTextForGroup(self.loadingThreads[_index].groupId, 'Погрузка прервана. Выход за пределы зоны погрузки.',  _delay);
		end
	end
	for _index = 1, #_thread_to_delete do
		table.remove(self.loadingThreads, _thread_to_delete[_index]);
	end
end; 

function mission.controller:hasAirDefenceUnitDead(_air_defence_groupments, _dead_unit_id)
	error('MSF Error: mission.controller:hasAirDefenceUnitDead(_air_defence_groupments, _dead_unit_id) is not implemented yet');
	return false, nil, nil;
end;

-- работа ПВО:
function mission.controller:airDefenceWorkCycle(_air_defence_groupments)
	error('MSF Error: mission.controller:airDefenceWorkCycle(_air_defence_groupments) is not implemented yet');
end;

function mission.controller:mainCycle(_arguments)	
	-- debug:
	--[[ local _iteration_start_time_status, _iteration_start_time = pcall(function()
			local _start_time = os.clock();
			return _start_time;
		end);
	if not _iteration_start_time_status then 
		env.error('MSF Error: error on attempt to execute os.clock(): ' .. tostring(_iteration_start_time), false);
		_iteration_start_time = 0;
	end --]]
	-- end debug

	mission.controller.iteration = timer.getTime();
	
	if (true == self.stopMainCycle) then 
		return nil;
	end
	
	for _unit_key, _unit in pairs(mission.model.units) do
		local _update_unit_status, _update_unit_result = pcall(function() _unit:update();  return 0; end); -- close _pcall function
		
		if not _update_unit_status then 
			trigger.action.outText('MSF ERROR: Epic fail with update unit cycle. Error:\n\r' .. _update_unit_result, 20);
			env.error('MSF ERROR: Epic fail with update unit cycle. Error:\n\r' .. _update_unit_result, false);
		end
	end
	
	local _refresh_radio_menu_status, _refresh_radio_menu_result = pcall(function()
			for _unit_key, _unit in pairs(mission.model.units) do
				_unit:refreshRadioMenu();
			end
			return 0; 
		end);
	if not _refresh_radio_menu_status then 
		trigger.action.outText('MSF ERROR: Epic fail with function _unit:refreshRadioMenu(). Error:\n\r' .. _refresh_radio_menu_result, 20);
		env.error('MSF ERROR: Epic fail with function _unit:refreshRadioMenu(). Error:\n\r' .. _refresh_radio_menu_result, false);
	end
	
	local _update_pickuping_threads_status, _update_pickuping_threads_result = pcall(function()
			local _pickuping_threads_to_stop = {};
			for _pickuping_thread_index, _pickuping_thread_context in ipairs(mission.model.pickupingThreads) do
				if mission.controller.updatePickupingThread(_pickuping_thread_context) then table.insert(_pickuping_threads_to_stop, _pickuping_thread_index);
				end
			end
			for _index, _thread_index_to_stop in ipairs(_pickuping_threads_to_stop) do
				table.remove(mission.model.pickupingThreads, _thread_index_to_stop);
			end
			return 0;
		end);
		
	if not _update_pickuping_threads_status then 
		trigger.action.outText('MSF ERROR: Epic fail with function mission.controller.updatePickupingThread(). Error:\n\r' .. _update_pickuping_threads_result, 20);
		env.error('MSF ERROR: Epic fail with function mission.controller.updatePickupingThread(). Error:\n\r' .. _update_pickuping_threads_result, false);
	end
	
	local _update_cargoplaces_status, _update_cargoplaces_result = pcall(function()
			for _cargo_place_key, _cargo_place in pairs(mission.model.cargoPlaces) do
				_cargo_place:update(2);
			end
			return 0;
		end);
	if not _update_cargoplaces_status then 
		trigger.action.outText('MSF ERROR: Epic fail with function _cargo_place:update(). Error:\n\r' .. _update_cargoplaces_result, 20);
		env.error('MSF ERROR: Epic fail with function _cargo_place:update(). Error:\n\r' .. _update_cargoplaces_result, false);
	end
	
	local _unload_cargo_status, _unload_cargo_result = pcall(function()
			for _unit_key, _unit in pairs(mission.model.units) do
				_unit:tryUnloadCargo(2);
			end
			return 0;
		end);
	if not _unload_cargo_status then 
		trigger.action.outText('MSF ERROR: Epic fail with function _unit:tryUnloadCargo(). Error:\n\r' .. _unload_cargo_result, 20);
		env.error('MSF ERROR: Epic fail with function _unit:tryUnloadCargo(). Error:\n\r' .. _unload_cargo_result, false);
	end
	
	local _air_defence_status, _air_defence_error = xpcall(function() 
			for _air_defence_groupment_key, _air_defence_groupment in pairs(mission.model.airDefenceGroupments) do _air_defence_groupment:update(); end
			return 0; 
		end, 
		function(_air_defence_error) return _air_defence_error .. '\n\r' .. debug.traceback(); end);
	if not _air_defence_status then
		trigger.action.outText('MSF ERROR: Epic fail with air defence groupments update, cause of ' 
			.. assert(_air_defence_error, 'air defence groupment update has failed, but did not return error information'), 30);
		env.error('MSF ERROR: Epic fail with air defence groupments update, cause of ' 
			.. assert(_air_defence_error, 'air defence groupment update has failed, but did not return error information'), false);
	end
	
	local _display_text_status,  _display_text_error = pcall(function() mission.view.displayText(); return 0; end);
	
	if not _display_text_status then 
		trigger.action.outText('MSF ERROR: Epic fail with function mission.view.displayText(). Error:\n\r' .. _display_text_error, 20);
		env.error('MSF ERROR: Epic fail with function mission.view.displayText(). Error:\n\r' .. _display_text_error, false);
	end
	
	local _flags_status, _ = pcall(function()
		for _flag_key, _flag in pairs(mission.model.flags) do
			local _old_flag_value = _flag:getValue();
			
			_flag:update();
			
			local _flag_value = _flag:getValue();
			
			if ((0 == _old_flag_value or nil == _old_flag_value) and 1 == _flag_value) then
				_flag:flagOn({});
			end
			if (1 == _old_flag_value and (0 == _flag_value or nil == _flag_value)) then
				_flag:flagOff({});
			end
			if (_old_flag_value ~= _flag_value) then
				_flag:change({ newValue = _flag_value, oldValue = _old_flag_value });
			end
		end
		
		return 0;
	end);
	
	if not _flags_status then
		mission.view.printDebugMessage('MSF ERROR: Epic fail with mission.model.flags test cycle');
		env.error('MSF ERROR: Epic fail with mission.model.flags test cycle', false);
	end
	
	
	local _zones_status, _ = pcall(function()
		for _zone_key, _zone in pairs(mission.model.zones) do
			for _unit_key, _unit in pairs(mission.model.units) do
				if (false == _zone:isUnitInsideZone(_unit)) then
					if (true == _unit:isUnitInsideZone(_zone:getDCSWorldTriggerZone())) then
						_zone:unitEnter(_unit);
					end
				else
					if (false == _unit:isUnitInsideZone(_zone:getDCSWorldTriggerZone())) then
						_zone:unitLeave(_unit);
					end
				end
			end
		end
		
		return 0;
	end);
	
	if not _flags_status then
		env.error('MSF ERROR: Epic fail with mission.model.zones test cycle', false);
	end
	
	-- debug:
	--[[ local _iteration_time_status, _iteration_time = pcall(function()
			mission.controller.iterationTime = os.clock() - _iteration_start_time;
			return mission.controller.iterationTime;
		end);
	if not _iteration_time_status then 
		env.error('MSF Error: error on attempt to execute os.clock(): ' .. tostring(_iteration_time), false);
		mission.controller.iterationTime = 0;
	end --]]
	-- end debug
	
	return (timer.getTime() + 2);
end;

-- controller:
function mission.controller:start() 
	self.stopMainCycle = false;
	
	for _group_key, _group in pairs(mission.model.groups) do
		_group:init({});
	end
	
	env.info('MSF INFO: mission start', false);
	timer.scheduleFunction(self.mainCycle, {}, (timer.getTime() + 1)); 
	-- timer.scheduleFunction(mission.controller.mainCycleStorage, {}, (timer.getTime() + 32)); 
end;

function mission.controller:stop()
	slef.stopMainCycle = true;
end;

function mission.controller:beginPhase(phaseName)
	if (nil ~= mission.model.phases[phaseName]) then 
		mission.model.phases[phaseName]:activate({});
	else
		trigger.action.outText("error: could not find phase with name: [" .. phaseName .. "]", 10);
	end
end;

function mission.controller:endPhase(phaseName) 
	if (nil ~= mission.model.phases[phaseName]) then 
		mission.model.phases[phaseName]:deactivate({});
	else
		trigger.action.outText("error: could not find phase with name: [" .. phaseName .. "]", 10);
	end
end;

function mission.model:addUnit(unitName)
	self.units[unitName] = Factory.create(msfUnit, { newUnit = Unit.getByName(unitName), newUnitName = unitName });
	
	for _coalition_key, _coalition in pairs(env.mission.coalition) do
		for _country_index, _country in ipairs(_coalition.country) do
			if _country.helicopter and _country.helicopter.group then 
				for _group_index, _group in ipairs(_country.helicopter.group) do
					if _group.task == 'Transport' then 
						for _unit_index, _unit in ipairs(_group.units) do
							if _unit.name == unitName then 
								self.units[unitName]:setTransportTask(true); 
								break;
							end
						end
					end
				end
			end
			if _country.plane and _country.plane.group then
				for _group_index, _group in ipairs(_country.plane.group) do
					if _group.task == 'Intercept' then 
						for _unit_index, _unit in ipairs(_group.units) do
							if _unit.name == unitName then 
								self.units[unitName]:setInterceptorTask(true); 
								break;
							end
						end
					end
				end
			end
		end
	end
end;

function mission.model:addGroup(groupName)
	self.groups[groupName] = Factory.create(msfGroup, { newGroup = Group.getByName(groupName), newGroupName = groupName });
end;

function mission.model:createGroup(_category, _country, _descriptor)
	local _new_group = coalition.addGroup(_country, _category, _descriptor);
	self.groups[_descriptor.name] = Factory.create(msfGroup, { newGroup = _new_group, newGroupName = _descriptor.name });
	
	return self.groups[_descriptor.name];
end;

function mission.model:addPhase(_phaseName, _description)
	mission.model.phases[_phaseName] = Factory.create(msfMissionPhase, { description = _description });
end; 

function mission.model:addCargoPlace(_name, _point, _radius, _coalition)
	table.insert(self.cargoPlaces, Factory.create(msfCargoPlace, { name = _name, point = _point, radius = _radius, coalition = _coalition }));
	
	local _new_cargo_place_id = #self.cargoPlaces;
	return _new_cargo_place_id;
end;

function mission.model:addZone(_zoneName)
	parameters = {};
	parameters.newTriggerZone = trigger.misc.getZone(_zoneName);
	
	if (nil ~= parameters.newTriggerZone) then
		self.zones[_zoneName] = Factory.create(msfZone, parameters);
	else
		trigger.action.outText("error: could not find trigger zone with name: \'" .. _zoneName .. "\'", 10);
	end
end;

function mission.model:addAirDefenceGroupment(_name, _display_name, _is_activated, _auto_deactivation)
	self.airDefenceGroupments[_name] = Factory.create(msfAirDefenceGroupment, { 
		name = _name
		, displayName = _display_name
		, parent = nil
		, isActivated = _is_activated 
		, deactivateIfNoTargets = _auto_deactivation });
end;

function mission.model:addFlag(_name, _number, _value)
	self.flags[_name] = Factory.create(msfFlag, { flagName = _name, flagNumber = _number, flagValue = _value });
end;

function mission.model:addSound(_key, _path, _length)
	self.sounds[_key] = { path = _path, length = _length };
end;

local _customEventHandler = {};

local function _msfOnSelectCargoPlaceTemplate(_data)
	_data.unit:setSelectedCargoTemplateName(_data.cargoTemplateName);
end;
local function _msfOnResetCargoPlaceTemplate(_data)
	_data.unit:resetSelectedCargoTemplateName();
end;

function _customEventHandler:onEvent(event)
	
	if (world.event.S_EVENT_ENGINE_STARTUP == event.id) then
		local _status, _initiator_name = pcall(function() local _name = event.initiator:getName(); return _name; end);
		if not _status then
			env.error('MSF ERROR: Could not get initiator name of event S_EVENT_ENGINE_STARTUP, cause of ' .. tostring(_initiator_name), false);
			return;
		end
		if mission.model.units[_initiator_name] and mission.model.units[_initiator_name]:getDCSWorldUnit() then
			local _event_handler_status, _event_handler_result = pcall(function() 
					mission.model.units[_initiator_name]:engineStartUp(event); return 0;
				end);
			if not _event_handler_status then 
				env.error('MSF ERROR: MSF event ' .. tostring(_initiator_name) 
					.. ':engineStartUp() failed, cause of ' .. tostring(_event_handler_result), false);
			end
		end
	end
	
	if (world.event.S_EVENT_CRASH == event.id) then
		local _status, _initiator_name = pcall(function() local _name = event.initiator:getName(); return _name; end);
		if not _status then
			env.error('MSF ERROR: Could not get initiator name of event S_EVENT_CRASH, cause of ' .. tostring(_initiator_name), false);
			return;
		end
		if mission.model.units[_initiator_name] then mission.model.units[_initiator_name]:clearCargo();
		end
	end
	
	if (world.event.S_EVENT_EJECTION == event.id) then
		local _status, _initiator_name = pcall(function() local _name = event.initiator:getName(); return _name; end);
		if not _status then
			env.error('MSF ERROR: Could not get initiator name of event S_EVENT_EJECTION, cause of ' .. tostring(_initiator_name), false);
			return;
		end
		if mission.model.units[_initiator_name] then mission.model.units[_initiator_name]:clearCargo();
		end
	end
	
	if (world.event.S_EVENT_PILOT_DEAD == event.id) then
		local _status, _initiator_name = pcall(function() local _name = event.initiator:getName(); return _name; end);
		if not _status then
			env.error('MSF ERROR: Could not get initiator name of event S_EVENT_PILOT_DEAD, cause of ' .. tostring(_initiator_name), false);
			return;
		end
		if mission.model.units[_initiator_name] then 
			local _event_handler_status, _event_handler_result = pcall(function()
					mission.model.units[_initiator_name]:dead(); return 0; 
				end);
			if not _event_handler_status then 
				env.error('MSF ERROR: MSF event ' .. tostring(_initiator_name) .. ':dead() failed, cause of ' .. tostring(_event_handler_result), false);
			end
			mission.model.units[_initiator_name]:clearCargo();
		end
	end
	
	if (world.event.S_EVENT_DEAD == event.id) then
		local _status, _initiator_name = pcall(function() local _name = event.initiator:getName(); return _name; end);
		if not _status then
			env.error('MSF ERROR: Could not get initiator name of event S_EVENT_DEAD, cause of ' .. tostring(_initiator_name), false);
			return;
		end
		local _group_status, _initiator_group_name  = pcall(function() 
				local _initiator_group = event.initiator:getGroup();
				local _group_name = _initiator_group:getName();
				return _group_name;
			end);
		if not _group_status then 
			env.error('MSF ERROR: Could not get initiator group name of event S_EVENT_DEAD, cause of ' .. tostring(_initiator_group_name), false);
			return;
		end
		if mission.model.units[_initiator_name] then 
			local _unit = mission.model.units[_initiator_name];
			local _event_handler_status, _event_handler_result = pcall(function() 
					_unit:dead(event); return 0;
				end);
			if not _event_handler_status then 
				env.error('MSF ERROR: MSF event ' .. tostring(_initiator_name) .. ':dead() failed, cause of ' .. tostring(_event_handler_result), false);
			end
			_unit:clearCargo();
			mission.controller.removeRadioCommandsForTransport(_unit);
		end
		if mission.model.groups[_initiator_group_name] then 
			local _group = mission.model.groups[_initiator_group_name];
			local _event_handler_status, _event_handler_result = pcall(function() 
					_group:dead(event); return 0;
				end);
			if not _event_handler_status then 
				env.error('MSF ERROR: MSF event ' 
					.. tostring(_initiator_group_name) .. ':dead() failed, cause of ' .. tostring(_event_handler_result), false);
			end
		end
	end
	
	if (world.event.S_EVENT_BIRTH == event.id) then
		env.info('S_EVENT_BIRTH has been raised', false);
	end
	
	if (world.event.S_EVENT_LAND == event.id) then
		env.info('S_EVENT_LAND has been raised', false);
		local _status, _initiator_name = pcall(function() local _name = event.initiator:getName(); return _name; end);
		if not _status then
			env.error('MSF ERROR: Could not get initiator name of event S_EVENT_TAKEOFF, cause of ' .. tostring(_initiator_name), false);
			return;
		end
		local _initiator_group_status, _initiator_group_name = pcall(function() 
				local _dcsw_group = event.initiator:getGroup();
				local _dcsw_group_name = _dcsw_group:getName();
				local _dcsw_group_coalition = _dcsw_group:getCoalition();
				return _dcsw_group_name;
			end);
		if not _initiator_group_status then
			env.error('MSF ERROR: Could not get initiator group name on event S_EVENT_TAKEOFF, cause of ' .. tostring(_initiator_group_name), false);
			return;
		end
	end
	
	if (world.event.S_EVENT_TAKEOFF == event.id) then
		local _status, _initiator_name = pcall(function() local _name = event.initiator:getName(); return _name; end);
		if not _status then
			env.error('MSF ERROR: Could not get initiator name of event S_EVENT_TAKEOFF, cause of ' .. tostring(_initiator_name), false);
			return;
		end		
		local _initiator_group_status, _initiator_group_name = pcall(function() 
				local _dcsw_group = event.initiator:getGroup();
				local _dcsw_group_name = _dcsw_group:getName();
				return _dcsw_group_name;
			end);
		if not _initiator_group_status then
			env.error('MSF ERROR: Could not get initiator group name on event S_EVENT_TAKEOFF, cause of ' .. tostring(_initiator_group_name), false);
			return;
		end
		if not mission.model.groups[_initiator_group_name] then
			mission.model:addGroup(_initiator_group_name);
		end
		if mission.model.groups[_initiator_group_name] and mission.model.groups[_initiator_group_name]:getDCSWorldGroup() then
			do
				local _event_handler_status, _event_handler_result = pcall(function()
						mission.model.groups[_initiator_group_name]:takeOff(event); return 0;
					end);
				if not _event_handler_status then 
					env.error('MSF ERROR: MSF event '.. tostring(_initiator_group_name) 
						.. ':takeOff() failed, cause of ' .. tostring(_event_handler_result), false);
				end
			end
		end
		local _unit = mission.model.units[_initiator_name];
		if mission.model.units[_initiator_name] and mission.model.units[_initiator_name]:getDCSWorldUnit() then 
			do
				local _event_handler_status, _event_handler_result = pcall(function()
						_unit:takeOff(event); return 0;
					end);
				if not _event_handler_status then 
					env.error('MSF ERROR: MSF event ' .. tostring(_initiator_name) 
						.. ':takeOff() failed, cause of ' .. tostring(_event_handler_result), false);
				end
			end
		end -- if unit exists
	end
	
	if (world.event.S_EVENT_PLAYER_LEAVE_UNIT == event.id) then
		env.info('S_EVENT_PLAYER_LEAVE_UNIT has been raised', false);
	end
	
	if (world.event.S_EVENT_PLAYER_ENTER_UNIT == event.id) then
		env.info('S_EVENT_PLAYER_ENTER_UNIT has been raised', false);
	end
	
	if (world.event.S_EVENT_HIT == event.id) then				
		do -- handle units
			local _status, _event_data = pcall(function()
					local _target_name, _initiator_name = event.target:getName(), event.initiator:getName();
					return { targetName = _target_name, initiatorName = _initiator_name };
				end);
			if _status then
				if mission.model.units[_event_data.targetName] and mission.model.units[_event_data.targetName]:getDCSWorldUnit() then 
					do 
						local _event_handler_status, _event_handler_result = pcall(function() 
								mission.model.units[_event_data.targetName]:hit(event); return 0;	
							end);
						if not _event_handler_status then 
							env.error('MSF ERROR: MSF event ' .. tostring(_event_data.targetName)
								.. ':hit() failed, cause of ' .. tostring(_event_handler_result), false);
						end
					end
				end
				if mission.model.units[_event_data.initiatorName] and mission.model.units[_event_data.initiatorName]:getDCSWorldUnit() then
					do
						local _event_handler_status, _event_handler_result = pcall(function() 
								mission.model.units[_event_data.initiatorName]:hitting(event); return 0;	
							end);
						if not _event_handler_status then 
							env.error('MSF ERROR: MSF event ' .. tostring(_event_data.initiatorName)
								.. ':hitting() failed, cause of ' .. tostring(_event_handler_result), false);
						end
					end
				end
			else
				env.error('MSF ERROR: Could not get initiator or target name on event S_EVENT_HIT, cause of ' .. tostring(_event_data), false);
			end
		end
		do -- handle groups
			local _status, _event_data = pcall(function()
					local _target_group, _initiator_group = event.target:getGroup(), event.initiator:getGroup();
					local _target_name, _initiator_name = _target_group:getName(), _initiator_group:getName();
					return { targetGroupName = _target_name, initiatorGroupName = _initiator_name };
				end);
			if _status then 
				if mission.model.groups[_event_data.targetGroupName] and mission.model.groups[_event_data.targetGroupName]:getDCSWorldGroup() then 
					do 
						local _event_handler_status, _event_handler_result = pcall(function() 
								mission.model.groups[_event_data.targetGroupName]:hit(event); return 0;	
							end);
						if not _event_handler_status then 
							env.error('MSF ERROR: MSF event ' .. tostring(_event_data.targetGroupName)
								.. ':hit() failed, cause of ' .. tostring(_event_handler_result), false);
						end
					end
				end
				if mission.model.groups[_event_data.initiatorGroupName] and mission.model.groups[_event_data.initiatorGroupName]:getDCSWorldGroup() then 
					do 
						local _event_handler_status, _event_handler_result = pcall(function() 
								-- mission.model.groups[_event_data.initiatorGroupName]:hitting(event); 
								return 0;	
							end);
						if not _event_handler_status then 
							env.error('MSF ERROR: MSF event ' .. tostring(_event_data.initiatorGroupName)
								.. ':hitting() failed, cause of ' .. tostring(_event_handler_result), false);
						end
					end
				end
			else
				env.error('MSF ERROR: Could not get initiator or target group name on event S_EVENT_HIT, cause of ' .. tostring(_event_data), false);
			end
		end
	end
	
	if (world.event.S_EVENT_SHOT == event.id or world.event.S_EVENT_SHOOTING_START == event.id) then
		do -- handle units
			local _status, _initiator_name = pcall(function()
					local _name = event.initiator:getName(); return _name;
				end);
			if _status then 
				if mission.model.units[_initiator_name] and mission.model.units[_initiator_name]:getDCSWorldUnit() then
					local _event_handler_status, _event_handler_result = pcall(function() mission.model.units[_initiator_name]:shot(event); return 0;
						end);
					if not _event_handler_status then 
						env.error('MSF ERROR: MSF event ' .. tostring(_initiator_name) 
							.. ':shot() failed, cause of ' .. tostring(_event_handler_result), false);
					end
				end
			else
				env.error('MSF ERROR: Could not get initiator name of event S_EVENT_SHOT, cause of ' .. tostring(_initiator_name), false);
			end
		end
		do -- handle groups
			local _status, _initiator_group_name = pcall(function()
					local _initiator_group = event.initiator:getGroup(); 
					local _name = _initiator_group:getName();
					return _name;
				end);
			if _status then 
				if mission.model.groups[_initiator_group_name] and mission.model.groups[_initiator_group_name]:getDCSWorldGroup() then 
					local _group = mission.model.groups[_initiator_group_name];
					if _group:isEnableMarkTargetsBySmoke() and _group.tasks.msf_MarkBySmoke then 
						_group:enableMarkTargetBySmoke(false);
						_group.tasks.msf_MarkBySmoke:StartAfter(5);
					end
					
					local _event_handler_status, _event_handler_result = pcall(function() mission.model.groups[_initiator_group_name]:shot(event); return 0;
						end);
					if not _event_handler_status then 
						env.error('MSF ERROR: MSF event ' .. tostring(initiator_group_name)
							.. ':shot() failed, cause of ' .. tostring(_event_handler_result), false);
					end
				end
			else
				env.error('MSF ERROR: Could not get initiator group name of event S_EVENT_SHOT, cause of ' .. tostring(_initiator_group_name), false);
			end
		end
	end
	
	if (world.event.S_EVENT_REFUELING == event.id) then env.info('started refueling', false); -- mission.model.messages.push('started refueling');
	end
	
	if (world.event.S_EVENT_REFUELING_STOP == event.id) then env.info('stoped refueling', false); -- mission.model.messages.push('stoped refueling');
	end
end;

world.addEventHandler(_customEventHandler);

--trigger.action.outText("msf_internal_version.lua file, version: 1.0.2.0144.52 changed 15.05.2015", 5);

-- version: a.b.c.dddd.r
-- r - revision number
-- dddd - bug fixes
-- c - minor update
-- b - major update
-- a - epic update
-- changed by Chichco